/*
===============================================================================

	FILE:	arithmeticencoder.cpp
	
	CONTENTS:
			
    see corresponding header file

  PROGRAMMERS:
	
		martin isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		(c) 2001 martin isenburg@cs.unc.edu
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
    see corresponding header file
	
===============================================================================
*/
#include "arithmeticencoder.h"

#include "arithmeticmodel.h"

#include <iostream.h>

ArithmeticEncoder::ArithmeticEncoder()
{
	bits = new int[10000];
	
	number_bits = 0;
	allocated_bits = 10000;

	low = 0;
	high = 65535; // Top_value = 65535
	fbits = 0;
}

ArithmeticEncoder::~ArithmeticEncoder()
{
	if (bits)
	{
		delete [] bits;
	}
}

void ArithmeticEncoder::encode(ArithmeticModel* am, int sym)
{
	if (sym<0 || sym>=am->nsym)
	{
		cerr << "encode: symbol out of range" << endl;
	}

	__int64 low_to_high = (__int64)(high-low)+1;
	high = low + (low_to_high*am->cfreq[sym])/am->cfreq[0]-1;
	low = low + (low_to_high*am->cfreq[sym+1])/am->cfreq[0];

	encode();
	
	am->update(sym);
}
			
void ArithmeticEncoder::encode(ArithmeticModel* am, int sym, int low_sym)
{
	if (low_sym<0 || low_sym>=am->nsym)
	{
		cerr << "encode: lowest symbol out of range. low_sym: " << low_sym << endl;
	}
	if (sym<low_sym || sym>=am->nsym)
	{
		cerr << "encode: symbol out of range. sym: " << sym << " low_sym: " << low_sym << endl;
	}

	__int64 low_to_high = (__int64)(high-low)+1;
	high = low + (low_to_high*am->cfreq[sym])/am->cfreq[low_sym]-1;
	low = low + (low_to_high*am->cfreq[sym+1])/am->cfreq[low_sym];

	encode();
	
	am->update(sym);
}

void ArithmeticEncoder::encode(int range, int sym)
{
	if (sym<0 || sym>=range)
	{
		cerr << "encode: symbol out of range" << endl;
	}

	__int64 low_to_high = (__int64)(high-low)+1;
	high = low + (low_to_high*(range-sym))/range-1;
	low = low + (low_to_high*(range-sym-1))/range;
	
	encode();
}

void ArithmeticEncoder::done()
{
	fbits += 1;
	if (low < 16384)
	{
		bit_plus_follow(0);
	}
	else
	{
		bit_plus_follow(1);
	}
}
		
int* ArithmeticEncoder::getBits()
{
	int* tmp_bits = bits;
	bits = 0;
	return tmp_bits;
}
	
int ArithmeticEncoder::getNumberBits()
{
	return number_bits;
}

void ArithmeticEncoder::encode()
{
	for (;;)
	{
		if (high<32768) // Half = 32768
		{
			bit_plus_follow(0);
		}
		else if (low>=32768)
		{
			bit_plus_follow(1);
			low -= 32768;
			high -= 32768;
		}
		else if (low>=16384 && high<49152) // First_qtr = 16384  Third_qtr = 49152
		{
			fbits += 1;
			low -= 16384;
			high -= 16384;
		}
		else
		{
			break;
		}
		low = 2*low;
		high = 2*high+1;
	}
}
	
void ArithmeticEncoder::output_bit(int bit)
{
	if (number_bits == allocated_bits)
	{
		int* newbits = new int[allocated_bits*2];
		for (int i = 0; i < allocated_bits; i++)
		{
			newbits[i] = bits[i];
		}
		delete [] bits;
		bits = newbits;
		allocated_bits = allocated_bits*2;
	}
	bits[number_bits++] = bit;
}

void ArithmeticEncoder::bit_plus_follow(int bit)
{
	output_bit(bit);
	while (fbits > 0)
	{
		output_bit(!bit);
		fbits -= 1;
	}
}
