/*
===============================================================================

	FILE:	codec.h
	
	CONTENTS:
	
    An interface that is used by encoder and decoder to turn code symbols into
    a compressed bit-stream and vice-versa by calling the arithmetic coder 
	
	PROGRAMMERS:
	
		martin isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		(c) 2002 martin isenburg@cs.unc.edu
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
		20 Feb 2002 -- created
	
===============================================================================
*/
#ifndef CF_CODEC_H
#define CF_CODEC_H

class ArithmeticModel;
class ArithmeticEncoder;
class ArithmeticDecoder;

class CFcodec
{
public:
	int be;
	int ie;

	int ib_on;
	int ib_off;
	int av_on;
	int av_off;

	int* boundary_degree_dist;
	int* interior_degree_dist;

	ArithmeticEncoder* ae_conn;
	ArithmeticEncoder* ae_geom;

	ArithmeticDecoder* ad;

	void initEnc(int boundary_degrees, int* has_boundary_degree, int interior_degrees, int* has_interior_degree, int pos_bits);
	void initDec();

	void doneEnc();
	void doneDec();

	void writeBoundaryEdgeDegree(int degree);
	void writeLargeBoundaryEdgeDegree(int degree);
	void writeInteriorEdgeDegree(int degree);
	void writeLargeInteriorEdgeDegree(int degree);
	void writeIsBoundary(int truefalse);
	void writeAlreadyVisited(int truefalse);
	void writeRange(int range, int index);
  
	void writePositionCorrector(int* corr);
	void writePositionCorrectorOther(int* corr);

  CFcodec();
  ~CFcodec();

protected:
	// prob tables for connectivity
	ArithmeticModel* amBoundaryEdgeDegree;
	ArithmeticModel* amLargeBoundaryEdgeDegree;
	ArithmeticModel* amInteriorEdgeDegree;
	ArithmeticModel* amLargeInteriorEdgeDegree;
	ArithmeticModel* amIsBoundary;
	ArithmeticModel* amAlreadyVisited;

	// prob tables for geometry
	ArithmeticModel* amSign;
	ArithmeticModel* amBits[5];

	ArithmeticModel* amSignOther;
	ArithmeticModel* amBitsOther[5];

  int bitsMask[5];
	int bitsUsed[5];
};

#endif
