/*
===============================================================================

	FILE:	encoder.cpp
	
	CONTENTS:
	
    see corresponding header file
	
	PROGRAMMERS:
	
		martin isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		(c) 2002 martin isenburg@cs.unc.edu
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
    see corresponding header file
	
===============================================================================
*/
#include <iostream.h>

#include "encoder.h"

#include "mesh.h"
#include "vector.h"
#include "vertex.h"
#include "spinedge.h"
#include "codec.h"
#include "positionquantizer.h"
#include "vec3fv.h"

CFencoder::CFencoder()
{
	verticesTETdone = 0;
	verticesTETcount = 0;
	verticesTETnumber = 0;

	verticesHEXdone = 0;
	verticesHEXcount = 0;
	verticesHEXnumber = 0;

	facesTET[0] = facesTET[1] = facesTET[2] = facesTET[3] = facesTET[4] = 0;
	verticesTET = 0;
	facesHEX[0] = facesHEX[1] = facesHEX[2] = facesHEX[3] = facesHEX[4] = 0;
	verticesHEX = 0;
}


CFencoder::~CFencoder()
{
}

CFspinEdge* CFencoder::pick_best_gateTET(CFmesh* mesh)
{
	int i;
  CFspinEdge* e;
  CFspinEdge* e_min = NULL;

  int min_edge_slot = 100000;
	int max_zero_slot = 0;

	for (i = 0; i < mesh->total_edges; i++)
	{
		e = (CFspinEdge*)mesh->es->elementAt(i);
		if ((e->slots != -1) && (e->inv->group != -1))
		{
			if (e->slots <= min_edge_slot)
			{
				if ((e->slots < min_edge_slot) || (((e->slots == 0) + (e->next->slots == 0) + (e->next->next->slots == 0)) > max_zero_slot))
				{
					e_min = e;
					min_edge_slot = e_min->slots;
					max_zero_slot = ((e->slots == 0) + (e->next->slots == 0) + (e->next->next->slots == 0));
				}
			}
		}
	}
	if (min_edge_slot == 0 && max_zero_slot && e_min->next->next->slots == 0)
	{
		e_min = e_min->next->next;
	}
	return (e_min ? e_min->inv : NULL);
}

CFspinEdge* CFencoder::pick_best_gateHEX(CFmesh* mesh)
{
	int i;
  CFspinEdge* e;
  CFspinEdge* e_min = NULL;

  int min_edge_slot = 100000;
	int max_zero_slot = 0;

	for (i = 0; i < mesh->total_edges; i++)
	{
		e = (CFspinEdge*)mesh->es->elementAt(i);
		if ((e->slots != -1) && (!e->marker))
		{
			if (e->slots <= min_edge_slot)
			{
				if (min_edge_slot == 0)
				{
					if ((e->next->slots == 0) + (e->next->next->slots == 0) + (e->next->next->next->slots == 0) > max_zero_slot)
					{
						max_zero_slot = (e->next->slots == 0) + (e->next->next->slots == 0) + (e->next->next->next->slots == 0);
						e_min = e;
					}
				}
				else
				{
					min_edge_slot = e->slots;
					max_zero_slot = (e->next->slots == 0) + (e->next->next->slots == 0) + (e->next->next->next->slots == 0);
					e_min = e;
				}
			}
		}
	}
	if (e_min)
	{
		if (e_min->inv->group != -1 && e_min->next->next->next->slots == 0)
		{
			e_min = e_min->next->next->next;
		}
		if (e_min->inv->group != -1 && e_min->next->next->next->slots == 0)
		{
			e_min = e_min->next->next->next;
		}
		return e_min->inv;
	}
	else
	{
		return 0;
	}
}

int CFencoder::encodeStep(CFmesh* mesh, CFcodec* codec, PositionQuantizer* pq)
{
	if (mesh->ts->size())
	{
		return encodeStepTET(mesh, codec);
	}
	else
	{
		return encodeStepHEX(mesh, codec, pq);
	}
}

int CFencoder::encodeStepTET(CFmesh* mesh, CFcodec* codec)
{
  int i;
	int slots,degree;
	int is_on_boundary;
  CFspinEdge* run;
	CFspinEdge *e;
  int component_count = 1;

	CFspinEdge *e0;
	CFspinEdge *e1;
	CFspinEdge *e2;

	CFspinEdge *e00;
	CFspinEdge *e01;
	CFspinEdge *e02;

	CFspinEdge *e10;
	CFspinEdge *e11;
	CFspinEdge *e12;

	CFspinEdge *e20;
	CFspinEdge *e21;
	CFspinEdge *e22;

	bool face0adjacency;
	bool face1adjacency;
	bool face2adjacency;

	bool edge0adjacency; // at spinedges e21<->e01
	bool edge1adjacency; // at spinedges e02<->e12
	bool edge2adjacency; // at spinedges e10<->e20

	bool vertex3adjacency; // at spinedges e02, e10, e21

	if (mesh->gate == 0)
	{	
		case0 = case1 = case2a = case2b = case3a = case3b = case4 = 0;
		pred_null = pred_last = pred_ray = pred_para = pred_lorenzo = 0;
		dist_null = dist_last = dist_ray = dist_para = dist_lorenzo = 0.0f;
		facesTET[0] = facesTET[1] = facesTET[2] = facesTET[3] = facesTET[4] = 0;
		if (verticesTET)
		{
			delete [] verticesTET;
		}
		verticesTETdone = 0;
		verticesTETcount = 0;
		verticesTETnumber = mesh->vs->size();
		verticesTET = new CFspinEdge*[verticesTETnumber];
		for (i = 0; i < verticesTETnumber; i++)
		{
			verticesTET[i] = 0;
		}
		for (i = 0; i < mesh->total_edges; i++)
		{
			run = (CFspinEdge*)mesh->es->elementAt(i);
			run->component = 0;
			run->marker = 1;
			run->degree = 0;
			run->slots = -1;
			run->Lnext = 0;
			run->is_on_boundary = 0;
			run->vertex->active = 0;
		}
		
		for (i = 0; i < mesh->total_edges; i++)
		{
			run = (CFspinEdge*) mesh->es->elementAt(i);
			if (run->marker == 1)
			{
				degree = 0;
				is_on_boundary = 0;
				while (run->marker == 1)
				{
					degree++;
					run->marker = 0;
					run->inv->marker = 0;
					if (run->group == -1)
					{
						is_on_boundary = 1;
					}
					if ((run->vertex != run->spin->vertex) || (run->inv->vertex != run->inv->spin->vertex))
					{
						cout << "problem" << endl;
					}
					run = run->spin;
				}
				while (run->degree == 0)
				{
					run->degree = degree;
					run->inv->degree = degree;
					run->is_on_boundary = is_on_boundary;
					run->inv->is_on_boundary = is_on_boundary;
					run = run->spin;
				}
			}
		}

		run = 0; // search for a good border face (try to pick a corner -> face with lowest degree count)
		for (i = 0; i < mesh->total_edges; i++)
		{
			e = (CFspinEdge*) mesh->es->elementAt(i);
			if (e->inv->group == -1)
			{
				if ((run == 0) || (e->degree < run->degree))
				{
					run = e;
				}
			}
		}

		mesh->gate = run;

		e0 = run;
		e1 = e0->next->next;
		e2 = e0->next;

		e00 = e0->spin->inv;
		e11 = e1->spin->inv;
		e22 = e2->spin->inv;

// mark active vertices

		e0->vertex->active = 1;
		e1->vertex->active = 1;
		e2->vertex->active = 1;

// renumber vertices

		e0->vertex->index = verticesTETcount++;
		e1->vertex->index = verticesTETcount++;
		e2->vertex->index = verticesTETcount++;

// the starting face is *always* a boundary face

		if (e0->inv->group != -1)
		{
			cout << "ERROR: starting assumption violated" << endl;
			return 0;
		}

// mark its edges as *known* boundary edges

		e0->is_on_boundary = 2;
		e00->is_on_boundary = 2;
		e1->is_on_boundary = 2;
		e11->is_on_boundary = 2;
		e2->is_on_boundary = 2;
		e22->is_on_boundary = 2;

// record their edge degrees

		codec->writeBoundaryEdgeDegree(e0->degree);
		e0->slots = e0->inv->slots = e0->degree - 1;
		codec->writeBoundaryEdgeDegree(e1->degree);
		e1->slots = e1->inv->slots = e1->degree - 1;
		codec->writeBoundaryEdgeDegree(e2->degree);
		e2->slots = e2->inv->slots = e2->degree - 1;

		return 1;
	}

	if (mesh->gate->group == -1)
	{
		cout << "ERROR: i hit a border face as e0" << endl;
		return 0;
	}

	e0 = mesh->gate;
	e1 = e0->next->next;
	e2 = e0->next;

	e00 = e0->spin->inv;
	e01 = e00->next->next;
	e02 = e00->next;

	e11 = e1->spin->inv;
	e12 = e11->next->next;
	e10 = e11->next;

	e22 = e2->spin->inv;
	e20 = e22->next->next;
	e21 = e22->next;

	e0->component = component_count;
	e1->component = component_count;
	e2->component = component_count;
	e00->component = component_count;
	e01->component = component_count;
	e02->component = component_count;
	e10->component = component_count;
	e11->component = component_count;
	e12->component = component_count;
	e20->component = component_count;
	e21->component = component_count;
	e22->component = component_count;

	if (e0->group == -1)
	{
		cout << "i hit the border" << endl;
		return 0;
	}

	// check only
	if (e00->inv->component)
	{
		face0adjacency = true;
	}
	else
	{
		face0adjacency = false;
	}
	if (e11->inv->component)
	{
		face1adjacency = true;
	}
	else
	{
		face1adjacency = false;
	}
	if (e22->inv->component)
	{
		face2adjacency = true;
	}
	else
	{
		face2adjacency = false;
	}
	if ((e02->vertex != e10->vertex) || (e02->vertex != e21->vertex))
	{
		cout << "fourth vertex not the same for e02, e10, and e21" << endl;
	}
	if (e02->vertex->active)
	{
		vertex3adjacency = true;
	}
	e = e21->spin;
	edge0adjacency = false;
	while (e != e21)
	{
		if (e->slots != -1)
		{
			edge0adjacency = true;
			break;
		}
		e = e->spin;
	}
	e = e02->spin;
	edge1adjacency = false;
	while (e != e02)
	{
		if (e->slots != -1)
		{
			edge1adjacency = true;
			break;
		}
		e = e->spin;
	}
	e = e10->spin;
	edge2adjacency = false;
	while (e != e10)
	{
		if (e->slots != -1)
		{
			edge2adjacency = true;
			break;
		}
		e = e->spin;
	}
	if (face0adjacency)
	{
		if (e0->inv->slots != 0)
		{
			cout << "ERROR: e0->inv->slots should be zero: " << e0->inv->slots << endl;
		}
		if (!edge0adjacency || !edge1adjacency)
		{
			cout << "ERROR: edge0adjacency and edge1adjacency should be on" << endl;
		}
	}
	if (face1adjacency)
	{
		if (e1->inv->slots != 0)
		{
			cout << "ERROR: e1->inv->slots should be zero: " << e1->inv->slots << endl;
		}
		if (!edge1adjacency || !edge2adjacency)
		{
			cout << "ERROR: edge1adjacency and edge2adjacency should be on" << endl;
		}
	}
	if (face2adjacency)
	{
		if (e0->inv->slots != 0)
		{
			cout << "ERROR: e2->inv->slots should be zero: " << e2->inv->slots << endl;
		}
		if (!edge2adjacency || !edge0adjacency)
		{
			cout << "ERROR: edge2adjacency and edge0adjacency should be on" << endl;
		}
	}

// work

	if (!face0adjacency && !face1adjacency && !face2adjacency)
	{
		case0++;

// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e00->slots = e0->inv->slots - 1;
		e11->slots = e1->inv->slots - 1;
		e22->slots = e2->inv->slots - 1;

// update the old boundary spinedges on the other end with the decreased slot count

		e = e0->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e0->inv->slots)
		{
			cout << "other side of e0 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

		e = e1->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e1->inv->slots)
		{
			cout << "other side of e1 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

		e = e2->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e2->inv->slots)
		{
			cout << "other side of e2 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;

// propagate all boundary information (there is some avoidable redundancy with the following per face boundary specification here)

		if (e00->is_on_boundary && !e00->slots && e00->inv->group == -1)
		{
			e01->is_on_boundary = 2;
			e21->is_on_boundary = 2;

			e02->is_on_boundary = 2;
			e12->is_on_boundary = 2;
		}

		if (e11->is_on_boundary && !e11->slots && e11->inv->group == -1)
		{
			e12->is_on_boundary = 2;
			e02->is_on_boundary = 2;

			e10->is_on_boundary = 2;
			e20->is_on_boundary = 2;
		}

		if (e22->is_on_boundary && !e22->slots && e22->inv->group == -1)
		{
			e20->is_on_boundary = 2;
			e10->is_on_boundary = 2;

			e21->is_on_boundary = 2;
			e01->is_on_boundary = 2;
		}

// process free faces & edges (first those that are connected) in a smart order

// specify whether face f0 is a boundary face

		if (e00->is_on_boundary)
		{
			if (e00->slots)
			{
				if (e00->inv->group == -1)
				{
					codec->writeIsBoundary(1);

					e01->is_on_boundary = 2;
					e21->is_on_boundary = 2;

					e02->is_on_boundary = 2;
					e12->is_on_boundary = 2;
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e00
			}
		}

// specify whether face f1 is a boundary face

		if (e11->is_on_boundary)
		{
			if (e11->slots)
			{
				if (e11->inv->group == -1)
				{
					codec->writeIsBoundary(1);

					e12->is_on_boundary = 2;
					e02->is_on_boundary = 2;

					e10->is_on_boundary = 2;
					e20->is_on_boundary = 2;
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e11
			}
		}

// specify whether face f2 is a boundary face

		if (e22->is_on_boundary)
		{
			if (e22->slots)
			{
				if (e22->inv->group == -1)
				{
					codec->writeIsBoundary(1);

					e20->is_on_boundary = 2;
					e10->is_on_boundary = 2;

					e21->is_on_boundary = 2;
					e01->is_on_boundary = 2;
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e22
			}
		}

	// specify edge0

		if (!edge0adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e01->is_on_boundary)
			{
				if (e01->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e01->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e01->degree);
			}
			e01->slots = e21->slots = e01->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case0: edge0adjacency" << endl;
			e21->vertex->active = 2;

			e = e01->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e21->slots = slots;

			e = e21->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e01->slots = slots;
		}

// specify edge1

		if (!edge1adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e12->is_on_boundary)
			{
				if (e12->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e12->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e12->degree);
			}
			e12->slots = e02->slots = e12->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case0: edge1adjacency" << endl;
			e02->vertex->active = 2;

			e = e12->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e02->slots = slots;

			e = e02->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e12->slots = slots;
		}

		if (!edge2adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e20->is_on_boundary)
			{
				if (e20->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e20->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e20->degree);
			}
			e20->slots = e10->slots = e20->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case0: edge2adjacency" << endl;
			e10->vertex->active = 2;

			e = e20->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e10->slots = slots;

			e = e10->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e20->slots = slots;
		}

// process (and renumber) the vertices

		if (e02->vertex->active)
		{
			if (e02->vertex->active == 2)
			{
				e02->vertex->active = 1;
			}
			else
			{
				cout << "case0: vertex adjacency" << endl;
			}
		}
		else
		{
			e02->vertex->active = 1;
			e02->vertex->index = verticesTETcount++;
		}

// add the edges to their vertices

		addEdgeVertexTET(e00);
		addEdgeVertexTET(e01);
		addEdgeVertexTET(e02);

		addEdgeVertexTET(e10);
		addEdgeVertexTET(e11);
		addEdgeVertexTET(e12);

		addEdgeVertexTET(e20);
		addEdgeVertexTET(e21);
		addEdgeVertexTET(e22);

// add (inverse) faces to the priority list

		addFaceTET(e00);
		addFaceTET(e11);
		addFaceTET(e22);
	}
	else if (face0adjacency && !face1adjacency && !face2adjacency)
	{
// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e11->slots = e1->inv->slots - 1;
		e22->slots = e2->inv->slots - 1;
		e21->slots = e01->inv->slots - 1;
		e12->slots = e02->inv->slots - 1;

// update the old boundary spinedges on the other end with the decreased slot count

		e = e1->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e1->inv->slots)
		{
			cout << "other side of e1 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

		e = e2->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e2->inv->slots)
		{
			cout << "other side of e2 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

		e = e01->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e01->inv->slots)
		{
			cout << "other side of e01.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

		e = e02->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e02->inv->slots)
		{
			cout << "other side of e02.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;

		e00->inv->slots = -1;
		e01->inv->slots = -1;
		e02->inv->slots = -1;

// propagate all boundary information (there is some avoidable redundancy with the following per face boundary specification here)

		if ((e11->is_on_boundary && !e11->slots && e11->inv->group == -1) || (e12->is_on_boundary && !e12->slots && e12->inv->group == -1))
		{
			e10->is_on_boundary = 2;
			e20->is_on_boundary = 2;
		}

		if ((e22->is_on_boundary && !e22->slots && e22->inv->group == -1) || (e21->is_on_boundary && !e21->slots && e21->inv->group == -1))
		{
			e10->is_on_boundary = 2;
			e20->is_on_boundary = 2;
		}

// process free faces & edges (first those that are connected) in a smart order

// specify whether face f1 is a boundary face

		if (e11->is_on_boundary && e12->is_on_boundary)
		{
			if (e11->slots && e12->slots)
			{
				if (e11->inv->group == -1)
				{
					codec->writeIsBoundary(1);

					e10->is_on_boundary = 2;
					e20->is_on_boundary = 2;
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e11
			}
		}

// specify whether face f2 is a boundary face

		if (e22->is_on_boundary && e21->is_on_boundary)
		{
			if (e22->slots && e21->slots)
			{
				if (e22->inv->group == -1)
				{
					codec->writeIsBoundary(1);

					e20->is_on_boundary = 2;
					e10->is_on_boundary = 2;
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e22
			}
		}

// specify edge 2

		if (!edge2adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e20->is_on_boundary)
			{
				if (e20->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e20->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e20->degree);
			}
			e20->slots = e10->slots = e20->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case1: edge2adjacency" << endl;

			e = e20->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e10->slots = slots;

			e = e10->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e20->slots = slots;
		}

// add the edges to their vertices

		addEdgeVertexTET(e10);
		addEdgeVertexTET(e11);
		addEdgeVertexTET(e12);

		addEdgeVertexTET(e20);
		addEdgeVertexTET(e21);
		addEdgeVertexTET(e22);

// add (inverse) faces to the priority list

		addFaceTET(e11);
		addFaceTET(e22);
	}
	else if (face0adjacency && face1adjacency && !face2adjacency)
	{
// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e22->slots = e2->inv->slots - 1;
		e20->slots = e10->inv->slots - 1;
		e21->slots = e01->inv->slots - 1;

// update the old boundary spinedges on the other end with the decreased slot count

		e = e2->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e2->inv->slots)
		{
			cout << "other side of e2 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

		e = e10->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e10->inv->slots)
		{
			cout << "other side of e10.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

		e = e01->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e01->inv->slots)
		{
			cout << "other side of e01.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceTET(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;

		e00->inv->slots = -1;
		e01->inv->slots = -1;
		e02->inv->slots = -1;

		e10->inv->slots = -1;
		e11->inv->slots = -1;
		e12->inv->slots = -1;

// process free faces & edges (first those that are connected) in a smart order

// specify whether face f2 is a boundary face

		if (e22->is_on_boundary && e21->is_on_boundary  && e20->is_on_boundary)
		{
			if (e22->slots && e21->slots && e20->slots)
			{
				if (e22->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e22
			}
		}

// add the edges to their vertices

		addEdgeVertexTET(e20);
		addEdgeVertexTET(e21);
		addEdgeVertexTET(e22);

// add (inverse) faces to the priority list

		addFaceTET(e22);
	}
	else if (face0adjacency && face1adjacency && face2adjacency)
	{
		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;

		e00->inv->slots = -1;
		e01->inv->slots = -1;
		e02->inv->slots = -1;

		e10->inv->slots = -1;
		e11->inv->slots = -1;
		e12->inv->slots = -1;

		e20->inv->slots = -1;
		e21->inv->slots = -1;
		e22->inv->slots = -1;
	}
	else
	{
		cout << "oops" << endl;
	}

//	mesh->gate = pick_best_gateTET(mesh);

	mesh->gate = getFaceTET();
	if (mesh->gate == NULL) mesh->gate = getEdgeVertexTET();

	return (mesh->gate != NULL);
}

int CFencoder::encodeStepHEX(CFmesh* mesh, CFcodec* codec, PositionQuantizer* pq)
{
  int i;
	int degree;
	int slots;
	int is_on_boundary;
  CFspinEdge* run;
	CFspinEdge *e;
  int component_count = 1;

	bool face0adjacency;
	bool face1adjacency;
	bool face2adjacency;
	bool face3adjacency;
	bool face4adjacency;

	bool edge0adjacency; // at spinedges e01<->e32
	bool edge1adjacency; // at spinedges e12<->e03
	bool edge2adjacency; // at spinedges e23<->e10
	bool edge3adjacency; // at spinedges e30<->e21

	bool edge4adjacency; // at spinedges e4<->e20
	bool edge5adjacency; // at spinedges e5<->e13
	bool edge6adjacency; // at spinedges e6<->e02
	bool edge7adjacency; // at spinedges e7<->e31

	bool vertex4adjacency; // at spinedges e02, e10, e21
	bool vertex5adjacency; // at spinedges e02, e10, e21
	bool vertex6adjacency; // at spinedges e02, e10, e21
	bool vertex7adjacency; // at spinedges e02, e10, e21

	CFspinEdge *e0;
	CFspinEdge *e1;
	CFspinEdge *e2;
	CFspinEdge *e3;

	CFspinEdge *e00;
	CFspinEdge *e01;
	CFspinEdge *e02;
	CFspinEdge *e03;

	CFspinEdge *e11;
	CFspinEdge *e12;
	CFspinEdge *e13;
	CFspinEdge *e10;

	CFspinEdge *e22;
	CFspinEdge *e23;
	CFspinEdge *e20;
	CFspinEdge *e21;

	CFspinEdge *e33;
	CFspinEdge *e30;
	CFspinEdge *e31;
	CFspinEdge *e32;

	CFspinEdge *e4;
	CFspinEdge *e5;
	CFspinEdge *e6;
	CFspinEdge *e7;

	if (mesh->gate == 0)
	{
		case0 = case1 = case2a = case2b = case3a = case3b = case4 = 0;
		pred_null = pred_last = pred_ray = pred_para = pred_lorenzo = 0;
		dist_null = dist_last = dist_ray = dist_para = dist_lorenzo = 0.0f;
		facesHEX[0] = facesHEX[1] = facesHEX[2] = facesHEX[3] = facesHEX[4] = 0;
		if (verticesHEX)
		{
			delete [] verticesHEX;
		}
		verticesHEXdone = 0;
		verticesHEXcount = 0;
		verticesHEXnumber = mesh->vs->size();
		verticesHEX = new CFspinEdge*[verticesHEXnumber];
		for (i = 0; i < verticesHEXnumber; i++)
		{
			verticesHEX[i] = 0;
		}
		for (i = 0; i < mesh->total_edges; i++)
		{
			run = (CFspinEdge*)mesh->es->elementAt(i);
			run->component = 0;
			run->marker = 1;
			run->degree = 0;
			run->slots = -1;
			run->Lnext = 0;
			run->is_on_boundary = 0;
			run->vertex->active = 0;
		}
		for (i = 0; i < mesh->total_edges; i++)
		{
			run = (CFspinEdge*) mesh->es->elementAt(i);
			if (run->marker == 1)
			{
				degree = 0;
				is_on_boundary = 0;
				while (run->marker == 1)
				{
					degree++;
					run->marker = 0;
					run->inv->marker = 0;
					if (run->group == -1)
					{
						is_on_boundary = 1;
					}
					if ((run->vertex != run->spin->vertex) || (run->inv->vertex != run->inv->spin->vertex))
					{
						cout << "problem" << endl;
					}
					run = run->spin;
				}
				while (run->degree == 0)
				{
					run->degree = degree;
					run->inv->degree = degree;
					run->is_on_boundary = is_on_boundary;
					run->inv->is_on_boundary = is_on_boundary;
					run = run->spin;
				}
			}
		}
		run = 0; // search for a good start face, try to pick a corner (e.g. three boundary faces)
		for (i = 0; i < mesh->total_edges; i++)
		{
			e = (CFspinEdge*) mesh->es->elementAt(i);
			if (e->inv->group == -1)
			{
				if ((run == 0) || (e->degree < run->degree))
				{
					run = e;
				}
				if ((e->spin->group == -1) && (e->next->next->next->spin->group == -1))
				{
					run = e;
					break;
				}
			}
		}

		mesh->gate = run;

		e0 = run;
		e1 = e0->next->next->next;
		e2 = e0->next->next;
		e3 = e0->next;

		e00 = e0->spin->inv;
		e11 = e1->spin->inv;
		e22 = e2->spin->inv;
		e33 = e3->spin->inv;

// renumber vertices

		e0->vertex->index = verticesHEXcount++;
		e1->vertex->index = verticesHEXcount++;
		e2->vertex->index = verticesHEXcount++;
		e3->vertex->index = verticesHEXcount++;

// mark active vertices

		e0->vertex->active = 1;
		encodePosition0(codec,pq,e0->vertex);
		e1->vertex->active = 1;
		encodePosition1(codec,pq,e1->vertex,e0->vertex);
		e2->vertex->active = 1;
		encodePosition2(codec,pq,e2->vertex,e1->vertex);
		e3->vertex->active = 1;
		encodePosition3(codec,pq,e3->vertex,e2->vertex,e1->vertex,e0->vertex);

// the starting face is *always* a boundary face

		if (e0->inv->group != -1)
		{
			cout << "ERROR: starting assumption violated" << endl;
			return 0;
		}

// mark its edges as *known* boundary edges

		e0->is_on_boundary = 2;
		e00->is_on_boundary = 2;
		e1->is_on_boundary = 2;
		e11->is_on_boundary = 2;
		e2->is_on_boundary = 2;
		e22->is_on_boundary = 2;
		e3->is_on_boundary = 2;
		e33->is_on_boundary = 2;

// record their edge degrees

		codec->writeBoundaryEdgeDegree(e0->degree);
		e0->slots = e0->inv->slots = e0->degree - 1;
		codec->writeBoundaryEdgeDegree(e1->degree);
		e1->slots = e1->inv->slots = e1->degree - 1;
		codec->writeBoundaryEdgeDegree(e2->degree);
		e2->slots = e2->inv->slots = e2->degree - 1;
		codec->writeBoundaryEdgeDegree(e3->degree);
		e3->slots = e3->inv->slots = e3->degree - 1;

		return 1;
	}

	if (mesh->gate->group == -1)
	{
		cout << "ERROR: i hit a border face as e0" << endl;
		return 0;
	}

	e0 = mesh->gate;
	e1 = e0->next->next->next;
	e2 = e0->next->next;
	e3 = e0->next;

	e00 = e0->spin->inv;
	e01 = e00->next->next->next;
	e02 = e00->next->next;
	e03 = e00->next;

	e11 = e1->spin->inv;
	e12 = e11->next->next->next;
	e13 = e11->next->next;
	e10 = e11->next;

	e22 = e2->spin->inv;
	e23 = e22->next->next->next;
	e20 = e22->next->next;
	e21 = e22->next;

	e33 = e3->spin->inv;
	e30 = e33->next->next->next;
	e31 = e33->next->next;
	e32 = e33->next;

	e4 = e20->spin->inv;
	e5 = e4->next->next->next;
	e6 = e4->next->next;
	e7 = e4->next;

// mark visited spinedges
				
	e0->component = component_count;
	e1->component = component_count;
	e2->component = component_count;
	e3->component = component_count;

	e00->component = component_count;
	e01->component = component_count;
	e02->component = component_count;
	e03->component = component_count;

	e10->component = component_count;
	e11->component = component_count;
	e12->component = component_count;
	e13->component = component_count;

	e20->component = component_count;
	e21->component = component_count;
	e22->component = component_count;
	e23->component = component_count;

	e30->component = component_count;
	e31->component = component_count;
	e32->component = component_count;
	e33->component = component_count;

	e4->component = component_count;
	e5->component = component_count;
	e6->component = component_count;
	e7->component = component_count;

// check adjacencies

	if (e00->inv->component)
	{
		face0adjacency = true;
	}
	else
	{
		face0adjacency = false;
	}

	if (e11->inv->component)
	{
		face1adjacency = true;
	}
	else
	{
		face1adjacency = false;
	}

	if (e22->inv->component)
	{
		face2adjacency = true;
	}
	else
	{
		face2adjacency = false;
	}

	if (e33->inv->component)
	{
		face3adjacency = true;
	}
	else
	{
		face3adjacency = false;
	}

	if (e4->inv->component)
	{
		face4adjacency = true;
	}
	else
	{
		face4adjacency = false;
	}

	if (e4->vertex->active)
	{
		vertex4adjacency = true;
	}
	else
	{
		vertex4adjacency = false;
	}

	if (e5->vertex->active)
	{
		vertex5adjacency = true;
	}
	else
	{
		vertex5adjacency = false;
	}

	if (e6->vertex->active)
	{
		vertex6adjacency = true;
	}
	else
	{
		vertex6adjacency = false;
	}

	if (e7->vertex->active)
	{
		vertex7adjacency = true;
	}
	else
	{
		vertex7adjacency = false;
	}

	e = e4->spin;
	edge4adjacency = false;
	while (e != e4)
	{
		if (e->slots != -1)
		{
			edge4adjacency = true;
			break;
		}
		e = e->spin;
	}

	e = e5->spin;
	edge5adjacency = false;
	while (e != e5)
	{
		if (e->slots != -1)
		{
			edge5adjacency = true;
			break;
		}
		e = e->spin;
	}

	e = e6->spin;
	edge6adjacency = false;
	while (e != e6)
	{
		if (e->slots != -1)
		{
			edge6adjacency = true;
			break;
		}
		e = e->spin;
	}

	e = e7->spin;
	edge7adjacency = false;
	while (e != e7)
	{
		if (e->slots != -1)
		{
			edge7adjacency = true;
			break;
		}
		e = e->spin;
	}

	e = e01->spin;
	edge0adjacency = false;
	while (e != e01)
	{
		if (e->slots != -1)
		{
			edge0adjacency = true;
			break;
		}
		e = e->spin;
	}

	e = e12->spin;
	edge1adjacency = false;
	while (e != e12)
	{
		if (e->slots != -1)
		{
			edge1adjacency = true;
			break;
		}
		e = e->spin;
	}

	e = e23->spin;
	edge2adjacency = false;
	while (e != e23)
	{
		if (e->slots != -1)
		{
			edge2adjacency = true;
			break;
		}
		e = e->spin;
	}

	e = e30->spin;
	edge3adjacency = false;
	while (e != e30)
	{
		if (e->slots != -1)
		{
			edge3adjacency = true;
			break;
		}
		e = e->spin;
	}

// just a consistency check
	
	if (face0adjacency)
	{
		if (e0->inv->slots != 0)
		{
			cout << "ERROR: e0->inv->slots should be zero: " << e0->inv->slots << endl;
		}
		if (!edge0adjacency || !edge1adjacency)
		{
			cout << "ERROR: edge0adjacency and edge1adjacency should be on" << endl;
		}
		if (!vertex6adjacency || !vertex7adjacency)
		{
			cout << "ERROR: vertex6adjacency and vertex7adjacency should be on" << endl;
		}
	}
	if (face1adjacency)
	{
		if (e1->inv->slots != 0)
		{
			cout << "ERROR: e1->inv->slots should be zero: " << e1->inv->slots << endl;
		}
		if (!edge1adjacency || !edge2adjacency)
		{
			cout << "ERROR: edge1adjacency and edge2adjacency should be on" << endl;
		}
		if (!vertex5adjacency || !vertex6adjacency)
		{
			cout << "ERROR: vertex5adjacency and vertex6adjacency should be on" << endl;
		}
	}
	if (face2adjacency)
	{
		if (e2->inv->slots != 0)
		{
			cout << "ERROR: e2->inv->slots should be zero: " << e2->inv->slots << endl;
		}
		if (!edge2adjacency || !edge3adjacency)
		{
			cout << "ERROR: edge2adjacency and edge3adjacency should be on" << endl;
		}
		if (!vertex4adjacency || !vertex5adjacency)
		{
			cout << "ERROR: vertex4adjacency and vertex5adjacency should be on" << endl;
		}
	}
	if (face3adjacency)
	{
		if (e3->inv->slots != 0)
		{
			cout << "ERROR: e3->inv->slots should be zero: " << e3->inv->slots << endl;
		}
		if (!edge3adjacency || !edge0adjacency)
		{
			cout << "ERROR: edge3adjacency and edge0adjacency should be on" << endl;
		}
		if (!vertex7adjacency || !vertex4adjacency)
		{
			cout << "ERROR: vertex7adjacency and vertex4adjacency should be on" << endl;
		}
	}
	if (face4adjacency)
	{
		if (!edge4adjacency || !edge5adjacency || !edge6adjacency || !edge7adjacency)
		{
			cout << "ERROR: edge4adjacency, edge5adjacency, edge6adjacency, and edge7adjacency should be on" << endl;
		}
		if (!vertex4adjacency || !vertex5adjacency || !vertex6adjacency || !vertex7adjacency)
		{
			cout << "ERROR: vertex4adjacency, vertex5adjacency, vertex6adjacency, and vertex7adjacency should be on" << endl;
		}
	}

// here is the switch where we do the actual work

	if (!face0adjacency && !face1adjacency && !face2adjacency && !face3adjacency) // case 0
	{
		case0++;
// cout << "case 0" << endl;

// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e00->slots = e0->inv->slots - 1;
		e11->slots = e1->inv->slots - 1;
		e22->slots = e2->inv->slots - 1;
		e33->slots = e3->inv->slots - 1;

// update the old boundary spinedges on the other end with the decreased slot count

		e = e0->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e0->inv->slots)
		{
			cout << "other side of e0 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e1->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e1->inv->slots)
		{
			cout << "other side of e1 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
      updateFaceHEX(e);
		}

		e = e2->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e2->inv->slots)
		{
			cout << "other side of e2 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
      updateFaceHEX(e);
		}

		e = e3->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e3->inv->slots)
		{
			cout << "other side of e3 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
      updateFaceHEX(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;
		e3->inv->slots = -1;

// propagate all boundary information (there is some avoidable redundancy with the following per face boundary specification here)

		if (e00->is_on_boundary && !e00->slots && e00->inv->group == -1)
		{
			e01->is_on_boundary = 2;
			e32->is_on_boundary = 2;

			e02->is_on_boundary = 2;
			e6->is_on_boundary = 2;

			e03->is_on_boundary = 2;
			e12->is_on_boundary = 2;
		}

		if (e11->is_on_boundary && !e11->slots && e11->inv->group == -1)
		{
			e12->is_on_boundary = 2;
			e03->is_on_boundary = 2;

			e13->is_on_boundary = 2;
			e5->is_on_boundary = 2;

			e10->is_on_boundary = 2;
			e23->is_on_boundary = 2;
		}

		if (e22->is_on_boundary && !e22->slots && e22->inv->group == -1)
		{
			e23->is_on_boundary = 2;
			e10->is_on_boundary = 2;

			e20->is_on_boundary = 2;
			e4->is_on_boundary = 2;

			e21->is_on_boundary = 2;
			e30->is_on_boundary = 2;
		}

		if (e33->is_on_boundary && !e33->slots && e33->inv->group == -1)
		{
			e30->is_on_boundary = 2;
			e21->is_on_boundary = 2;

			e31->is_on_boundary = 2;
			e7->is_on_boundary = 2;

			e32->is_on_boundary = 2;
			e01->is_on_boundary = 2;
		}

// process free faces & edges (first those that are connected) in a smart order

// specify whether face f0 is a boundary face

		if (e00->is_on_boundary)
		{
			if (e00->slots)
			{
				if (e00->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e00
			}
			if (e00->inv->group == -1)
			{
				e01->is_on_boundary = 2;
				e32->is_on_boundary = 2;

				e02->is_on_boundary = 2;
				e6->is_on_boundary = 2;

				e03->is_on_boundary = 2;
				e12->is_on_boundary = 2;
			}
		}

// specify edge0

		if (!edge0adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e01->is_on_boundary)
			{
				if (e01->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e01->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e01->degree);
			}
			e01->slots = e32->slots = e01->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case0: edge0adjacency" << endl;
			e32->vertex->active = 2;

			e = e01->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e32->slots = slots;

			e = e32->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e01->slots = slots;
		}

// specify edge1

		if (!edge1adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e12->is_on_boundary)
			{
				if (e12->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e12->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e12->degree);
			}
			e12->slots = e03->slots = e12->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case0: edge1adjacency" << endl;
			e03->vertex->active = 2;

			e = e12->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e03->slots = slots;

			e = e03->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e12->slots = slots;
		}
				
// specify whether face f1 is a boundary face

		if (e11->is_on_boundary && e12->is_on_boundary)
		{
			if (e11->slots && e12->slots)
			{
				if (e11->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e11 or e12
			}
			if (e11->inv->group == -1)
			{
				e13->is_on_boundary = 2;
				e5->is_on_boundary = 2;

				e10->is_on_boundary = 2;
				e23->is_on_boundary = 2;
			}
		}

// specify edge2

		if (!edge2adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e23->is_on_boundary)
			{
				if (e23->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e23->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e23->degree);
			}
			e23->slots = e10->slots = e23->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case0: edge2adjacency" << endl;
			e10->vertex->active = 2;

			e = e23->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e10->slots = slots;

			e = e10->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e23->slots = slots;
		}

// specify whether face f2 is a boundary face

		if (e22->is_on_boundary && e23->is_on_boundary)
		{
			if (e22->slots && e23->slots)
			{
				if (e22->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across 22 or e23
			}
			if (e22->inv->group == -1)
			{
				e20->is_on_boundary = 2;
				e4->is_on_boundary = 2;

				e21->is_on_boundary = 2;
				e30->is_on_boundary = 2;
			}
		}

// specify edge3

		if (!edge3adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e30->is_on_boundary)
			{
				if (e30->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e30->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e30->degree);
			}
			e30->slots = e21->slots = e30->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case0: edge3adjacency" << endl;
			e21->vertex->active = 2;

			e = e30->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e21->slots = slots;

			e = e21->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e30->slots = slots;
		}

// specify whether face f3 is a boundary face

		if (e33->is_on_boundary && e30->is_on_boundary && e32->is_on_boundary)
		{
			if (e33->slots && e30->slots && e32->slots)
			{
				if (e33->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e33, e30, or e32 
			}
			if (e33->inv->group == -1)
			{
				e31->is_on_boundary = 2;
				e7->is_on_boundary = 2;
			}
		}

// the following if ... else ... switch makes handling case M a little 'cleaner'

		if (!face4adjacency)
		{
			codec->writeAlreadyVisited(0); // it was not case M

// process free faces & edges (now those that were (initially) connected)

// specify edge6

			if (!edge6adjacency)
			{
				codec->writeAlreadyVisited(0);
				if (e6->is_on_boundary)
				{
					if (e6->is_on_boundary != 2)
					{
						codec->writeIsBoundary(1); // i did not know that
					}
					codec->writeBoundaryEdgeDegree(e6->degree);
				}
				else
				{
					codec->writeIsBoundary(0);
					codec->writeInteriorEdgeDegree(e6->degree);
				}
				e6->slots = e02->slots = e6->degree - 2;
			}
			else
			{
				codec->writeAlreadyVisited(1);
				cout << "case0: edge6adjacency" << endl;
				e6->vertex->active = 2;
				e02->vertex->active = 2;

				e = e6->spin->spin;
				slots = 0;
				while (e->slots == -1)
				{
					slots++;
					e = e->spin;
				}
				cout << "# of slots is " << slots << " of " << e->slots;
				e->slots = slots;
				e02->slots = slots;

				e = e02->spin->spin;
				slots = 0;
				while (e->slots == -1)
				{
					slots++;
					e = e->spin;
				}
				cout << "and " << slots << " of " << e->slots << endl;
				e->slots = slots;
				e6->slots = slots;
			}

// specify whether face f4 is a boundary face

			if (e6->is_on_boundary)
			{
				if (e6->slots)
				{
					if (e6->inv->group == -1)
					{
						codec->writeIsBoundary(1);
					}
					else
					{
						codec->writeIsBoundary(0);
					}
				}
				else
				{
					// it's the same as what is across e6 
				}
				if (e6->inv->group == -1)
				{
					e7->is_on_boundary = 2;
					e31->is_on_boundary = 2;

					e4->is_on_boundary = 2;
					e20->is_on_boundary = 2;

					e5->is_on_boundary = 2;
					e13->is_on_boundary = 2;
				}
			}

// specify edge5

			if (!edge5adjacency)
			{
				codec->writeAlreadyVisited(0);
				if (e5->is_on_boundary)
				{
					if (e5->is_on_boundary != 2)
					{
						codec->writeIsBoundary(1); // i did not know that
					}
					codec->writeBoundaryEdgeDegree(e5->degree);
				}
				else
				{
					codec->writeIsBoundary(0);
					codec->writeInteriorEdgeDegree(e5->degree);
				}
				e5->slots = e13->slots = e5->degree - 2;
			}
			else
			{
				codec->writeAlreadyVisited(1);
				cout << "case0: edge5adjacency" << endl;
				e5->vertex->active = 2;
				e13->vertex->active = 2;

				e = e5->spin->spin;
				slots = 0;
				while (e->slots == -1)
				{
					slots++;
					e = e->spin;
				}
				cout << "# of slots is " << slots << " of " << e->slots;
				e->slots = slots;
				e13->slots = slots;

				e = e13->spin->spin;
				slots = 0;
				while (e->slots == -1)
				{
					slots++;
					e = e->spin;
				}
				cout << "and " << slots << " of " << e->slots << endl;
				e->slots = slots;
				e5->slots = slots;
			}

// specify edge4

			if (!edge4adjacency)
			{
				codec->writeAlreadyVisited(0);
				if (e4->is_on_boundary)
				{
					if (e4->is_on_boundary != 2)
					{
						codec->writeIsBoundary(1); // i did not know that
					}
					codec->writeBoundaryEdgeDegree(e4->degree);
				}
				else
				{
					codec->writeIsBoundary(0);
					codec->writeInteriorEdgeDegree(e4->degree);
				}
				e4->slots = e20->slots = e4->degree - 2;
			}
			else
			{
				codec->writeAlreadyVisited(1);
				cout << "case0: edge4adjacency" << endl;
				e4->vertex->active = 2;
				e20->vertex->active = 2;

				e = e4->spin->spin;
				slots = 0;
				while (e->slots == -1)
				{
					slots++;
					e = e->spin;
				}
				cout << "# of slots is " << slots << " of " << e->slots;
				e->slots = slots;
				e20->slots = slots;

				e = e20->spin->spin;
				slots = 0;
				while (e->slots == -1)
				{
					slots++;
					e = e->spin;
				}
				cout << "and " << slots << " of " << e->slots << endl;
				e->slots = slots;
				e4->slots = slots;
			}

// specify edge7

			if (!edge7adjacency)
			{
				codec->writeAlreadyVisited(0);
				if (e7->is_on_boundary)
				{
					if (e7->is_on_boundary != 2)
					{
						codec->writeIsBoundary(1); // i did not know that
					}
					codec->writeBoundaryEdgeDegree(e7->degree);
				}
				else
				{
					codec->writeIsBoundary(0);
					codec->writeInteriorEdgeDegree(e7->degree);
				}
				e7->slots = e31->slots = e7->degree - 2;
			}
			else
			{
				codec->writeAlreadyVisited(1);
				cout << "case0: edge7adjacency" << endl;
				e7->vertex->active = 2;
				e31->vertex->active = 2;

				e = e7->spin->spin;
				slots = 0;
				while (e->slots == -1)
				{
					slots++;
					e = e->spin;
				}
				cout << "# of slots is " << slots << " of " << e->slots;
				e->slots = slots;
				e31->slots = slots;

				e = e31->spin->spin;
				slots = 0;
				while (e->slots == -1)
				{
					slots++;
					e = e->spin;
				}
				cout << "and " << slots << " of " << e->slots << endl;
				e->slots = slots;
				e7->slots = slots;
			}
		}
		else
		{
			codec->writeAlreadyVisited(1); // it was case M
			cout << "case0: face4adjacency" << endl;

			int candidates = getSlot1EdgeHEX(e4);
			codec->writeRange((verticesHEXcount-verticesHEXdone+1),e4->inv->vertex->index-verticesHEXdone);
			codec->writeRange(candidates,0);

// advance around the old edges setting an decreased slot count for the new boundary spinedges

			e20->slots = e4->inv->slots - 1;
			e13->slots = e5->inv->slots - 1;
			e02->slots = e6->inv->slots - 1;
			e31->slots = e7->inv->slots - 1;

		// update the old boundary spinedges on the other end with the decreased slot count

			e = e4->spin;
			while (e->slots == -1)
			{
				e = e->spin;
			}
			if (e->slots != e4->inv->slots)
			{
				cout << "other side of e4.inv wrong slot count" << endl;
			}
			else
			{
				e->slots--;
				updateFaceHEX(e);
			}

			e = e5->spin;
			while (e->slots == -1)
			{
				e = e->spin;
			}
			if (e->slots != e5->inv->slots)
			{
				cout << "other side of e5.inv wrong slot count" << endl;
			}
			else
			{
				e->slots--;
				updateFaceHEX(e);
			}

			e = e6->spin;
			while (e->slots == -1)
			{
				e = e->spin;
			}
			if (e->slots != e6->inv->slots)
			{
				cout << "other side of e6.inv wrong slot count" << endl;
			}
			else
			{
				e->slots--;
				updateFaceHEX(e);
			}

			e = e7->spin;
			while (e->slots == -1)
			{
				e = e->spin;
			}
			if (e->slots != e7->inv->slots)
			{
				cout << "other side of e7.inv wrong slot count" << endl;
			}
			else
			{
				e->slots--;
				updateFaceHEX(e);
			}

// delete the (now) interior spinedges from further consideration

			e4->inv->slots = -1;
			e5->inv->slots = -1;
			e6->inv->slots = -1;
			e7->inv->slots = -1;

			e4->vertex->active = 2;
			e5->vertex->active = 2;
			e6->vertex->active = 2;
			e7->vertex->active = 2;
		}

// process (and renumber) the vertices

		if (e7->vertex->active)
		{
			if (e7->vertex->active == 2)
			{
				e7->vertex->active = 1;
			}
			else
			{
				cout << "case 0: non-manifold vertex 7" << endl;
			}
		}
		else
		{
			e7->vertex->active = 1;
			if (e0->inv->group != -1)
			{
				encodePosition7(codec,pq,e7->vertex,e0->vertex,e0->inv->spin->next->next->vertex);
			}
			else
			{
				encodePosition7(codec,pq,e7->vertex,e0->vertex);
			}
			e7->vertex->index = verticesHEXcount++;
		}

		if (e6->vertex->active)
		{
			if (e6->vertex->active == 2)
			{
				e6->vertex->active = 1;
			}
			else
			{
				cout << "case 0: non-manifold vertex 6" << endl;
			}
		}
		else
		{
			e6->vertex->active = 1;
			encodePosition6(codec,pq,e6->vertex,e1->vertex,e0->vertex,e7->vertex);
			e6->vertex->index = verticesHEXcount++;
		}

		if (e5->vertex->active)
		{
			if (e5->vertex->active == 2)
			{
				e5->vertex->active = 1;
			}
			else
			{
				cout << "case 0: non-manifold vertex 5" << endl;
			}
		}
		else
		{
			e5->vertex->active = 1;
			encodePosition5(codec,pq,e5->vertex,e0->vertex,e1->vertex,e2->vertex,e3->vertex,e6->vertex,e7->vertex);
			e5->vertex->index = verticesHEXcount++;
		}

		if (e4->vertex->active)
		{
			if (e4->vertex->active == 2)
			{
				e4->vertex->active = 1;
			}
			else
			{
				cout << "case 0: non-manifold vertex 4" << endl;
			}
		}
		else
		{
			e4->vertex->active = 1;
			encodePosition4(codec,pq,e4->vertex,e0->vertex,e2->vertex,e3->vertex,e5->vertex,e6->vertex,e7->vertex,e1->vertex);
			e4->vertex->index = verticesHEXcount++;
		}
	
// add the edges to their vertices

		addEdgeVertexHEX(e00);
		addEdgeVertexHEX(e01);
		addEdgeVertexHEX(e02);
		addEdgeVertexHEX(e03);

		addEdgeVertexHEX(e10);
		addEdgeVertexHEX(e11);
		addEdgeVertexHEX(e12);
		addEdgeVertexHEX(e13);

		addEdgeVertexHEX(e20);
		addEdgeVertexHEX(e21);
		addEdgeVertexHEX(e22);
		addEdgeVertexHEX(e23);

		addEdgeVertexHEX(e30);
		addEdgeVertexHEX(e31);
		addEdgeVertexHEX(e32);
		addEdgeVertexHEX(e33);

		if (!face4adjacency)
		{
			addEdgeVertexHEX(e4);
			addEdgeVertexHEX(e5);
			addEdgeVertexHEX(e6);
			addEdgeVertexHEX(e7);
		}

// add (inverse) faces to the priority list

		addFaceHEX(e00);
		addFaceHEX(e11);
		addFaceHEX(e22);
		addFaceHEX(e33);

		if (!face4adjacency)
		{
			addFaceHEX(e4);
		}
	}
	else if (face0adjacency && !face1adjacency && !face2adjacency && !face3adjacency && !face4adjacency && !edge4adjacency) // case 1 (not complete)
	{
		case1++;
		// cout << "case 1" << endl;

// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e11->slots = e1->inv->slots - 1;
		e22->slots = e2->inv->slots - 1;
		e33->slots = e3->inv->slots - 1;
		e32->slots = e01->inv->slots - 1;
		e6->slots = e02->inv->slots - 1;
		e12->slots = e03->inv->slots - 1;

// update the old boundary spinedges on the other end with the decreased slot count

		e = e1->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e1->inv->slots)
		{
			cout << "other side of e1 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e2->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e2->inv->slots)
		{
			cout << "other side of e2 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e3->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e3->inv->slots)
		{
			cout << "other side of e3 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e01->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e01->inv->slots)
		{
			cout << "other side of e01.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e02->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e02->inv->slots)
		{
			cout << "other side of e02.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e03->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e03->inv->slots)
		{
			cout << "other side of e03.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;
		e3->inv->slots = -1;

		e00->inv->slots = -1;
		e01->inv->slots = -1;
		e02->inv->slots = -1;
		e03->inv->slots = -1;

// propagate all boundary information (there is some avoidable redundancy with the following per face boundary specification here)

		if (e11->inv->group == -1 && (!e11->slots || !e12->slots))
		{
			e13->is_on_boundary = 2;
			e5->is_on_boundary = 2;

			e10->is_on_boundary = 2;
			e23->is_on_boundary = 2;
		}

		if (e22->inv->group == -1 && (!e22->slots))
		{
			e23->is_on_boundary = 2;
			e10->is_on_boundary = 2;

			e20->is_on_boundary = 2;
			e4->is_on_boundary = 2;

			e21->is_on_boundary = 2;
			e30->is_on_boundary = 2;
		}

		if (e33->inv->group == -1 && (!e33->slots || !e32->slots))
		{
			e30->is_on_boundary = 2;
			e21->is_on_boundary = 2;

			e31->is_on_boundary = 2;
			e7->is_on_boundary = 2;
		}

		if (e6->inv->group == -1 && (!e6->slots))
		{
			e7->is_on_boundary = 2;
			e31->is_on_boundary = 2;

			e4->is_on_boundary = 2;
			e20->is_on_boundary = 2;

			e5->is_on_boundary = 2;
			e13->is_on_boundary = 2;
		}

// process free faces & edges (first those that are connected) in a smart order

// specify edge2

		if (!edge2adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e23->is_on_boundary)
			{
				if (e23->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
					// mesh->active = e23;
				}
				codec->writeBoundaryEdgeDegree(e23->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e23->degree);
			}
			e23->slots = e10->slots = e23->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case1: edge2adjacency" << endl;
			e10->vertex->active = 2;

			e = e23->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e10->slots = slots;

			e = e10->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e23->slots = slots;
		}

// specify edge3

		if (!edge3adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e30->is_on_boundary)
			{
				if (e30->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e30->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e30->degree);
			}
			e30->slots = e21->slots = e30->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case1: edge3adjacency" << endl;
			e21->vertex->active = 2;

			e = e30->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e21->slots = slots;

			e = e21->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e30->slots = slots;
		}

// specify whether face f1 is a boundary face

		if (e11->is_on_boundary && e12->is_on_boundary  && e10->is_on_boundary)
		{
			if (e11->slots && e12->slots && e10->slots)
			{
				if (e11->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e11, e12, or e10
			}
			if (e11->inv->group == -1)
			{
				e13->is_on_boundary = 2;
				e5->is_on_boundary = 2;
			}
		}

// specify whether face f2 is a boundary face

		if (e22->is_on_boundary && e23->is_on_boundary && e21->is_on_boundary)
		{
			if (e22->slots && e23->slots && e21->slots)
			{
				if (e22->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e22, e23, or e21
			}
			if (e22->inv->group == -1)
			{
				e20->is_on_boundary = 2;
				e4->is_on_boundary = 2;
			}
		}

// specify whether face f3 is a boundary face

		if (e33->is_on_boundary && e32->is_on_boundary && e30->is_on_boundary)
		{
			if (e33->slots && e32->slots && e30->slots)
			{
				if (e33->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e33, e32, or e30 
			}
			if (e33->inv->group == -1)
			{
				e31->is_on_boundary = 2;
				e7->is_on_boundary = 2;
			}
		}

// specify edge7

		if (!edge7adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e7->is_on_boundary)
			{
				if (e7->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
					// mesh->active = e7;
				}
				codec->writeBoundaryEdgeDegree(e7->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e7->degree);
			}
			e7->slots = e31->slots = e7->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case1: edge7adjacency" << endl;
			e31->vertex->active = 2;

			e = e7->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e31->slots = slots;

			e = e31->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e7->slots = slots;
		}

// specify edge5

		if (!edge5adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e5->is_on_boundary)
			{
				if (e5->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
					// mesh->active = e5;
				}
				codec->writeBoundaryEdgeDegree(e5->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e5->degree);
			}
			e5->slots = e13->slots = e5->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case1: edge5adjacency" << endl;
			e5->vertex->active = 2;

			e = e5->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e13->slots = slots;

			e = e13->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e5->slots = slots;
		}

// specify whether face f4 is a boundary face

		if (e6->is_on_boundary && e5->is_on_boundary && e7->is_on_boundary)
		{
			if (e6->slots && e5->slots && e7->slots)
			{
				if (e6->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e6, e5, or e7
			}
			if (e6->inv->group == -1)
			{
				e4->is_on_boundary = 2;
				e20->is_on_boundary = 2;
			}
		}

// process free edges (now those that are not connected)

// specify edge4

		if (!edge4adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e4->is_on_boundary)
			{
				if (e4->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
					// mesh->active = e4;
				}
				codec->writeBoundaryEdgeDegree(e4->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e4->degree);
			}
			e4->slots = e20->slots = e4->degree - 2;
		}
		else
		{
			cout << "case1: edge4adjacency >>>not handled<<<" << endl;
		}

// process (and renumber) the vertices

		if (e5->vertex->active)
		{
			if (e5->vertex->active == 2)
			{
				e5->vertex->active = 1;
			}
			else
			{
				cout << "case1: non-manifold vertex 5" << endl;
			}
		}
		else
		{
			e5->vertex->active = 1;
			encodePosition5(codec,pq,e5->vertex,e0->vertex,e1->vertex,e2->vertex,e3->vertex,e6->vertex,e7->vertex);
			e5->vertex->index = verticesHEXcount++;
		}

		if (e4->vertex->active)
		{
			if (e4->vertex->active == 2)
			{
				e4->vertex->active = 1;
			}
			else
			{
				cout << "case1: non-manifold vertex 4" << endl;
			}
		}
		else
		{
			e4->vertex->active = 1;
			encodePosition4(codec,pq,e4->vertex,e0->vertex,e2->vertex,e3->vertex,e5->vertex,e6->vertex,e7->vertex,e1->vertex);
			e4->vertex->index = verticesHEXcount++;
		}

// add the edges to their vertices

		addEdgeVertexHEX(e10);
		addEdgeVertexHEX(e11);
		addEdgeVertexHEX(e12);
		addEdgeVertexHEX(e13);

		addEdgeVertexHEX(e20);
		addEdgeVertexHEX(e21);
		addEdgeVertexHEX(e22);
		addEdgeVertexHEX(e23);

		addEdgeVertexHEX(e30);
		addEdgeVertexHEX(e31);
		addEdgeVertexHEX(e32);
		addEdgeVertexHEX(e33);

		addEdgeVertexHEX(e4);
		addEdgeVertexHEX(e5);
		addEdgeVertexHEX(e6);
		addEdgeVertexHEX(e7);

// add (inverse) faces to the priority list

		addFaceHEX(e11);
		addFaceHEX(e22);
		addFaceHEX(e33);
		addFaceHEX(e4);
	}
	else if (face0adjacency && face1adjacency && !face2adjacency && !face3adjacency && !face4adjacency) // case 2a
	{
		case2a++;
		// cout << "case 2a" << endl;

// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e22->slots = e2->inv->slots - 1;
		e33->slots = e3->inv->slots - 1;
		e32->slots = e01->inv->slots - 1;
		e6->slots = e02->inv->slots - 1;
		e5->slots = e13->inv->slots - 1;
		e23->slots = e10->inv->slots - 1;

// update the old boundary spinedges on the other end with the decreased slot count

		e = e2->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e2->inv->slots)
		{
			cout << "other side of e2 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e3->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e3->inv->slots)
		{
			cout << "other side of e3 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e01->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e01->inv->slots)
		{
			cout << "other side of e01.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e02->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e02->inv->slots)
		{
			cout << "other side of e02.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e13->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e13->inv->slots)
		{
			cout << "other side of e13.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e10->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e10->inv->slots)
		{
			cout << "other side of e10.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;
		e3->inv->slots = -1;

		e00->inv->slots = -1;
		e01->inv->slots = -1;
		e02->inv->slots = -1;
		e03->inv->slots = -1;

		e10->inv->slots = -1;
		e11->inv->slots = -1;
		e12->inv->slots = -1;
		e13->inv->slots = -1;

// process free faces & edges (all are connected) in a smart order

// specify whether face f2 is a boundary face

		if (e22->is_on_boundary && e23->is_on_boundary)
		{
			if (e22->slots && e23->slots)
			{
				if (e22->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across 22 or 23
			}
			if (e22->inv->group == -1)
			{
				e20->is_on_boundary = 2;
				e4->is_on_boundary = 2;

				e21->is_on_boundary = 2;
				e30->is_on_boundary = 2;
			}
		}

// specify whether face f3 is a boundary face

		if (e33->is_on_boundary && e32->is_on_boundary)
		{
			if (e33->slots && e32->slots)
			{
				if (e33->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e33 or e32 
			}
			if (e33->inv->group == -1)
			{
				e30->is_on_boundary = 2;
				e21->is_on_boundary = 2;

				e31->is_on_boundary = 2;
				e7->is_on_boundary = 2;
			}
		}

// specify whether face f4 is a boundary face

		if (e6->is_on_boundary && e5->is_on_boundary)
		{
			if (e6->slots && e5->slots)
			{
				if (e6->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e6 or e5
			}
			if (e6->inv->group == -1)
			{
				e7->is_on_boundary = 2;
				e31->is_on_boundary = 2;

				e4->is_on_boundary = 2;
				e20->is_on_boundary = 2;
			}
		}

// specify edge3

		if (!edge3adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e30->is_on_boundary)
			{
				if (e30->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
					// mesh->active = e30;
				}
				codec->writeBoundaryEdgeDegree(e30->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e30->degree);
			}
			e30->slots = e21->slots = e30->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case2a: edge3adjacency" << endl;
			e21->vertex->active = 2;

			e = e30->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e21->slots = slots;

			e = e21->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e30->slots = slots;
		}

// specify edge7

		if (!edge7adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e7->is_on_boundary)
			{
				if (e7->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
					// mesh->active = e7;
				}
				codec->writeBoundaryEdgeDegree(e7->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e7->degree);
			}
			e7->slots = e31->slots = e7->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case2a: edge7adjacency" << endl;
			e31->vertex->active = 2;

			e = e7->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e31->slots = slots;

			e = e31->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e7->slots = slots;
		}

// specify edge4

		if (!edge4adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e4->is_on_boundary)
			{
				if (e4->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
					// mesh->active = e4;
				}
				codec->writeBoundaryEdgeDegree(e4->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e4->degree);
			}
			e4->slots = e20->slots = e4->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case2a: edge4adjacency" << endl;
			e4->vertex->active = 2;

			e = e4->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e20->slots = slots;

			e = e20->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e4->slots = slots;
		}

// process (and renumber) the vertices

		if (e4->vertex->active)
		{
			if (e4->vertex->active == 2)
			{
				e4->vertex->active = 1;
			}
			else
			{
				cout << "case 2a: non-manifold vertex 4" << endl;
			}
		}
		else
		{
			e4->vertex->active = 1;
			encodePosition4(codec,pq,e4->vertex,e0->vertex,e2->vertex,e3->vertex,e5->vertex,e6->vertex,e7->vertex,e1->vertex);
			e4->vertex->index = verticesHEXcount++;
		}

// add the edges to their vertices

		addEdgeVertexHEX(e20);
		addEdgeVertexHEX(e21);
		addEdgeVertexHEX(e22);
		addEdgeVertexHEX(e23);

		addEdgeVertexHEX(e30);
		addEdgeVertexHEX(e31);
		addEdgeVertexHEX(e32);
		addEdgeVertexHEX(e33);

		addEdgeVertexHEX(e4);
		addEdgeVertexHEX(e5);
		addEdgeVertexHEX(e6);
		addEdgeVertexHEX(e7);

// add (inverse) faces to the priority list

		addFaceHEX(e22);
		addFaceHEX(e33);
		addFaceHEX(e4);
	}
	else if (face0adjacency && !face1adjacency && face2adjacency && !face3adjacency && !face4adjacency) // case 2b
	{
		case2b++;
		// cout << "case 2b" << endl;

// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e11->slots = e1->inv->slots - 1;
		e10->slots = e23->inv->slots - 1;
		e4->slots = e20->inv->slots - 1;
		e30->slots = e21->inv->slots - 1;
		e33->slots = e3->inv->slots - 1;
		e32->slots = e01->inv->slots - 1;
		e6->slots = e02->inv->slots - 1;
		e12->slots = e03->inv->slots - 1;

// update the old boundary spinedges on the other end with the decreased slot count

		e = e1->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e1->inv->slots)
		{
			cout << "other side of e1 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e23->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e23->inv->slots)
		{
			cout << "other side of e23.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e20->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e20->inv->slots)
		{
			cout << "other side of e20.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e21->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e21->inv->slots)
		{
			cout << "other side of e21.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e3->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e3->inv->slots)
		{
			cout << "other side of e3 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e01->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e01->inv->slots)
		{
			cout << "other side of e01.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e02->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e02->inv->slots)
		{
			cout << "other side of e02.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e03->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e03->inv->slots)
		{
			cout << "other side of e03.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;
		e3->inv->slots = -1;

		e00->inv->slots = -1;
		e01->inv->slots = -1;
		e02->inv->slots = -1;
		e03->inv->slots = -1;

		e20->inv->slots = -1;
		e21->inv->slots = -1;
		e22->inv->slots = -1;
		e23->inv->slots = -1;

// process free faces & edges (all are connected) in a smart order

// specify whether face f1 is a boundary face

		if (e11->is_on_boundary && e12->is_on_boundary && e10->is_on_boundary)
		{
			if (e11->slots && e12->slots && e10->slots)
			{
				if (e11->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e11, e12, or e10
			}
			if (e11->inv->group == -1)
			{
				e13->is_on_boundary = 2;
				e5->is_on_boundary = 2;
			}
		}

// specify whether face f3 is a boundary face

		if (e33->is_on_boundary && e30->is_on_boundary && e32->is_on_boundary)
		{
			if (e33->slots && e30->slots && e32->slots)
			{
				if (e33->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e33, e30, or e32 
			}
			if (e33->inv->group == -1)
			{
				e31->is_on_boundary = 2;
				e7->is_on_boundary = 2;
			}
		}

// specify whether face f4 is a boundary face

		if (e6->is_on_boundary && e4->is_on_boundary)
		{
			if (e6->slots && e4->slots)
			{
				if (e6->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e6 or e4
			}
			if (e6->inv->group == -1)
			{
				e7->is_on_boundary = 2;
				e31->is_on_boundary = 2;

				e5->is_on_boundary = 2;
				e13->is_on_boundary = 2;
			}
		}

// specify edge5

		if (!edge5adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e5->is_on_boundary)
			{
				if (e5->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e5->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e5->degree);
			}
			e5->slots = e13->slots = e5->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case2b: edge5adjacency" << endl;

			e = e5->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e13->slots = slots;

			e = e13->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e5->slots = slots;
		}

// specify edge7

		if (!edge7adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e7->is_on_boundary)
			{
				if (e7->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e7->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e7->degree);
			}
			e7->slots = e31->slots = e7->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case2b: edge7adjacency" << endl;
			e31->vertex->active = 2;

			e = e7->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e31->slots = slots;

			e = e31->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e7->slots = slots;
		}

// add the edges to their vertices

		addEdgeVertexHEX(e10);
		addEdgeVertexHEX(e11);
		addEdgeVertexHEX(e12);
		addEdgeVertexHEX(e13);

		addEdgeVertexHEX(e30);
		addEdgeVertexHEX(e31);
		addEdgeVertexHEX(e32);
		addEdgeVertexHEX(e33);

		addEdgeVertexHEX(e4);
		addEdgeVertexHEX(e5);
		addEdgeVertexHEX(e6);
		addEdgeVertexHEX(e7);

// add (inverse) faces to the priority list

		addFaceHEX(e11);
		addFaceHEX(e33);
		addFaceHEX(e4);
	}
	else if (face0adjacency && face1adjacency && face2adjacency && !face3adjacency && !face4adjacency) // case 3a
	{
		case3a++;
		// cout << "case 3a" << endl;

// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e33->slots = e3->inv->slots - 1;
		e32->slots = e01->inv->slots - 1;
		e6->slots = e02->inv->slots - 1;
		e5->slots = e13->inv->slots - 1;
		e4->slots = e20->inv->slots - 1;
		e30->slots = e21->inv->slots - 1;

// update the old boundary spinedges on the other end with the decreased slot count

		e = e3->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e3->inv->slots)
		{
			cout << "other side of e3 wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e01->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e01->inv->slots)
		{
			cout << "other side of e01.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e02->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e02->inv->slots)
		{
			cout << "other side of e02.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e13->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e13->inv->slots)
		{
			cout << "other side of e13.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e20->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e20->inv->slots)
		{
			cout << "other side of e20.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e21->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e21->inv->slots)
		{
			cout << "other side of e21.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;
		e3->inv->slots = -1;

		e00->inv->slots = -1;
		e01->inv->slots = -1;
		e02->inv->slots = -1;
		e03->inv->slots = -1;

		e10->inv->slots = -1;
		e11->inv->slots = -1;
		e12->inv->slots = -1;
		e13->inv->slots = -1;

		e20->inv->slots = -1;
		e21->inv->slots = -1;
		e22->inv->slots = -1;
		e23->inv->slots = -1;

// process free faces & edges (all are connected) in a smart order

// specify whether face f3 is a boundary face

		if (e33->is_on_boundary && e32->is_on_boundary && e30->is_on_boundary)
		{
			if (e33->slots && e32->slots && e30->slots)
			{
				if (e33->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e33, e32, or e30 
			}
			if (e33->inv->group == -1)
			{
				e31->is_on_boundary = 2;
				e7->is_on_boundary = 2;
			}
		}

// specify whether face f4 is a boundary face

		if (e6->is_on_boundary && e5->is_on_boundary && e4->is_on_boundary)
		{
			if (e6->slots && e5->slots && e4->slots)
			{
				if (e6->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e6, e5, or e4
			}
			if (e6->inv->group == -1)
			{
				e7->is_on_boundary = 2;
				e31->is_on_boundary = 2;
			}
		}

// specify edge7

		if (!edge7adjacency)
		{
			codec->writeAlreadyVisited(0);
			if (e7->is_on_boundary)
			{
				if (e7->is_on_boundary != 2)
				{
					codec->writeIsBoundary(1); // i did not know that
				}
				codec->writeBoundaryEdgeDegree(e7->degree);
			}
			else
			{
				codec->writeIsBoundary(0);
				codec->writeInteriorEdgeDegree(e7->degree);
			}
			e7->slots = e31->slots = e7->degree - 2;
		}
		else
		{
			codec->writeAlreadyVisited(1);
			cout << "case3a: edge7adjacency" << endl;
			e31->vertex->active = 2;

			e = e7->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "# of slots is " << slots << " of " << e->slots;
			e->slots = slots;
			e31->slots = slots;

			e = e31->spin->spin;
			slots = 0;
			while (e->slots == -1)
			{
				slots++;
				e = e->spin;
			}
			cout << "and " << slots << " of " << e->slots << endl;
			e->slots = slots;
			e7->slots = slots;
		}

// add the edges to their vertices

		addEdgeVertexHEX(e30);
		addEdgeVertexHEX(e31);
		addEdgeVertexHEX(e32);
		addEdgeVertexHEX(e33);

		addEdgeVertexHEX(e4);
		addEdgeVertexHEX(e5);
		addEdgeVertexHEX(e6);
		addEdgeVertexHEX(e7);

// add (inverse) faces to the priority list

		addFaceHEX(e33);
		addFaceHEX(e4);
	}
	else if (face0adjacency && !face1adjacency && face2adjacency && !face3adjacency && face4adjacency) // case 3b
	{
		case3b++;
		// cout << "case 3b" << endl;

// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e11->slots = e1->inv->slots - 1;
		e12->slots = e03->inv->slots - 1;
		e13->slots = e5->inv->slots - 1;
		e10->slots = e23->inv->slots - 1;

		e33->slots = e3->inv->slots - 1;
		e30->slots = e21->inv->slots - 1;
		e31->slots = e7->inv->slots - 1;
		e32->slots = e01->inv->slots - 1;

// update the old boundary spinedges on the other end with the decreased slot count

		e = e1->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e1->inv->slots)
		{
			cout << "other side of e1.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e03->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e03->inv->slots)
		{
			cout << "other side of e03.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e5->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e5->inv->slots)
		{
			cout << "other side of e02.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e23->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e23->inv->slots)
		{
			cout << "other side of e23.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e3->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e3->inv->slots)
		{
			cout << "other side of e3.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e21->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e21->inv->slots)
		{
			cout << "other side of e21.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e7->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e7->inv->slots)
		{
			cout << "other side of e7.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e01->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e01->inv->slots)
		{
			cout << "other side of e01.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;
		e3->inv->slots = -1;

		e00->inv->slots = -1;
		e01->inv->slots = -1;
		e02->inv->slots = -1;
		e03->inv->slots = -1;

		e20->inv->slots = -1;
		e21->inv->slots = -1;
		e22->inv->slots = -1;
		e23->inv->slots = -1;

		e4->inv->slots = -1;
		e5->inv->slots = -1;
		e6->inv->slots = -1;
		e7->inv->slots = -1;

// process free faces & edges (all are connected) in a smart order

// specify whether face f1 is a boundary face

		if (e11->is_on_boundary && e12->is_on_boundary && e10->is_on_boundary && e13->is_on_boundary)
		{
			if (e11->slots && e12->slots && e10->slots && e13->slots)
			{
				if (e11->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across
			}
		}

// specify whether face f3 is a boundary face

		if (e33->is_on_boundary && e32->is_on_boundary && e30->is_on_boundary && e31->is_on_boundary)
		{
			if (e33->slots && e32->slots && e30->slots && e31->slots)
			{
				if (e33->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across
			}
		}

// add the edges to their vertices

		addEdgeVertexHEX(e10);
		addEdgeVertexHEX(e11);
		addEdgeVertexHEX(e12);
		addEdgeVertexHEX(e13);

		addEdgeVertexHEX(e30);
		addEdgeVertexHEX(e31);
		addEdgeVertexHEX(e32);
		addEdgeVertexHEX(e33);

// add (inverse) faces to the priority list

		addFaceHEX(e11);
		addFaceHEX(e33);
	}
	else if (face0adjacency && face1adjacency && face2adjacency && face3adjacency && !face4adjacency) // case 4
	{
		case4++;
		// cout << "case 4" << endl;

// advance around the old edges setting an decreased slot count for the new boundary spinedges

		e7->slots = e31->inv->slots - 1;
		e6->slots = e02->inv->slots - 1;
		e5->slots = e13->inv->slots - 1;
		e4->slots = e20->inv->slots - 1;

	// update the old boundary spinedges on the other end with the decreased slot count

		e = e31->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e31->inv->slots)
		{
			cout << "other side of e31.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e02->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e02->inv->slots)
		{
			cout << "other side of e02.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e13->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e13->inv->slots)
		{
			cout << "other side of e13.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

		e = e20->spin;
		while (e->slots == -1)
		{
			e = e->spin;
		}
		if (e->slots != e20->inv->slots)
		{
			cout << "other side of e20.inv wrong slot count" << endl;
		}
		else
		{
			e->slots--;
			updateFaceHEX(e);
		}

// delete the (now) interior spinedges from further consideration

		e0->inv->slots = -1;
		e1->inv->slots = -1;
		e2->inv->slots = -1;
		e3->inv->slots = -1;

		e00->inv->slots = -1;
		e01->inv->slots = -1;
		e02->inv->slots = -1;
		e03->inv->slots = -1;

		e10->inv->slots = -1;
		e11->inv->slots = -1;
		e12->inv->slots = -1;
		e13->inv->slots = -1;

		e20->inv->slots = -1;
		e21->inv->slots = -1;
		e22->inv->slots = -1;
		e23->inv->slots = -1;

		e30->inv->slots = -1;
		e31->inv->slots = -1;
		e32->inv->slots = -1;
		e33->inv->slots = -1;

// process free faces & edges (all are connected) in a smart order

// specify whether face f4 is a boundary face

		if (e6->is_on_boundary && e5->is_on_boundary && e4->is_on_boundary && e7->is_on_boundary)
		{
			if (e6->slots && e5->slots && e4->slots && e7->slots)
			{
				if (e6->inv->group == -1)
				{
					codec->writeIsBoundary(1);
				}
				else
				{
					codec->writeIsBoundary(0);
				}
			}
			else
			{
				// it's the same as what is across e6
			}
		}

// add the edges to their vertices

		addEdgeVertexHEX(e4);
		addEdgeVertexHEX(e5);
		addEdgeVertexHEX(e6);
		addEdgeVertexHEX(e7);

// add (inverse) faces to the priority list

		addFaceHEX(e4);
	}
	else
	{
		if (face0adjacency)
		{
			cout << "face0adjacency" << endl;
		}
		if (face1adjacency)
		{
			cout << "face1adjacency" << endl;
		}
		if (face2adjacency)
		{
			cout << "face2adjacency" << endl;
		}
		if (face3adjacency)
		{
			cout << "face3adjacency" << endl;
		}
		if (face4adjacency)
		{
			cout << "face4adjacency" << endl;
		}

		if (vertex4adjacency)
		{
			cout << "vertex 4 is active" << endl;
		}
		if (vertex5adjacency)
		{
			cout << "vertex 5 is active" << endl;
		}
		if (vertex6adjacency)
		{
			cout << "vertex 6 is active" << endl;
		}
		if (vertex7adjacency)
		{
			cout << "vertex 7 is active" << endl;
		}

		if (edge0adjacency)
		{
			cout << "edge0adjacency" << endl;
		}
		if (edge1adjacency)
		{
			cout << "edge1adjacency" << endl;
		}
		if (edge2adjacency)
		{
			cout << "edge2adjacency" << endl;
		}
		if (edge3adjacency)
		{
			cout << "edge3adjacency" << endl;
		}
		if (edge4adjacency)
		{
			cout << "edge4adjacency" << endl;
		}
		if (edge5adjacency)
		{
			cout << "edge5adjacency" << endl;
		}
		if (edge6adjacency)
		{
			cout << "edge6adjacency" << endl;
		}
		if (edge7adjacency)
		{
			cout << "edge7adjacency" << endl;
		}
	}

	mesh->gate = getFaceHEX();
	if (mesh->gate == NULL) mesh->gate = getEdgeVertexHEX();

//	if (mesh->active) return 0;

	return (mesh->gate != NULL);
}

void CFencoder::addEdgeVertexHEX(CFspinEdge* edge)
{
	int index = edge->vertex->index;
	edge->Lnext = verticesHEX[index];
	verticesHEX[index] = edge;
}

CFspinEdge* CFencoder::getEdgeVertexHEX()
{
	while (verticesHEXdone < verticesHEXnumber)
	{
		while (verticesHEX[verticesHEXdone])
		{
			if ((verticesHEX[verticesHEXdone]->slots != -1) && (verticesHEX[verticesHEXdone]->inv->group != -1))
			{
				return verticesHEX[verticesHEXdone]->inv;
			}
			else
			{
				verticesHEX[verticesHEXdone] = verticesHEX[verticesHEXdone]->Lnext;
			}
		}
		verticesHEXdone++;
	}
	return 0;
}

int CFencoder::getSlot1EdgeHEX(CFspinEdge* edge)
{
	getEdgeVertexHEX(); // discard completed vertices

	CFspinEdge* e1 = verticesHEX[edge->inv->vertex->index];
	CFspinEdge* e2 = 0;

	int count = 0;

	while (e1)
	{
		if (e1->slots >= 1)
		{
			count++;
			if (edge->inv == e1)
			{
				e2 = e1;
			}
		}
		e1 = e1->Lnext;
	}
	if (e2 == 0)
	{
		cout << "PROBLEM here" << endl;
	}
	else
	{
		cout << "candidate count: " << count << endl;
		cout << "verticesHEXcount-verticesHEXdone+1: " << verticesHEXcount-verticesHEXdone+1 << endl;
	}
	return count;
}

void CFencoder::addFaceHEX(CFspinEdge* face)
{
	if (face->inv->group != -1)
	{
		int count = (face->slots == 0)+(face->next->slots == 0)+(face->next->next->slots == 0)+(face->next->next->next->slots == 0);
		if (count)
		{
			face->inv->Lnext = facesHEX[count];
			facesHEX[count] = face->inv;
		}
		else
		{
			count = 4;
			while (count--)
			{
				if (face->is_on_boundary && (face->slots == 1))
				{
					if (facesHEX[0] == 0)
					{
						facesHEX[0] = face->inv;
						facesHEX[5] = 0;
					}
					if (facesHEX[5] != 0)
					{
						facesHEX[5]->Lnext = face->inv;
					}
					facesHEX[5] = face->inv;
					return;
				}
				face = face->next;
			}
		}
	}
	checkFaceHEX();
}

void CFencoder::updateFaceHEX(CFspinEdge* face)
{
	if (face->inv->group != -1)
	{
		if (face->slots == 0)
		{
			int count = 1+(face->next->slots == 0)+(face->next->next->slots == 0)+(face->next->next->next->slots == 0);
			while (face->inv->Lnext) face = face->next;
			face->inv->Lnext = facesHEX[count];
			facesHEX[count] = face->inv;
		}
		else if (face->is_on_boundary && face->slots == 1)
		{
			if (face->inv->Lnext == 0)
			{
				if (facesHEX[0] == 0)
				{
					facesHEX[0] = face->inv;
					facesHEX[5] = 0;
				}
				if (facesHEX[5] != 0)
				{
					facesHEX[5]->Lnext = face->inv;
				}
				facesHEX[5] = face->inv;
			}
		}
	}
	checkFaceHEX();
}

CFspinEdge* CFencoder::getFaceHEX()
{
	CFspinEdge* face;
	int count = 5;
	while (count--)
	{
		while (facesHEX[count])
		{
			face = facesHEX[count];
			facesHEX[count] = face->Lnext;
			if (face->inv->slots != -1)
			{
				if (count)
				{
					while (face->inv->slots != 0)
					{
						face = face->next;
					}
					if (face->next->inv->slots == 0)
					{
						face = face->next;
					}
					if (face->next->inv->slots == 0)
					{
						face = face->next;
					}
					if (face->next->inv->slots == 0)
					{
						face = face->next;
					}
				}
				return face;
			}
		}
	}
	return 0;
}

void CFencoder::checkFaceHEX()
{
	return;
	CFspinEdge* face;
	int count = 5;
	if (facesHEX[0])
	{
		if (facesHEX[5]->Lnext)
		{
			cout << "PROBLEM" << endl;
		}
	}
	while (count--)
	{
		face = facesHEX[count];
		while (face)
		{
			if (face->inv->slots != -1)
			{
				if (count)
				{
					if ((face->inv->slots == 0)+(face->next->inv->slots == 0)+(face->next->next->inv->slots == 0)+(face->next->next->next->inv->slots == 0) < count)
					{
						cout << "PROBLEM" << endl;
					}
				}
				else
				{
					if ((face->is_on_boundary && face->inv->slots == 1)+(face->next->is_on_boundary && face->next->inv->slots == 1)+(face->next->next->is_on_boundary && face->next->next->inv->slots == 1)+(face->next->next->next->is_on_boundary && face->next->next->next->inv->slots == 1) == 0)
					{
						cout << "PROBLEM" << endl;
					}
				}
			}
			face = face->Lnext;
		}
	}
}

void CFencoder::encodePosition7(CFcodec* codec, PositionQuantizer* pq, CFvertex* V7, CFvertex* V0) // predict as V0
{
	pred_last++;

	if (!V0->active)
	{
		cout << "WARNING: vertex not active" << endl; 
	}

	dist_last += VecDistance3fv(V0->v, V7->v);

  int Corr[3];
	pq->GetCorrector(V0->vq, V7->vq, Corr);

	codec->writePositionCorrector(Corr);
}

void CFencoder::encodePosition7(CFcodec* codec, PositionQuantizer* pq, CFvertex* V7, CFvertex* V0, CFvertex* VV) // uses vertex from previous layer
{
	pred_ray++;

	if (!V0->active || !VV->active)
	{
		cout << "WARNING: vertex not active" << endl; 
	}

	float P[3];

	VecSubtract3fv(P, V0->v, VV->v);
	VecSelfAdd3fv(P, V0->v);

	dist_ray += VecDistance3fv(P, V7->v);

  int Pred[3];
	int Corr[3];

	pq->EnQuantize(P, Pred);
	pq->GetCorrector(Pred, V7->vq, Corr);

	codec->writePositionCorrector(Corr);
}

void CFencoder::encodePosition6(CFcodec* codec, PositionQuantizer* pq, CFvertex* V6, CFvertex* V1, CFvertex* V0, CFvertex* V7)
{
	pred_para++;

	if (!V0->active || !V1->active || !V7->active)
	{
		cout << "WARNING: vertex not active" << endl; 
	}

	float P[3] = {0.0f, 0.0f, 0.0f};

	VecSelfAdd3fv(P, V1->v);
	VecSelfSubtract3fv(P, V0->v);
	VecSelfAdd3fv(P, V7->v);

	dist_para += VecDistance3fv(P, V6->v);

  int Pred[3];
	int Corr[3];

	pq->EnQuantize(P, Pred);
	pq->GetCorrector(Pred, V6->vq, Corr);

	codec->writePositionCorrector(Corr);
}

int CFencoder::encodePosition5(CFcodec* codec, PositionQuantizer* pq, CFvertex* V5, CFvertex* V0, CFvertex* V1, CFvertex* V2, CFvertex* V3, CFvertex* V6, CFvertex* V7)
{
	pred_para++;

  if (!V0->active || !V1->active || !V2->active || !V3->active || !V6->active || !V7->active)
	{
		cout << "WARNING: vertex not active" << endl; 
	}

	float P[3] = {0.0f, 0.0f, 0.0f};

	VecSelfAdd3fv(P, V2->v);
	VecSelfSubtract3fv(P, V1->v);
	VecSelfAdd3fv(P, V6->v);

	dist_para += VecDistance3fv(P, V5->v);

  int Pred[3];
	int Corr[3];

	pq->EnQuantize(P, Pred);
	pq->GetCorrector(Pred, V5->vq, Corr);

	codec->writePositionCorrector(Corr);

	return 1;
}

#define LORENZO_PREDICTION

void CFencoder::encodePosition4(CFcodec* codec, PositionQuantizer* pq, CFvertex* V4, CFvertex* V0, CFvertex* V2, CFvertex* V3, CFvertex* V5, CFvertex* V6, CFvertex* V7, CFvertex* V1)
{
/*
	fprintf(stderr, "predicting  v4: %g %g %g\n", V4->v[0], V4->v[1], V4->v[2]); 
	fprintf(stderr, "from        v0: %g %g %g\n", V0->v[0], V0->v[1], V0->v[2]); 
	fprintf(stderr, "from        v1: %g %g %g\n", V1->v[0], V1->v[1], V1->v[2]); 
	fprintf(stderr, "from        v2: %g %g %g\n", V2->v[0], V2->v[1], V2->v[2]); 
	fprintf(stderr, "from        v3: %g %g %g\n", V3->v[0], V3->v[1], V3->v[2]); 
	fprintf(stderr, "from        v5: %g %g %g\n", V5->v[0], V5->v[1], V5->v[2]); 
	fprintf(stderr, "from        v6: %g %g %g\n", V6->v[0], V6->v[1], V6->v[2]); 
	fprintf(stderr, "from        v7: %g %g %g\n", V7->v[0], V7->v[1], V7->v[2]); 
  exit(1);
*/

	if (!V0->active || !V2->active || !V3->active || !V5->active || !V6->active || !V7->active || !V1->active)
	{
		cout << "WARNING: vertex not active" << endl; 
	}

	float P[3] = {0.0f, 0.0f, 0.0f};

#ifdef LORENZO_PREDICTION
	pred_lorenzo++;

  VecSelfAdd3fv(P, V3->v);
  VecSelfAdd3fv(P, V7->v);
  VecSelfAdd3fv(P, V5->v);
  VecSelfSubtract3fv(P, V6->v);
  VecSelfSubtract3fv(P, V0->v);
  VecSelfSubtract3fv(P, V2->v);
  VecSelfAdd3fv(P, V1->v);

	int Pred[3];
	int Corr[3];

	dist_lorenzo += VecDistance3fv(P, V4->v);

	pq->EnQuantize(P, Pred);
	pq->GetCorrector(Pred, V4->vq, Corr);

	codec->writePositionCorrectorOther(Corr);
#else
	pred_para++;

	VecSelfAddScalarMult3fv(P, V5->v,  1.0f);
	VecSelfAddScalarMult3fv(P, V6->v, -1.0f);
	VecSelfAddScalarMult3fv(P, V7->v,  1.0f);

  int Pred[3];
	int Corr[3];

	dist_para += VecDistance3fv(P, V4->v);

	pq->EnQuantize(P, Pred);
	pq->GetCorrector(Pred, V4->vq, Corr);

	codec->writePositionCorrector(Corr);
#endif
}

void CFencoder::encodePosition3(CFcodec* codec, PositionQuantizer* pq, CFvertex* V3, CFvertex* V2, CFvertex* V1, CFvertex* V0)
{
	pred_para++;

	if (!V0->active || !V1->active || !V2->active)
	{
		cout << "WARNING: vertex not active" << endl; 
	}

	float P[3] = {0.0f, 0.0f, 0.0f};

	VecSelfAddScalarMult3fv(P, V0->v,  1.0f);
	VecSelfAddScalarMult3fv(P, V1->v, -1.0f);
	VecSelfAddScalarMult3fv(P, V2->v,  1.0f);

	int Pred[3];
	int Corr[3];

	pq->EnQuantize(P, Pred);
	pq->GetCorrector(Pred, V3->vq, Corr);

	codec->writePositionCorrector(Corr);
}

void CFencoder::encodePosition2(CFcodec* codec, PositionQuantizer* pq, CFvertex* V2, CFvertex* V1)
{
	pred_last++;

	if (!V1->active)
	{
		cout << "WARNING: vertex not active" << endl; 
	}

	int Corr[3];
	pq->GetCorrector(V1->vq, V2->vq, Corr);

	codec->writePositionCorrector(Corr);
}

void CFencoder::encodePosition1(CFcodec* codec, PositionQuantizer* pq, CFvertex* V1, CFvertex* V0)
{
	pred_last++;

	if (!V0->active)
	{
		cout << "WARNING: vertex not active" << endl; 
	}

	int Corr[3];
	pq->GetCorrector(V0->vq, V1->vq, Corr);

	codec->writePositionCorrector(Corr);
}

void CFencoder::encodePosition0(CFcodec* codec, PositionQuantizer* pq, CFvertex* V0)
{
	pred_null++;

	int Corr[3];
	pq->GetCorrector(0, V0->vq, Corr);

	codec->writePositionCorrector(Corr);
}

void CFencoder::addEdgeVertexTET(CFspinEdge* edge)
{
	int index = edge->vertex->index;
	edge->Lnext = verticesTET[index];
	verticesTET[index] = edge;
}

CFspinEdge* CFencoder::getEdgeVertexTET()
{
	while (verticesTETdone < verticesTETnumber)
	{
		while (verticesTET[verticesTETdone])
		{
			if ((verticesTET[verticesTETdone]->slots != -1) && (verticesTET[verticesTETdone]->inv->group != -1))
			{
				return verticesTET[verticesTETdone]->inv;
			}
			else
			{
				verticesTET[verticesTETdone] = verticesTET[verticesTETdone]->Lnext;
			}
		}
		verticesTETdone++;
	}
	return 0;
}

void CFencoder::checkFaceTET()
{
	return;
	CFspinEdge* face;
	int count = 4;
	while (count--)
	{
		face = facesTET[count];
		while (face)
		{
			if (face->inv->slots != -1)
			{
				if (count)
				{
					if ((face->inv->slots == 0)+(face->next->inv->slots == 0)+(face->next->next->inv->slots == 0) < count)
					{
						cout << "PROBLEM" << endl;
					}
				}
				else
				{
					if ((face->is_on_boundary && face->inv->slots == 1)+(face->next->is_on_boundary && face->next->inv->slots == 1)+(face->next->next->is_on_boundary && face->next->next->inv->slots == 1)+(face->inv->slots == 0)+(face->next->inv->slots == 0)+(face->next->next->inv->slots == 0) == 0)
					{
						cout << "PROBLEM" << endl;
					}
				}
			}
			face = face->Lnext;
		}
	}
}

void CFencoder::addFaceTET(CFspinEdge* face)
{
	if (face->inv->group != -1)
	{
		int count = (face->slots == 0)+(face->next->slots == 0)+(face->next->next->slots == 0);
		if (count)
		{
			if (face->inv->Lnext == 0)
			{
				face->inv->Lnext = facesTET[count];
				facesTET[count] = face->inv;
			}
			else
			{
				cout << "PROBLEM" << endl;
			}
		}
		else
		{
			count = 3;
			while (count--)
			{
				if (face->is_on_boundary && (face->slots == 1))
				{
					if (face->inv->Lnext == 0)
					{
						face->inv->Lnext = facesTET[0];
						facesTET[0] = face->inv;
						return;
					}
					else
					{
						cout << "PROBLEM" << endl;
					}
				}
				face = face->next;
			}
		}
	}
	checkFaceTET();
}

void CFencoder::updateFaceTET(CFspinEdge* face)
{
	if (face->inv->group != -1)
	{
		if (face->slots == 0)
		{
			int count = 1+(face->next->slots == 0)+(face->next->next->slots == 0);
			if (face->inv->Lnext == 0)
			{
				face->inv->Lnext = facesTET[count];
				facesTET[count] = face->inv;
			}
			else
			{
//				cout << "PROBLEM" << endl;
				if (face->inv->Lnext) face = face->next;
				if (face->inv->Lnext) face = face->next;
				if (face->inv->Lnext)
				{
					cout << "SERIOUS PROBLEM" << endl;
				}
				face->inv->Lnext = facesTET[count];
				facesTET[count] = face->inv;

/*
				CFspinEdge* edge;
				int count = 4;
				while (count--)
				{
					edge = facesTET[count];
					while (edge)
					{
						if (edge == face->inv)
						{
									cout << "PROBLEM " << count << endl;
						}
						if (edge->inv->slots != -1)
						{
							if (count)
							{
								if ((edge->inv->slots == 0)+(edge->next->inv->slots == 0)+(edge->next->next->inv->slots == 0) < count)
								{
									cout << "PROBLEM" << endl;
								}
							}
							else
							{
								if ((edge->is_on_boundary && edge->inv->slots == 1)+(edge->next->is_on_boundary && edge->next->inv->slots == 1)+(edge->next->next->is_on_boundary && edge->next->next->inv->slots == 1)+(edge->inv->slots == 0)+(edge->next->inv->slots == 0)+(edge->next->next->inv->slots == 0) == 0)
								{
									cout << "PROBLEM" << endl;
								}
							}
						}
						edge = edge->Lnext;
					}
				}
*/
			}
		}
		else if (face->is_on_boundary && face->slots == 1)
		{
			if (face->inv->Lnext == 0)
			{
				face->inv->Lnext = facesTET[0];
				facesTET[0] = face->inv;
			}
			else
			{
				cout << "PROBLEM" << endl;
			}
		}
	}
	checkFaceTET();
}

CFspinEdge* CFencoder::getFaceTET()
{
	CFspinEdge* face;
	int count = 4;
	while (count--)
	{
		while (facesTET[count])
		{
			face = facesTET[count];
			facesTET[count] = face->Lnext;
			if (face->inv->slots != -1)
			{
				if (count)
				{
					while (face->inv->slots != 0)
					{
						face = face->next;
					}
					if (face->next->inv->slots == 0)
					{
						face = face->next;
					}
					if (face->next->inv->slots == 0)
					{
						face = face->next;
					}
				}
				return face;
			}
		}
	}
	return 0;
}
