/*
===============================================================================

	FILE:	encoder.h
	
	CONTENTS:
	
    This is the actual encoder. It can encode TET and HEX meshes. It can do
    so step by step (which is used for visualization).
	
	PROGRAMMERS:
	
		martin isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		(c) 2002 martin isenburg@cs.unc.edu
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
		20 May 2002 -- created
	
===============================================================================
*/
#ifndef CF_ENCODER_H
#define CF_ENCODER_H

class CFmesh;
class CFcodec;
class CFvertex;
class CFspinEdge;
class PositionQuantizer;

class CFencoder
{
public:

	int case0;
	int case1;
	int case2a;
	int case2b;
	int case3a;
	int case3b;
	int case4;

	int pred_null;
	int pred_last;
	int pred_ray;
	int pred_para;
  int pred_lorenzo;

  float dist_null;
  float dist_last;
  float dist_ray;
	float dist_para;
  float dist_lorenzo;

  int encodeStep(CFmesh*, CFcodec*, PositionQuantizer*);

  int encodeStepTET(CFmesh*, CFcodec*);
  int encodeStepHEX(CFmesh*, CFcodec*, PositionQuantizer*);

	CFspinEdge* pick_best_gateTET(CFmesh* mesh);

	int verticesTETdone;
	int verticesTETcount;
	int verticesTETnumber;
	CFspinEdge** verticesTET;
	void addEdgeVertexTET(CFspinEdge*);
	CFspinEdge* getEdgeVertexTET();
	int getSlot2EdgeTET(CFspinEdge* edge);

	CFspinEdge* facesTET[5];
	void addFaceTET(CFspinEdge*);
	void updateFaceTET(CFspinEdge*);
	CFspinEdge* getFaceTET();
	void checkFaceTET();

	CFspinEdge* pick_best_gateHEX(CFmesh* mesh);

	int verticesHEXdone;
	int verticesHEXcount;
	int verticesHEXnumber;
	CFspinEdge** verticesHEX;
	void addEdgeVertexHEX(CFspinEdge*);
	CFspinEdge* getEdgeVertexHEX();
	int getSlot1EdgeHEX(CFspinEdge* edge);

	CFspinEdge* facesHEX[6];
	void addFaceHEX(CFspinEdge*);
	void updateFaceHEX(CFspinEdge*);
	CFspinEdge* getFaceHEX();
	void checkFaceHEX();

	void encodePosition7(CFcodec*, PositionQuantizer*, CFvertex* V7, CFvertex* V0); // for now predict as V0
	void encodePosition7(CFcodec*, PositionQuantizer*, CFvertex* V7, CFvertex* V0, CFvertex* V1, CFvertex* V3);
	void encodePosition7(CFcodec*, PositionQuantizer*, CFvertex* V7, CFvertex* V0, CFvertex* VV);
	void encodePosition6(CFcodec*, PositionQuantizer*, CFvertex* V6, CFvertex* V0, CFvertex* V1, CFvertex* V7); 
//	int encodePosition5(CFcodec*, PositionQuantizer*, CFvertex* V5, CFvertex* V1, CFvertex* V2, CFvertex* V6);
	int encodePosition5(CFcodec*, PositionQuantizer*, CFvertex* V5, CFvertex* V0, CFvertex* V1, CFvertex* V2, CFvertex* V3, CFvertex* V6, CFvertex* V7);
	void encodePosition4(CFcodec*, PositionQuantizer*, CFvertex* V4, CFvertex* V0, CFvertex* V2, CFvertex* V3, CFvertex* V5, CFvertex* V6, CFvertex* V7, CFvertex* V1);
	void encodePosition3(CFcodec*, PositionQuantizer*, CFvertex* V3, CFvertex* V0, CFvertex* V1, CFvertex* V2);
	void encodePosition2(CFcodec*, PositionQuantizer*, CFvertex* V2, CFvertex* V1);
	void encodePosition1(CFcodec*, PositionQuantizer*, CFvertex* V1, CFvertex* V0);
	void encodePosition0(CFcodec*, PositionQuantizer*, CFvertex* V0);

  CFencoder();
  ~CFencoder();
};

#endif
