/*
===============================================================================

	FILE:	loader.cpp
	
	CONTENTS:
	
    see corresponding header file
	
	PROGRAMMERS:
	
		martin isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		(c) 2002 martin isenburg@cs.unc.edu
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
    see corresponding header file
	
===============================================================================
*/
#include "loader.h"

#include "vector.h"
#include "mesh.h"
#include "spinedge.h"
#include "vertex.h"
#include "normal.h"
#include "material.h"
#include "group.h"
#include "vec3fv.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream.h>

CFloader::CFloader()
{
}

CFloader::~CFloader()
{
}

static void quicksort_for_connectivity(void** a, int i, int j)
{
	int in_i = i;
	int in_j = j;
	void* w;
	__int64 key = ((CFspinEdge*)a[(i+j)/2])->marker;
	do
	{
		while ( ((CFspinEdge*)a[i])->marker > key ) i++;
		while ( ((CFspinEdge*)a[j])->marker < key ) j--;
		if (i<j)
		{
			w = a[i];
			a[i] = a[j];
			a[j] = w;
		}
	} while (++i<=--j);
	if (i == j+3)
	{
		i--;
		j++;
	}
	if (j>in_i) quicksort_for_connectivity(a, in_i, j);
	if (i<in_j) quicksort_for_connectivity(a, i, in_j);
}

int CFloader::loadTet(char* file_name, CFmesh* mesh)
{
	FILE *file;
	char line[255];
	
	int i;
	float x, y, z;
	int v0_i, v1_i, v2_i, v3_i, vhelp_i;
	__int64 v0_idx, v1_idx, v2_idx, v3_idx;
	CFvertex *v0, *v1, *v2, *v3;
	CFspinEdge *e0, *e1, *e2, *e00, *e01, *e02, *e10, *e11, *e12, *e20, *e21, *e22;

	int pos_vol = 0;
	int neg_vol = 0;
	int minus = 0;
	
	if (file = fopen(file_name, "r"))
	{
		cout << "opened file '" << file_name << "'" << endl;
	}
	else
	{
		cout << "cannot open file '" << file_name << "'" << endl;
		return 0;
	}
	
	// we use this group index for all faces in case the model has no group structure 	 
	char current_group_name[128];
	strcpy(current_group_name, "no group");
	int current_group = mesh->gs->size();
	mesh->gs->addElement((void*) new CFgroup(current_group_name));
	
	// we use this material index for all boundary in case the model has no material properties		
	char boundary_material_name[128];
	strcpy(boundary_material_name, "boundary");
	int boundary_material = mesh->ms->size();
	mesh->boundary_material = new CFmaterial(boundary_material_name, 0.0f, 0.0f, 1.0f);
	mesh->ms->addElement(mesh->boundary_material);

	// we use this material index for all interior in case the model has no material properties		
	char interior_material_name[128];
	strcpy(interior_material_name, "interior");
	int interior_material = mesh->ms->size();
	mesh->interior_material = new CFmaterial(interior_material_name);
	mesh->ms->addElement(mesh->interior_material);
	
	while (fgets(line, sizeof(char) * 255, file))
	{
		if (*line == 'v')
		{
			if (*(line+1) == ' ')
			{
				sscanf(line+2, "%f %f %f %f", &x, &y, &z);
				mesh->vs->addElement((void*) new CFvertex(x, y, z));
			}
			else
			{
				/* normals, textures, and else are not implemented */
			}
		}
		else if (*line == 'g')
		{
			if (sscanf(line+2, "%s", current_group_name) == 1) // prevent no-name groups
			{
				current_group = -1;
				for (i = 0; i < mesh->gs->size(); i++)
				{
					if (strcmp(&current_group_name[0], ((CFgroup*)mesh->gs->elementAt(i))->name) == 0)
					{
						current_group = i;
					}
				}
				if (current_group == -1)
				{
					current_group = mesh->gs->size();
					mesh->gs->addElement((void*) new CFgroup(current_group_name));
				}
			}
		}
		else if (*line == 't')
		{
			if (*(line+1) == ' ') // read in a tetrahedron
			{
				sscanf(line+2, "%d %d %d %d", &v0_i, &v1_i, &v2_i, &v3_i);

				v0 = (CFvertex*)mesh->vs->elementAt(v0_i - minus);
				v1 = (CFvertex*)mesh->vs->elementAt(v1_i - minus);
				v2 = (CFvertex*)mesh->vs->elementAt(v2_i - minus);
				v3 = (CFvertex*)mesh->vs->elementAt(v3_i - minus);

        v0->marker = 1;
        v1->marker = 1;
        v2->marker = 1;
        v3->marker = 1;

				float u1[3], u2[3], u3[3], u1u2[3];
				VecSubtract3fv(u1, v1->v, v0->v);
				VecSubtract3fv(u2, v2->v, v0->v);
				VecSubtract3fv(u3, v3->v, v0->v);
				VecCrossProd3fv(u1u2, u1, u2);
				if (VecDotProd3fv(u1u2, u3) > 0.0f)
				{
					pos_vol++;
				}
				else
				{
					neg_vol++;
					vhelp_i = v2_i;
					v2_i = v3_i;
					v3_i = vhelp_i;
				}
				// create all spin edges
				e0 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v0_i - minus));
				e1 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v1_i - minus));
				e2 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v2_i - minus));
				
				e00 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v1_i - minus));
				e01 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v0_i - minus));
				e02 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v3_i - minus));
				
				e10 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v3_i - minus));
				e11 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v2_i - minus));
				e12 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v1_i - minus));
				
				e20 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v2_i - minus));
				e21 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v3_i - minus));
				e22 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v0_i - minus));
				
				// preliminary next pointer (probably final pointer)
				e0->next = e2;
				e1->next = e0;
				e2->next = e1;
				
				e00->next = e02;
				e01->next = e00;
				e02->next = e01;
				
				e10->next = e12;
				e11->next = e10;
				e12->next = e11;
				
				e20->next = e22;
				e21->next = e20;
				e22->next = e21;
				
				// preliminary spin pointer
				e0->spin = e00;
				e1->spin = e11;
				e2->spin = e22;
				
				e00->spin = e0;
				e01->spin = e21;
				e02->spin = e12;
				
				e10->spin = e20;
				e11->spin = e1;
				e12->spin = e02;
				
				e20->spin = e10;
				e21->spin = e01;
				e22->spin = e2;

				// set key for all edges so we can recover connectivity with quicksort
				// (NOTE: assuming vertex index is < 1048576)
				// (NOTE: maximal 21 bits for vertex index)
				v0_idx = v0_i; v1_idx = v1_i; v2_idx = v2_i; v3_idx = v3_i;

				if (v0_idx > v1_idx)
				{
					e0->marker = (v2_idx << 42) | (v0_idx << 21) | (v1_idx);
				}
				else
				{
					e0->marker = (v2_idx << 42) | (v1_idx << 21) | (v0_idx);
				}
				if (v1_idx > v2_idx)
				{
					e1->marker = (v0_idx << 42) | (v1_idx << 21) | (v2_idx);
				}
				else
				{
					e1->marker = (v0_idx << 42) | (v2_idx << 21) | (v1_idx);
				}
				if (v2_idx > v0_idx)
				{
					e2->marker = (v1_idx << 42) | (v2_idx << 21) | (v0_idx);
				}
				else
				{
					e2->marker = (v1_idx << 42) | (v0_idx << 21) | (v2_idx);
				}
				if (v1_idx > v0_idx)
				{
					e00->marker = (v3_idx << 42) | (v1_idx << 21) | (v0_idx);
				}
				else
				{
					e00->marker = (v3_idx << 42) | (v0_idx << 21) | (v1_idx);
				}
				if (v0_idx > v3_idx)
				{
					e01->marker = (v1_idx << 42) | (v0_idx << 21) | (v3_idx);
				}
				else
				{
					e01->marker = (v1_idx << 42) | (v3_idx << 21) | (v0_idx);
				}
				if (v3_idx > v1_idx)
				{
					e02->marker = (v0_idx << 42) | (v3_idx << 21) | (v1_idx);
				}
				else
				{
					e02->marker = (v0_idx << 42) | (v1_idx << 21) | (v3_idx);
				}
				if (v3_idx > v2_idx)
				{
					e10->marker = (v1_idx << 42) | (v3_idx << 21) | (v2_idx);
				}
				else
				{
					e10->marker = (v1_idx << 42) | (v2_idx << 21) | (v3_idx);
				}
				if (v2_idx > v1_idx)
				{
					e11->marker = (v3_idx << 42) | (v2_idx << 21) | (v1_idx);
				}
				else
				{
					e11->marker = (v3_idx << 42) | (v1_idx << 21) | (v2_idx);
				}
				if (v1_idx > v3_idx)
				{
					e12->marker = (v2_idx << 42) | (v1_idx << 21) | (v3_idx);
				}
				else
				{
					e12->marker = (v2_idx << 42) | (v3_idx << 21) | (v1_idx);
				}
				if (v2_idx > v3_idx)
				{
					e20->marker = (v0_idx << 42) | (v2_idx << 21) | (v3_idx);
				}
				else
				{
					e20->marker = (v0_idx << 42) | (v3_idx << 21) | (v2_idx);
				}
				if (v3_idx > v0_idx)
				{
					e21->marker = (v2_idx << 42) | (v3_idx << 21) | (v0_idx);
				}
				else
				{
					e21->marker = (v2_idx << 42) | (v0_idx << 21) | (v3_idx);
				}
				if (v0_idx > v2_idx)
				{
					e22->marker = (v3_idx << 42) | (v0_idx << 21) | (v2_idx);
				}
				else
				{
					e22->marker = (v3_idx << 42) | (v2_idx << 21) | (v0_idx);
				}
				
				mesh->es->addElement(e0);
				mesh->es->addElement(e1);
				mesh->es->addElement(e2);
				mesh->es->addElement(e00);
				mesh->es->addElement(e01);
				mesh->es->addElement(e02);
				mesh->es->addElement(e10);
				mesh->es->addElement(e11);
				mesh->es->addElement(e12);
				mesh->es->addElement(e20);
				mesh->es->addElement(e21);
				mesh->es->addElement(e22);

				mesh->fs->addElement(e0);
				mesh->fs->addElement(e00);
				mesh->fs->addElement(e10);
				mesh->fs->addElement(e20);
				
				mesh->ts->addElement(e0);
			}
		}
		else if ((*line == 'u') && (*(line+1) == 's') && (*(line+2) == 'e'))
		{
			sscanf(line+6, "%s", interior_material_name);		
			interior_material = -1;
			for (i = 0; i < mesh->ms->size(); i++)
			{
				if (strcmp(interior_material_name, ((CFmaterial*)mesh->ms->elementAt(i))->name) == 0)
				{
					interior_material = i;
					break;
				}
			}
			if (interior_material == -1)
			{
				interior_material = mesh->ms->size()-1;
			}
		} 
		else if ((*line == 'm') && (*(line+1) == 'i') && (*(line+2) == 'n'))
		{
			sscanf(line+5, "%d", &minus);		
		}
	}
		
	cout << "# of pos_vol     : " << pos_vol << endl;
	cout << "# of neg_vol     : " << neg_vol << endl;

	cout << "# of vertices     : " << mesh->vs->size() << endl;
	cout << "# of tetrahedrons : " << mesh->ts->size() << endl;
	cout << "# of normals      : " << mesh->ns->size() << endl;
	cout << "# of groups       : " << mesh->gs->size() << endl;
	cout << "# of i-spinedges  : " << mesh->es->size() << endl;
	cout << "# of i-faces      : " << mesh->fs->size() << endl;
	
	quicksort_for_connectivity( mesh->es->data, 0, mesh->es->size()-1 );
		
	mesh->interior_edges = mesh->es->size();
	mesh->interior_faces = mesh->fs->size();
	
  minus = 0;
	for (i = 0; i < mesh->vs->size(); i++)
	{
		v0 = (CFvertex*)mesh->vs->elementAt(i);
    if (v0->marker)
    {
      minus++;
      v0->marker = 0;
    }
  }

  if (minus < mesh->vs->size()) cout << "number vertices actually used: " << minus << endl;


  // set inv pointer for all edges (and create boundary edges where needed)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		e1 = (CFspinEdge*)mesh->es->elementAt(i+1);
		e2 = (CFspinEdge*)mesh->es->elementAt(i+2);
		
		if (e0->inv == 0) // if we did not set the inv pointer yet
		{
			if ( e1 && (e0->marker == e1->marker) )
			{
				if ( e2 && (e1->marker == e2->marker) ) // non-manifold situation
				{
					cout << "non-manifold situation" << endl;
					return 0;
				}
				else
				{
					if ( e0->vertex == e1->vertex ) // non-orientated situation
					{
						cout << "non-orientated situation" << endl;
						return 0;
					}
					else
					{
						e0->inv = e1;
						e1->inv = e0;
						e0->next->inv = e1->next->next;
						e1->next->next->inv = e0->next;
						e0->next->next->inv = e1->next;
						e1->next->inv = e0->next->next;
						i++;
					}
				}
			}
			else
			{
				e00 = new CFspinEdge(-1, boundary_material, e0->next->next->vertex, 0); // group index -1 means hole
				e01 = new CFspinEdge(-1, boundary_material, e0->vertex, 0); // group index -1 means hole
				e02 = new CFspinEdge(-1, boundary_material, e0->next->vertex, 0); // group index -1 means hole
				e00->next = e02;
				e01->next = e00;
				e02->next = e01;
				mesh->es->addElement(e00);
				mesh->es->addElement(e01);
				mesh->es->addElement(e02);
				mesh->fs->addElement(e00);
				e0->inv = e00;
				e00->inv = e0;
				e0->next->inv = e01;
				e01->inv = e0->next;
				e0->next->next->inv = e02;
				e02->inv = e0->next->next;
			}
		}
	}
	
	mesh->total_edges = mesh->es->size();
	mesh->total_faces = mesh->fs->size();
	mesh->total_vertices = mesh->vs->size();

	cout << "# of b-spinedges  : " << (mesh->total_edges-mesh->interior_edges) << endl;
	cout << "# of b-faces      : " << (mesh->total_faces-mesh->interior_faces) << endl;
	
	// set spin pointer for all non-boundary spinedges (and set marker to zero)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		e0->spin = e0->spin->inv;
		e0->marker = 0;
	}
	
	// set spin pointer for all boundary spinedges
	for (i = mesh->interior_edges; i < mesh->total_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i); 		 
		e1 = e0->inv->spin;
		while (e1->group != -1)
		{
			e1 = e1->spin;
		}
		e0->spin = e1->inv;
	}
	return 1;
}

int CFloader::loadMtl(char* file_name, CFmesh* mesh)
{
	FILE *file;
	char line[255];  
	float r, g, b;
	CFmaterial* m;
	
	if (file = fopen(file_name, "r"))
	{
		//			cout << "opened file '" << file_name << "'" << endl;
	}
	else
	{
		cout << "cannot open file '" << file_name << "'" << endl;
	}
	
	while (fgets(line, sizeof(char) * 255, file))
	{
		if (*line == '\n')
		{
			// empty line
		}
		else if ((*line == 'n') && (*(line+1) == 'e') && (*(line+2) == 'w'))
		{
			m = new CFmaterial();
			sscanf(line+6, "%s", m->name);
			mesh->ms->addElement((void*) m);
		}
		else if (*line == 'K')
		{
			sscanf(line+3, "%f %f %f", &r, &g, &b);
			
			if ((*(line+1) == 'a'))
			{
				VecSet3fv(m->a, r, g, b);
			}
			else if ((*(line+1) == 'd'))
			{
				VecSet3fv(m->d, r, g, b);
			}
			else if ((*(line+1) == 's'))
			{
				VecSet3fv(m->s, r, g, b);
			}
			else if ((*(line+1) == 'e'))
			{
				VecSet3fv(m->e, r, g, b);
			}
			else
			{
				cout << "cannot understand: " << line << endl;
			}
		}
		else 
		{
			cout << "cannot understand: " << line << endl;
		}
	}
	cout << "# of materials : " << mesh->ms->size() << endl;
	return 1;
}

int CFloader::loadPol(char* file_name, CFmesh* mesh)
{
	FILE *file;
	char line[2550];
	
	int i;
	float x, y, z;
	int v_idx;
	__int64 v0_idx, v1_idx, v2_idx, v3_idx;
	CFspinEdge *e0, *e1, *e2, *e3, *e4;

	int pos_vol = 0;
	int neg_vol = 0;
	int minus = 1;
	
	if (file = fopen(file_name, "r"))
	{
		cout << "opened file '" << file_name << "'" << endl;
	}
	else
	{
		cout << "cannot open file '" << file_name << "'" << endl;
		return 0;
	}
	
	// we use this group index for all faces in case the model has no group structure 	 
	char current_group_name[128];
	strcpy(current_group_name, "no group");
	int current_group = mesh->gs->size();
	mesh->gs->addElement((void*) new CFgroup(current_group_name));
	
	// we use this material index for all boundary in case the model has no material properties		
	char boundary_material_name[128];
	strcpy(boundary_material_name, "boundary");
	int boundary_material = mesh->ms->size();
	mesh->boundary_material = new CFmaterial(boundary_material_name, 0.0f, 0.0f, 1.0f);
	mesh->ms->addElement(mesh->boundary_material);

	// we use this material index for all interior in case the model has no material properties		
	char interior_material_name[128];
	strcpy(interior_material_name, "interior");
	int interior_material = mesh->ms->size();
	mesh->interior_material = new CFmaterial(interior_material_name);
	mesh->ms->addElement(mesh->interior_material);

	while (fgets(line, sizeof(char) * 2550, file))
	{
		if (*line == 'v')
		{
			if (*(line+1) == ' ')
			{
				sscanf(line+2, "%f %f %f", &x, &y, &z);
				mesh->vs->addElement((void*) new CFvertex(x, y, z, mesh->vs->size()));
			}
			else
			{
				/* normals, textures, and else are not implemented */
			}
		}
		else if (*line == 'g')
		{
			if (sscanf(line+2, "%s", current_group_name) == 1) // prevent no-name groups
			{
				current_group = -1;
				for (i = 0; i < mesh->gs->size(); i++)
				{
					if (strcmp(&current_group_name[0], ((CFgroup*)mesh->gs->elementAt(i))->name) == 0)
					{
						current_group = i;
					}
				}
				if (current_group == -1)
				{
					current_group = mesh->gs->size();
					mesh->gs->addElement((void*) new CFgroup(current_group_name));
				}
			}
		}
		else if ((*(line) == 'p') && (*(line+1) == ' ')) // read in a polyhedron
		{
			int pol_num_faces, pol_num_vertices;
			
	    char* pointer = line + 1;

			sscanf(pointer, "%d", &pol_num_faces);
	    while ((*pointer  < '0') || (*pointer  > '9')) pointer++;
	    while ((*pointer >= '0') && (*pointer <= '9')) pointer++;

			int pol_num_edges_before = mesh->es->size();

			while (pol_num_faces)
			{
				sscanf(pointer, "%d", &pol_num_vertices);
		    while ((*pointer  < '0') || (*pointer  > '9')) pointer++;
		    while ((*pointer >= '0') && (*pointer <= '9')) pointer++;

				e2 = 0;

				while (pol_num_vertices)
				{
					sscanf(pointer, "%d", &v_idx);
			    while ((*pointer  < '0') || (*pointer  > '9')) pointer++;
			    while ((*pointer >= '0') && (*pointer <= '9')) pointer++;

					e1 = new CFspinEdge(current_group, interior_material, (CFvertex*)mesh->vs->elementAt(v_idx-1));

					if (e2)
					{
						e2->next = e1;
						v0_idx = e1->vertex->index;
						v1_idx = e2->vertex->index;
						if (v0_idx > v1_idx)
						{
							e1->marker = (v0_idx << 31) | (v1_idx);
						}
					  else
						{
							e1->marker = (v1_idx << 31) | (v0_idx);
						}
					}
					else
					{
						e0 = e1;
					}
		      mesh->es->addElement(e1);
		      e2 = e1;
					pol_num_vertices--;
				}
				mesh->fs->addElement(e0);
				e2->next = e0;
				v0_idx = e0->vertex->index;
				v1_idx = e2->vertex->index;
				if (v0_idx > v1_idx)
				{
					e0->marker = (v0_idx << 31) | (v1_idx);
				}
				else
				{
					e0->marker = (v1_idx << 31) | (v0_idx);
				}
				pol_num_faces--;
			}

			int pol_num_edges_after = mesh->es->size();

			quicksort_for_connectivity( mesh->es->data, pol_num_edges_before, pol_num_edges_after-1);

			// set inv pointer for all edges of this polyhedra

			for (i = pol_num_edges_before; i < pol_num_edges_after; i++)
			{
				e0 = (CFspinEdge*)mesh->es->elementAt(i);
				e1 = (CFspinEdge*)mesh->es->elementAt(i+1);
		
				if (e0->marker == e1->marker)
				{
					e0->inv = e1;
					e1->inv = e0;
					i++;
				}
				else
				{
					cout << "PROBLEM" << endl;
				}
			}

			mesh->ps->addElement(e0);
		}
	}
	
	cout << "# of vertices     : " << mesh->vs->size() << endl;
	cout << "# of polyhedrons  : " << mesh->ps->size() << endl;
	cout << "# of i-spinedges  : " << mesh->es->size() << endl;
	cout << "# of i-faces      : " << mesh->fs->size() << endl;

	mesh->interior_edges = mesh->es->size();
	mesh->interior_faces = mesh->fs->size();

	// set key for all edges so we can recover volume connectivity with quicksort
	// (NOTE: assuming vertex index is < 1048576)
	// (NOTE: maximal 21 bits for vertex index)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		v0_idx = e0->vertex->index;
		v1_idx = e0->next->vertex->index;
		v2_idx = e0->next->next->vertex->index;
		v3_idx = e0->next->next->next->vertex->index;
		
		if (v1_idx > v2_idx)
		{
			if (v0_idx > v3_idx)
			{
				e0->next->marker = (v0_idx << 42) | (v1_idx << 21) | (v2_idx);
			}
			else
			{
				e0->next->marker = (v3_idx << 42) | (v1_idx << 21) | (v2_idx);
			}
		}
		else
		{
			if (v0_idx > v3_idx)
			{
				e0->next->marker = (v0_idx << 42) | (v2_idx << 21) | (v1_idx);
			}
			else
			{
				e0->next->marker = (v3_idx << 42) | (v2_idx << 21) | (v1_idx);
			}
		}
	}

	quicksort_for_connectivity( mesh->es->data, 0, mesh->interior_edges-1 );
			
	// set inv pointer for all edges (and create boundary edges where needed)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		e1 = (CFspinEdge*)mesh->es->elementAt(i+1);
		e2 = (CFspinEdge*)mesh->es->elementAt(i+2);
		
		if (e0->spin == 0) // if we did not set the inv pointer yet
		{
			if ( e1 && (e0->marker == e1->marker) )
			{
				if ( e2 && (e1->marker == e2->marker) ) // non-manifold situation
				{
					cout << "non-manifold situation" << endl;
					return 0;
				}
				else
				{
					if ( e0->vertex == e1->vertex ) // non-orientated situation
					{
						cout << "non-orientated situation" << endl;
						return 0;
					}
					else
					{
						e0->spin = e1;
						e1->spin = e0;
						i++;
					}
				}
			}
			else
			{
				e1 = e0;
				e3 = 0;
				do
				{
					e2 = new CFspinEdge(-1, boundary_material, e1->vertex);
					mesh->es->addElement(e2);
					e1->spin = e2;
					e2->spin = e1;
					e1 = e1->next;
					if (e3)
					{
						e2->next = e3;
					}
					else
					{
						e4 = e2;
					}
					e3 = e2;
				} while (e1 != e0);
				e4->next = e3;
				mesh->fs->addElement(e4);
			}
		}
	}
	
	mesh->total_edges = mesh->es->size();
	mesh->total_faces = mesh->fs->size();
	mesh->total_vertices = mesh->vs->size();

	cout << "# of b-spinedges  : " << (mesh->total_edges-mesh->interior_edges) << endl;
	cout << "# of b-faces      : " << (mesh->total_faces-mesh->interior_faces) << endl;

	
	for (i = 0; i < mesh->total_vertices; i++)
	{
		((CFvertex*)mesh->vs->elementAt(i))->active = 0;
	}

	for (i = 0; i < mesh->interior_edges; i++)
	{
		((CFspinEdge*)mesh->es->elementAt(i))->vertex->active = 1;
	}

	mesh->active = (CFspinEdge*)mesh->es->elementAt(0);

	// recover inv pointers of boundary (assuming edge-manifold bouyndary)

	for (i = mesh->interior_edges; i < mesh->total_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);

		v0_idx = e0->vertex->index;
		v1_idx = e0->next->vertex->index;
		if (v0_idx > v1_idx)
		{
			e0->marker = (v0_idx << 31) | (v1_idx);
		}
		else
		{
			e0->marker = (v1_idx << 31) | (v0_idx);
		}
	}

	quicksort_for_connectivity( mesh->es->data, mesh->interior_edges, mesh->total_edges-1 );

	for (i = mesh->interior_edges; i < mesh->total_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		e1 = (CFspinEdge*)mesh->es->elementAt(i+1);

		if (e0->marker == e1->marker)
		{
			e0->inv = e1;
			e1->inv = e0;
			i++;
		}
		else
		{
			cout << "PROBLEM" << endl;
		}
	}

	return 1;
}

inline static void switch_endian_32(unsigned char* data)
{
  char help;
  help = data[0];
  data[0] = data[3];
  data[3] = help;
  help = data[1];
  data[1] = data[2];
  data[2] = help;
}

int CFloader::loadClg(char* file_name, CFmesh* mesh)	
{
  int i,j,k;
	FILE *file;
  int dimension[3];

	if (file = fopen(file_name, "rb"))
	{
		cout << "opened file '" << file_name << "'" << endl;
	}
	else
	{
		cout << "cannot open file '" << file_name << "'" << endl;
		return 0;
	}

  fread(dimension, sizeof(int), 3, file);
  switch_endian_32((unsigned char*)&dimension[0]);
  switch_endian_32((unsigned char*)&dimension[1]);
  switch_endian_32((unsigned char*)&dimension[2]);

	cout << "dimensionension " << dimension[0] << "x" << dimension[1] << "x" << dimension[2] << endl;

  int nverts = dimension[0] * dimension[1] * dimension[2];
  int nhexas = (dimension[0]-1) * (dimension[1]-1) * (dimension[2]-1);

	cout << "nverts: " << nverts << "nhexas: " << nhexas << endl;

  CFvertex *v0;

  // getting x coordinates
  for (i = 0; i < nverts; i++)
  {
    v0 = new CFvertex(0,0,0,i);
    mesh->vs->addElement(v0);
    fread(&(v0->v[0]), sizeof(float), 1, file);
    switch_endian_32((unsigned char*)&(v0->v[0]));
  }

  // getting y coordinates
  for (i = 0; i < nverts; i++)
  {
    v0 = (CFvertex*) mesh->vs->elementAt(i);
    fread(&(v0->v[1]), sizeof(float), 1, file);
    switch_endian_32((unsigned char*)&(v0->v[1]));
  }

  // getting z coordinates
  for (i = 0; i < nverts; i++)
  {
    v0 = (CFvertex*) mesh->vs->elementAt(i);
    fread(&(v0->v[2]), sizeof(float), 1, file);
    switch_endian_32((unsigned char*)&(v0->v[2]));
  }

  // getting t coordinates
  for (i = 0; i < nverts; i++)
  {
    v0 = (CFvertex*) mesh->vs->elementAt(i);
    fread(&(v0->v[3]), sizeof(float), 1, file);
    switch_endian_32((unsigned char*)&(v0->v[3]));
  }

  fclose(file);

  // create materials

  // we use this group index for all faces in case the model has no group structure 	 
	char current_group_name[128];
	strcpy(current_group_name, "no group");
	int current_group = mesh->gs->size();
	mesh->gs->addElement((void*) new CFgroup(current_group_name));
	
	// we use this material index for all boundary in case the model has no material properties		
	char boundary_material_name[128];
	strcpy(boundary_material_name, "boundary");
	int boundary_material = mesh->ms->size();
	mesh->boundary_material = new CFmaterial(boundary_material_name, 0.0f, 0.0f, 1.0f);
	mesh->ms->addElement(mesh->boundary_material);

	// we use this material index for all interior in case the model has no material properties		
	char interior_material_name[128];
	strcpy(interior_material_name, "interior");
	int interior_material = mesh->ms->size();
	mesh->interior_material = new CFmaterial(interior_material_name);
	mesh->ms->addElement(mesh->interior_material);

  // creating connectivity

	int v0_i, v1_i, v2_i, v3_i, v4_i, v5_i, v6_i, v7_i;
	__int64 v0_idx, v1_idx, v2_idx, v3_idx;
	CFvertex *v1, *v2, *v3, *v4, *v5, *v6, *v7;
	CFspinEdge *e0, *e1, *e2, *e3;
	CFspinEdge *e00, *e01, *e02, *e03;
	CFspinEdge *e10, *e11, *e12, *e13;
	CFspinEdge *e20, *e21, *e22, *e23;
	CFspinEdge *e30, *e31, *e32, *e33;
	CFspinEdge *e4, *e5, *e6, *e7;

	int pos_vol = 0;
	int neg_vol = 0;
	int format = -1; // 0 = CUBIT, 1 = GAMBIT, ...

  int layer = dimension[0]*dimension[1];
  int line = dimension[0];

  for (k = 0; k < (dimension[2]-1); k++) // layer by layer
  {
    for (j = 0; j < (dimension[1]-1); j++) // line by line
    {
      for (i = 0; i < (dimension[0]-1); i++) // hexa by hexa
      {
        v0_i = k * layer +   j   * line + i;
        v1_i = k * layer +   j   * line + i+1;
        v2_i = k * layer + (j+1) * line + i+1;
        v3_i = k * layer + (j+1) * line + i;
        v7_i = (k+1) * layer +   j   * line +  i;
        v6_i = (k+1) * layer +   j   * line + i+1;
        v5_i = (k+1) * layer + (j+1) * line + i+1;
        v4_i = (k+1) * layer + (j+1) * line +  i;

				v0 = (CFvertex*)mesh->vs->elementAt(v0_i);
				v1 = (CFvertex*)mesh->vs->elementAt(v1_i);
				v2 = (CFvertex*)mesh->vs->elementAt(v2_i);
				v3 = (CFvertex*)mesh->vs->elementAt(v3_i);

				v4 = (CFvertex*)mesh->vs->elementAt(v4_i);
				v5 = (CFvertex*)mesh->vs->elementAt(v5_i);
				v6 = (CFvertex*)mesh->vs->elementAt(v6_i);
				v7 = (CFvertex*)mesh->vs->elementAt(v7_i);

/*
				cout << v0->v[0] << " " << v0->v[1] << " " << v0->v[2] << endl;
				cout << v1->v[0] << " " << v1->v[1] << " " << v1->v[2] << endl;
				cout << v2->v[0] << " " << v2->v[1] << " " << v2->v[2] << endl;
				cout << v3->v[0] << " " << v3->v[1] << " " << v3->v[2] << endl;
				cout << v4->v[0] << " " << v4->v[1] << " " << v4->v[2] << endl;
				cout << v5->v[0] << " " << v5->v[1] << " " << v5->v[2] << endl;
				cout << v6->v[0] << " " << v6->v[1] << " " << v6->v[2] << endl;
				cout << v7->v[0] << " " << v7->v[1] << " " << v7->v[2] << endl;
*/
				// create all spin edges
				e0 = new CFspinEdge(current_group, interior_material, v0);
				e1 = new CFspinEdge(current_group, interior_material, v1);
				e2 = new CFspinEdge(current_group, interior_material, v2);
				e3 = new CFspinEdge(current_group, interior_material, v3);
				
				e00 = new CFspinEdge(current_group, interior_material, v1);
				e01 = new CFspinEdge(current_group, interior_material, v0);
				e02 = new CFspinEdge(current_group, interior_material, v7);
				e03 = new CFspinEdge(current_group, interior_material, v6);
				
				e10 = new CFspinEdge(current_group, interior_material, v5);
				e11 = new CFspinEdge(current_group, interior_material, v2);
				e12 = new CFspinEdge(current_group, interior_material, v1);
				e13 = new CFspinEdge(current_group, interior_material, v6);
				
				e20 = new CFspinEdge(current_group, interior_material, v5);
				e21 = new CFspinEdge(current_group, interior_material, v4);
				e22 = new CFspinEdge(current_group, interior_material, v3);
				e23 = new CFspinEdge(current_group, interior_material, v2);

				e30 = new CFspinEdge(current_group, interior_material, v3);
				e31 = new CFspinEdge(current_group, interior_material, v4);
				e32 = new CFspinEdge(current_group, interior_material, v7);
				e33 = new CFspinEdge(current_group, interior_material, v0);

				e4 = new CFspinEdge(current_group, interior_material, v4);
				e5 = new CFspinEdge(current_group, interior_material, v5);
				e6 = new CFspinEdge(current_group, interior_material, v6);
				e7 = new CFspinEdge(current_group, interior_material, v7);

				// preliminary next pointer (probably final pointer)
				e0->next = e3;
				e1->next = e0;
				e2->next = e1;
				e3->next = e2;
				
				e00->next = e03;
				e01->next = e00;
				e02->next = e01;
				e03->next = e02;
				
				e10->next = e13;
				e11->next = e10;
				e12->next = e11;
				e13->next = e12;
				
				e20->next = e23;
				e21->next = e20;
				e22->next = e21;
				e23->next = e22;
				
				e30->next = e33;
				e31->next = e30;
				e32->next = e31;
				e33->next = e32;
				
				e4->next = e7;
				e5->next = e4;
				e6->next = e5;
				e7->next = e6;

				// preliminary spin pointer
				e0->spin = e00;
				e1->spin = e11;
				e2->spin = e22;
				e3->spin = e33;
				
				e00->spin = e0;
				e01->spin = e32;
				e02->spin = e6;
				e03->spin = e12;
				
				e10->spin = e23;
				e11->spin = e1;
				e12->spin = e03;
				e13->spin = e5;
				
				e20->spin = e4;
				e21->spin = e30;
				e22->spin = e2;
				e23->spin = e10;

				e30->spin = e21;
				e31->spin = e7;
				e32->spin = e01;
				e33->spin = e3;

				e4->spin = e20;
				e5->spin = e13;
				e6->spin = e02;
				e7->spin = e31;

				mesh->es->addElement(e0);
				mesh->es->addElement(e1);
				mesh->es->addElement(e2);
				mesh->es->addElement(e3);

				mesh->es->addElement(e00);
				mesh->es->addElement(e01);
				mesh->es->addElement(e02);
				mesh->es->addElement(e03);

				mesh->es->addElement(e10);
				mesh->es->addElement(e11);
				mesh->es->addElement(e12);
				mesh->es->addElement(e13);

				mesh->es->addElement(e20);
				mesh->es->addElement(e21);
				mesh->es->addElement(e22);
				mesh->es->addElement(e23);

				mesh->es->addElement(e30);
				mesh->es->addElement(e31);
				mesh->es->addElement(e32);
				mesh->es->addElement(e33);

				mesh->es->addElement(e4);
				mesh->es->addElement(e5);
				mesh->es->addElement(e6);
				mesh->es->addElement(e7);

				mesh->fs->addElement(e0);
				mesh->fs->addElement(e00);
				mesh->fs->addElement(e11);
				mesh->fs->addElement(e22);
				mesh->fs->addElement(e33);
				mesh->fs->addElement(e4);
				
				mesh->hs->addElement(e0);
			}
		}
  }
		
	cout << "# of pos_vol     : " << pos_vol << endl;
	cout << "# of neg_vol     : " << neg_vol << endl;

	cout << "# of vertices     : " << mesh->vs->size() << endl;
	cout << "# of hexahedrons  : " << mesh->hs->size() << endl;
	cout << "# of normals      : " << mesh->ns->size() << endl;
	cout << "# of groups       : " << mesh->gs->size() << endl;
	cout << "# of i-spinedges  : " << mesh->es->size() << endl;
	cout << "# of i-faces      : " << mesh->fs->size() << endl;

	mesh->interior_edges = mesh->es->size();
	mesh->interior_faces = mesh->fs->size();
	
	// set key for all edges so we can recover connectivity with quicksort
	// (NOTE: assuming vertex index is < 1048576)
	// (NOTE: maximal 21 bits for vertex index)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		v0_idx = e0->vertex->index;
		v1_idx = e0->next->next->next->vertex->index;
		v2_idx = e0->next->next->vertex->index;
		v3_idx = e0->next->vertex->index;
		
		if (v0_idx > v1_idx)
		{
			if (v2_idx > v3_idx)
			{
				e0->marker = (v2_idx << 42) | (v0_idx << 21) | (v1_idx);
			}
			else
			{
				e0->marker = (v3_idx << 42) | (v0_idx << 21) | (v1_idx);
			}
		}
		else
		{
			if (v2_idx > v3_idx)
			{
				e0->marker = (v2_idx << 42) | (v1_idx << 21) | (v0_idx);
			}
			else
			{
				e0->marker = (v3_idx << 42) | (v1_idx << 21) | (v0_idx);
			}
		}
	}

	quicksort_for_connectivity( mesh->es->data, 0, mesh->interior_edges-1 );
			
	// set inv pointer for all edges (and create boundary edges where needed)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		e1 = (CFspinEdge*)mesh->es->elementAt(i+1);
		e2 = (CFspinEdge*)mesh->es->elementAt(i+2);
		
		if (e0->inv == 0) // if we did not set the inv pointer yet
		{
			if ( e1 && (e0->marker == e1->marker) )
			{
				if ( e2 && (e1->marker == e2->marker) ) // non-manifold situation
				{
					cout << "non-manifold situation" << endl;
					return 0;
				}
				else
				{
					if ( e0->vertex == e1->vertex ) // non-orientated situation
					{
						cout << "non-orientated situation" << endl;
						return 0;
					}
					else
					{
						e0->inv = e1;
						e1->inv = e0;
						e0->next->inv = e1->next->next->next;
						e1->next->next->next->inv = e0->next;
						e0->next->next->inv = e1->next->next;
						e1->next->next->inv = e0->next->next;
						e0->next->next->next->inv = e1->next;
						e1->next->inv = e0->next->next->next;
						i++;
					}
				}
			}
			else
			{
				e00 = new CFspinEdge(-1, boundary_material, e0->next->next->next->vertex); // group index -1 means hole
				e01 = new CFspinEdge(-1, boundary_material, e0->vertex); // group index -1 means hole
				e02 = new CFspinEdge(-1, boundary_material, e0->next->vertex); // group index -1 means hole
				e03 = new CFspinEdge(-1, boundary_material, e0->next->next->vertex); // group index -1 means hole
				e00->next = e03;
				e01->next = e00;
				e02->next = e01;
				e03->next = e02;
				mesh->es->addElement(e00);
				mesh->es->addElement(e01);
				mesh->es->addElement(e02);
				mesh->es->addElement(e03);
				mesh->fs->addElement(e00);
				e0->inv = e00;
				e00->inv = e0;
				e0->next->inv = e01;
				e01->inv = e0->next;
				e0->next->next->inv = e02;
				e02->inv = e0->next->next;
				e0->next->next->next->inv = e03;
				e03->inv = e0->next->next->next;
			}
		}
	}
	
	mesh->total_edges = mesh->es->size();
	mesh->total_faces = mesh->fs->size();
	mesh->total_vertices = mesh->vs->size();

	cout << "# of b-spinedges  : " << (mesh->total_edges-mesh->interior_edges) << endl;
	cout << "# of b-faces      : " << (mesh->total_faces-mesh->interior_faces) << endl;
	
	// set spin pointer for all non-boundary spinedges (and set marker to zero)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		e0->spin = e0->spin->inv;
		e0->marker = 0;
	}
	
	// set spin pointer for all boundary spinedges
	for (i = mesh->interior_edges; i < mesh->total_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i); 		 
		e1 = e0->inv->spin;
		while (e1->group != -1)
		{
			e1 = e1->spin;
		}
		e0->spin = e1->inv;
	}

  return 1;
}

int CFloader::loadHex(char* file_name, CFmesh* mesh)
{
	FILE *file;
	char line[255];
	
	int i;
	float x, y, z;
	int v0_i, v1_i, v2_i, v3_i, v4_i, v5_i, v6_i, v7_i;
	__int64 v0_idx, v1_idx, v2_idx, v3_idx;
	CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
	CFspinEdge *e0, *e1, *e2, *e3;
	CFspinEdge *e00, *e01, *e02, *e03;
	CFspinEdge *e10, *e11, *e12, *e13;
	CFspinEdge *e20, *e21, *e22, *e23;
	CFspinEdge *e30, *e31, *e32, *e33;
	CFspinEdge *e4, *e5, *e6, *e7;

	int pos_vol = 0;
	int neg_vol = 0;
	int minus = 1;
	int format = -1; // 0 = CUBIT, 1 = GAMBIT, ...
	
	if (file = fopen(file_name, "r"))
	{
		cout << "opened file '" << file_name << "'" << endl;
	}
	else
	{
		cout << "cannot open file '" << file_name << "'" << endl;
		return 0;
	}
	
	// we use this group index for all faces in case the model has no group structure 	 
	char current_group_name[128];
	strcpy(current_group_name, "no group");
	int current_group = mesh->gs->size();
	mesh->gs->addElement((void*) new CFgroup(current_group_name));
	
	// we use this material index for all boundary in case the model has no material properties		
	char boundary_material_name[128];
	strcpy(boundary_material_name, "boundary");
	int boundary_material = mesh->ms->size();
	mesh->boundary_material = new CFmaterial(boundary_material_name, 0.0f, 0.0f, 1.0f);
	mesh->ms->addElement(mesh->boundary_material);

	// we use this material index for all interior in case the model has no material properties		
	char interior_material_name[128];
	strcpy(interior_material_name, "interior");
	int interior_material = mesh->ms->size();
	mesh->interior_material = new CFmaterial(interior_material_name);
	mesh->ms->addElement(mesh->interior_material);
	
	while (fgets(line, sizeof(char) * 255, file))
	{
		if (*line == 'v')
		{
			if (*(line+1) == ' ')
			{
				sscanf(line+2, "%f %f %f", &x, &y, &z);
				mesh->vs->addElement((void*) new CFvertex(x, y, z, mesh->vs->size()));
			}
			else
			{
				/* normals, textures, and else are not implemented */
			}
		}
		else if (*line == 'g')
		{
			if (sscanf(line+2, "%s", current_group_name) == 1) // prevent no-name groups
			{
				current_group = -1;
				for (i = 0; i < mesh->gs->size(); i++)
				{
					if (strcmp(&current_group_name[0], ((CFgroup*)mesh->gs->elementAt(i))->name) == 0)
					{
						current_group = i;
					}
				}
				if (current_group == -1)
				{
					current_group = mesh->gs->size();
					mesh->gs->addElement((void*) new CFgroup(current_group_name));
				}
			}
		}
		else if (*line == 'h')
		{
			if (*(line+1) == ' ') // read in a tetrahedron
			{
				if (format == 0)
				{
					sscanf(line+2, "%d %d %d %d %d %d %d %d", &v6_i, &v7_i, &v4_i, &v5_i, &v1_i, &v0_i, &v3_i, &v2_i); // CUBIT
				}
				else if (format == 1)
				{
					sscanf(line+2, "%d %d %d %d %d %d %d %d", &v7_i, &v6_i, &v0_i, &v1_i, &v4_i, &v5_i, &v3_i, &v2_i); // GAMBIT
				}
				else
				{
					sscanf(line+2, "%d %d %d %d %d %d %d %d", &v0_i, &v1_i, &v2_i, &v3_i, &v4_i, &v5_i, &v6_i, &v7_i);
					cout << "no format specified! " << endl;
				}

				v0 = (CFvertex*)mesh->vs->elementAt(v0_i - minus);
				v1 = (CFvertex*)mesh->vs->elementAt(v1_i - minus);
				v2 = (CFvertex*)mesh->vs->elementAt(v2_i - minus);
				v3 = (CFvertex*)mesh->vs->elementAt(v3_i - minus);

				v4 = (CFvertex*)mesh->vs->elementAt(v4_i - minus);
				v5 = (CFvertex*)mesh->vs->elementAt(v5_i - minus);
				v6 = (CFvertex*)mesh->vs->elementAt(v6_i - minus);
				v7 = (CFvertex*)mesh->vs->elementAt(v7_i - minus);

/*
				cout << v0->v[0] << " " << v0->v[1] << " " << v0->v[2] << endl;
				cout << v1->v[0] << " " << v1->v[1] << " " << v1->v[2] << endl;
				cout << v2->v[0] << " " << v2->v[1] << " " << v2->v[2] << endl;
				cout << v3->v[0] << " " << v3->v[1] << " " << v3->v[2] << endl;
				cout << v4->v[0] << " " << v4->v[1] << " " << v4->v[2] << endl;
				cout << v5->v[0] << " " << v5->v[1] << " " << v5->v[2] << endl;
				cout << v6->v[0] << " " << v6->v[1] << " " << v6->v[2] << endl;
				cout << v7->v[0] << " " << v7->v[1] << " " << v7->v[2] << endl;

/*
				float u1[3], u2[3], u3[3], u1u2[3];
				VecSubtract3fv(u1, v1->v, v0->v);
				VecSubtract3fv(u2, v2->v, v0->v);
				VecSubtract3fv(u3, v3->v, v0->v);
				VecCrossProd3fv(u1u2, u1, u2);
				if (VecDotProd3fv(u1u2, u3) > 0.0f)
				{
					pos_vol++;
				}
				else
				{
					neg_vol++;
					vhelp_i = v2_i;
					v2_i = v3_i;
					v3_i = vhelp_i;
				}
*/
				// create all spin edges
				e0 = new CFspinEdge(current_group, interior_material, v0);
				e1 = new CFspinEdge(current_group, interior_material, v1);
				e2 = new CFspinEdge(current_group, interior_material, v2);
				e3 = new CFspinEdge(current_group, interior_material, v3);
				
				e00 = new CFspinEdge(current_group, interior_material, v1);
				e01 = new CFspinEdge(current_group, interior_material, v0);
				e02 = new CFspinEdge(current_group, interior_material, v7);
				e03 = new CFspinEdge(current_group, interior_material, v6);
				
				e10 = new CFspinEdge(current_group, interior_material, v5);
				e11 = new CFspinEdge(current_group, interior_material, v2);
				e12 = new CFspinEdge(current_group, interior_material, v1);
				e13 = new CFspinEdge(current_group, interior_material, v6);
				
				e20 = new CFspinEdge(current_group, interior_material, v5);
				e21 = new CFspinEdge(current_group, interior_material, v4);
				e22 = new CFspinEdge(current_group, interior_material, v3);
				e23 = new CFspinEdge(current_group, interior_material, v2);

				e30 = new CFspinEdge(current_group, interior_material, v3);
				e31 = new CFspinEdge(current_group, interior_material, v4);
				e32 = new CFspinEdge(current_group, interior_material, v7);
				e33 = new CFspinEdge(current_group, interior_material, v0);

				e4 = new CFspinEdge(current_group, interior_material, v4);
				e5 = new CFspinEdge(current_group, interior_material, v5);
				e6 = new CFspinEdge(current_group, interior_material, v6);
				e7 = new CFspinEdge(current_group, interior_material, v7);

				// preliminary next pointer (probably final pointer)
				e0->next = e3;
				e1->next = e0;
				e2->next = e1;
				e3->next = e2;
				
				e00->next = e03;
				e01->next = e00;
				e02->next = e01;
				e03->next = e02;
				
				e10->next = e13;
				e11->next = e10;
				e12->next = e11;
				e13->next = e12;
				
				e20->next = e23;
				e21->next = e20;
				e22->next = e21;
				e23->next = e22;
				
				e30->next = e33;
				e31->next = e30;
				e32->next = e31;
				e33->next = e32;
				
				e4->next = e7;
				e5->next = e4;
				e6->next = e5;
				e7->next = e6;

				// preliminary spin pointer
				e0->spin = e00;
				e1->spin = e11;
				e2->spin = e22;
				e3->spin = e33;
				
				e00->spin = e0;
				e01->spin = e32;
				e02->spin = e6;
				e03->spin = e12;
				
				e10->spin = e23;
				e11->spin = e1;
				e12->spin = e03;
				e13->spin = e5;
				
				e20->spin = e4;
				e21->spin = e30;
				e22->spin = e2;
				e23->spin = e10;

				e30->spin = e21;
				e31->spin = e7;
				e32->spin = e01;
				e33->spin = e3;

				e4->spin = e20;
				e5->spin = e13;
				e6->spin = e02;
				e7->spin = e31;

				mesh->es->addElement(e0);
				mesh->es->addElement(e1);
				mesh->es->addElement(e2);
				mesh->es->addElement(e3);

				mesh->es->addElement(e00);
				mesh->es->addElement(e01);
				mesh->es->addElement(e02);
				mesh->es->addElement(e03);

				mesh->es->addElement(e10);
				mesh->es->addElement(e11);
				mesh->es->addElement(e12);
				mesh->es->addElement(e13);

				mesh->es->addElement(e20);
				mesh->es->addElement(e21);
				mesh->es->addElement(e22);
				mesh->es->addElement(e23);

				mesh->es->addElement(e30);
				mesh->es->addElement(e31);
				mesh->es->addElement(e32);
				mesh->es->addElement(e33);

				mesh->es->addElement(e4);
				mesh->es->addElement(e5);
				mesh->es->addElement(e6);
				mesh->es->addElement(e7);

				mesh->fs->addElement(e0);
				mesh->fs->addElement(e00);
				mesh->fs->addElement(e11);
				mesh->fs->addElement(e22);
				mesh->fs->addElement(e33);
				mesh->fs->addElement(e4);
				
				mesh->hs->addElement(e0);
			}
		}
		else if ((*line == 'u') && (*(line+1) == 's') && (*(line+2) == 'e'))
		{
			sscanf(line+6, "%s", interior_material_name);		
			interior_material = -1;
			for (i = 0; i < mesh->ms->size(); i++)
			{
				if (strcmp(interior_material_name, ((CFmaterial*)mesh->ms->elementAt(i))->name) == 0)
				{
					interior_material = i;
					break;
				}
			}
			if (interior_material == -1)
			{
				interior_material = mesh->ms->size()-1;
			}
		} 
		else if ((*line == 'm') && (*(line+1) == 'i') && (*(line+2) == 'n'))
		{
			sscanf(line+5, "%d", &minus);		
		}
		else if ((*line == 'f') && (*(line+1) == 'o') && (*(line+2) == 'r') && (*(line+3) == 'm') && (*(line+4) == 'a') && (*(line+5) == 't'))
		{
			char format_string[20];
			sscanf(line+6, "%s", format_string);
			if (strcmp(&format_string[0], "CUBIT") == 0)
			{
				format = 0;
			}
			else if (strcmp(&format_string[0], "GAMBIT") == 0)
			{
				format = 1;
			}
		}
	}
		
	cout << "# of pos_vol     : " << pos_vol << endl;
	cout << "# of neg_vol     : " << neg_vol << endl;

	cout << "# of vertices     : " << mesh->vs->size() << endl;
	cout << "# of hexahedrons  : " << mesh->hs->size() << endl;
	cout << "# of normals      : " << mesh->ns->size() << endl;
	cout << "# of groups       : " << mesh->gs->size() << endl;
	cout << "# of i-spinedges  : " << mesh->es->size() << endl;
	cout << "# of i-faces      : " << mesh->fs->size() << endl;

	mesh->interior_edges = mesh->es->size();
	mesh->interior_faces = mesh->fs->size();
	
	// set key for all edges so we can recover connectivity with quicksort
	// (NOTE: assuming vertex index is < 1048576)
	// (NOTE: maximal 21 bits for vertex index)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		v0_idx = e0->vertex->index;
		v1_idx = e0->next->next->next->vertex->index;
		v2_idx = e0->next->next->vertex->index;
		v3_idx = e0->next->vertex->index;
		
		if (v0_idx > v1_idx)
		{
			if (v2_idx > v3_idx)
			{
				e0->marker = (v2_idx << 42) | (v0_idx << 21) | (v1_idx);
			}
			else
			{
				e0->marker = (v3_idx << 42) | (v0_idx << 21) | (v1_idx);
			}
		}
		else
		{
			if (v2_idx > v3_idx)
			{
				e0->marker = (v2_idx << 42) | (v1_idx << 21) | (v0_idx);
			}
			else
			{
				e0->marker = (v3_idx << 42) | (v1_idx << 21) | (v0_idx);
			}
		}
	}

	quicksort_for_connectivity( mesh->es->data, 0, mesh->interior_edges-1 );
			
	// set inv pointer for all edges (and create boundary edges where needed)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		e1 = (CFspinEdge*)mesh->es->elementAt(i+1);
		e2 = (CFspinEdge*)mesh->es->elementAt(i+2);
		
		if (e0->inv == 0) // if we did not set the inv pointer yet
		{
			if ( e1 && (e0->marker == e1->marker) )
			{
				if ( e2 && (e1->marker == e2->marker) ) // non-manifold situation
				{
					cout << "non-manifold situation" << endl;
					return 0;
				}
				else
				{
					if ( e0->vertex == e1->vertex ) // non-orientated situation
					{
						cout << "non-orientated situation" << endl;
						return 0;
					}
					else
					{
						e0->inv = e1;
						e1->inv = e0;
						e0->next->inv = e1->next->next->next;
						e1->next->next->next->inv = e0->next;
						e0->next->next->inv = e1->next->next;
						e1->next->next->inv = e0->next->next;
						e0->next->next->next->inv = e1->next;
						e1->next->inv = e0->next->next->next;
						i++;
					}
				}
			}
			else
			{
				e00 = new CFspinEdge(-1, boundary_material, e0->next->next->next->vertex); // group index -1 means hole
				e01 = new CFspinEdge(-1, boundary_material, e0->vertex); // group index -1 means hole
				e02 = new CFspinEdge(-1, boundary_material, e0->next->vertex); // group index -1 means hole
				e03 = new CFspinEdge(-1, boundary_material, e0->next->next->vertex); // group index -1 means hole
				e00->next = e03;
				e01->next = e00;
				e02->next = e01;
				e03->next = e02;
				mesh->es->addElement(e00);
				mesh->es->addElement(e01);
				mesh->es->addElement(e02);
				mesh->es->addElement(e03);
				mesh->fs->addElement(e00);
				e0->inv = e00;
				e00->inv = e0;
				e0->next->inv = e01;
				e01->inv = e0->next;
				e0->next->next->inv = e02;
				e02->inv = e0->next->next;
				e0->next->next->next->inv = e03;
				e03->inv = e0->next->next->next;
			}
		}
	}
	
	mesh->total_edges = mesh->es->size();
	mesh->total_faces = mesh->fs->size();
	mesh->total_vertices = mesh->vs->size();

	cout << "# of b-spinedges  : " << (mesh->total_edges-mesh->interior_edges) << endl;
	cout << "# of b-faces      : " << (mesh->total_faces-mesh->interior_faces) << endl;
	
	// set spin pointer for all non-boundary spinedges (and set marker to zero)
	for (i = 0; i < mesh->interior_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i);
		e0->spin = e0->spin->inv;
		e0->marker = 0;
	}
	
	// set spin pointer for all boundary spinedges
	for (i = mesh->interior_edges; i < mesh->total_edges; i++)
	{
		e0 = (CFspinEdge*)mesh->es->elementAt(i); 		 
		e1 = e0->inv->spin;
		while (e1->group != -1)
		{
			e1 = e1->spin;
		}
		e0->spin = e1->inv;
	}
	return 1;
}
