/*
===============================================================================

	FILE:	main.cpp
	
	CONTENTS:
	
		Edge-Based Volume Mesh Compression
	
	PROGRAMMERS:
	
		martin isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		Copyright (C) 2002  martin isenburg (isenburg@cs.unc.edu)
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
		20 February 2002 -- created while working at INRIA

===============================================================================
*/
#include <iostream.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <GL/glut.h>

#include "loader.h"
#include "encoder.h"
#include "mesh.h"
#include "codec.h"
#include "arithmeticmodel.h"
#include "arithmeticencoder.h"
#include "arithmeticdecoder.h"
#include "positionquantizer.h"
#include "vertex.h"
#include "normal.h"
#include "spinedge.h"
#include "material.h"
#include "vector.h"
#include "vec3fv.h"

// MOUSE INTERFACE
int LeftButtonDown=0;        
int MiddleButtonDown=0;        
int RightButtonDown=0;
int OldX,OldY,NewX,NewY;
float Elevation=0;
float Azimuth=0;
float DistX=0;
float DistY=0;
float DistZ=3;

// GLOBAL CONTROL VARIABLES
int WindowW=1280, WindowH=960;
//int WindowW=800, WindowH=600;
//int WindowW=960, WindowH=720;
float NearPlane=1.0f, FarPlane=5.0f; 
int CodingAnimationOn=0;
int CodingAnimationSpeed=0;
int CodingAnimationDirection=1;
int CodingAnimationStep=265;
int BoundaryOn=1;
int InteriorOn=0;
int EdgeDegreesOn = 0;
int TetrahedraOn = 0;
int CodingOn = 0;
int WireFrameOn=0;
int VertexNormalsOn=0;
int InteractionMode=0;
int ColourMode=0;
int ViewMode=0;
int ViewModeGroup=1;
int ViewModeComponent=1;

// GLOBAL FACE FIXER VARIABLES
CFloader *loader = 0;
CFmesh *mesh = 0;
CFencoder *encoder = 0;
CFcodec *codec = 0;
PositionQuantizer *pq = 0;

int boundary_degrees = 6;
int has_boundary_degree[] = {1,1,1,1,1,1};
int interior_degrees = 6;
int has_interior_degree[6] = {1,1,1,1,1,1};

// GLOBAL MATERIAL VARIABLES
float colours_ambient[10][4];
float colours_diffuse[10][4];
float colours_specular[4][4];
float colours_emission[5][4];

// LIGHT
float LightFrom[3] = {7.0f, 15.0f, 15.0f};

void countMeshDegrees(CFmesh* mesh)
{
	int i;
	int degree;
	bool is_on_boundary;
	int num_faces = 0;
	int num_b_faces = 0;
	int num_i_faces = 0;
	int num_edge_degrees = 0;
	int num_b_edge_degrees = 0;
	int num_i_edge_degrees = 0;
	int sum_edge_degrees = 0;
	int sum_b_edge_degrees = 0;
	int sum_i_edge_degrees = 0;
	int b_edge_degrees[20];
	int i_edge_degrees[20];
	int num_vertex_degrees = 0;
	int num_b_vertex_degrees = 0;
	int num_i_vertex_degrees = 0;
	int sum_vertex_degrees = 0;
	int sum_b_vertex_degrees = 0;
	int sum_i_vertex_degrees = 0;
	int b_vertex_degrees[20];
	int i_vertex_degrees[20];
	CFspinEdge* e;
	CFvertex* v;

	for (i = 0; i < 20; i++)
	{
		b_edge_degrees[i] = 0;
		i_edge_degrees[i] = 0;
		b_vertex_degrees[i] = 0;
		i_vertex_degrees[i] = 0;
	}

	for (i = 0; i < boundary_degrees; i++)
	{
		has_boundary_degree[i] = 0;
	}

	for (i = 0; i < interior_degrees; i++)
	{
		has_interior_degree[i] = 0;
	}

	for (i = 0; i < mesh->es->size(); i++)
	{
		e = (CFspinEdge*) mesh->es->elementAt(i);
		e->marker = 1;
	}
	for (i = 0; i < mesh->vs->size(); i++)
	{
		v = (CFvertex*) mesh->vs->elementAt(i);
		v->marker = -1;
		v->degree = 0;
		v->active = 0;
	}

	for (i = 0; i < mesh->es->size(); i++)
	{
		e = (CFspinEdge*) mesh->es->elementAt(i);
	
		if (e->marker == 1)
		{
			num_faces++;

			e->marker = 0;
			e->next->marker = 0;
			e->next->next->marker = 0;
			e->inv->marker = 0;
			e->inv->next->marker = 0;
			e->inv->next->next->marker = 0;

			if (e->group == -1)
			{
				if (e->next->group != -1 || e->next->next->group != -1)
				{
					cout << "1 ";
				}
				num_b_faces++;
			}
			else if (e->inv->group == -1)
			{
				if (e->inv->next->group != -1 || e->inv->next->next->group != -1)
				{
					cout << "2 ";
				}
				num_b_faces++;
			}
			else
			{
				if (e->next->group == -1 || e->next->next->group == -1 || e->inv->next->group == -1 || e->inv->next->next->group == -1)
				{
					cout << "3 ";
				}
				num_i_faces++;
			}
		}
	}
	cout << endl;

	for (i = 0; i < mesh->es->size(); i++)
	{
		e = (CFspinEdge*) mesh->es->elementAt(i);
	
		if (e->marker == 1)
		{
			cout << "problem";
		}
	}

	for (i = 0; i < mesh->es->size(); i++)
	{
		e = (CFspinEdge*) mesh->es->elementAt(i);
	
		if (e->marker == 0)
		{
			e->vertex->marker++; // increase vertex degree
			e->inv->vertex->marker++; // increase vertex degree
			degree = 0;
			is_on_boundary = false;
			while (e->marker == 0)
			{
				degree++;
				e->marker = 1;
				e->inv->marker = 1;
				if (e->group == -1)
				{
					is_on_boundary = true;
					e->vertex->active = 1;
					e->inv->vertex->active = 1;
				}
				if ((e->vertex != e->spin->vertex) || (e->inv->vertex != e->inv->spin->vertex))
				{
					cout << "problem" << endl;
				}
				e = e->spin;
			}
			e->vertex->degree+=degree;
			e->inv->vertex->degree+=degree;			
			num_edge_degrees++;
			sum_edge_degrees+=degree;
			if (is_on_boundary)
			{
				num_b_edge_degrees++;
				sum_b_edge_degrees+=degree;
				if (degree > 19) degree = 19;
				b_edge_degrees[degree]++;
				degree = degree - 2;
				if (degree >= boundary_degrees) degree = boundary_degrees-1;
				has_boundary_degree[degree] = 1;
				while (e->marker == 1)
				{
					e->is_on_boundary = 1;
					e->inv->is_on_boundary = 1;
					e->marker = 2;
					e = e->spin;
				}
			}
			else
			{
				num_i_edge_degrees++;
				sum_i_edge_degrees+=degree;
				if (degree > 19) degree = 19;
				i_edge_degrees[degree]++;
				degree = degree - 2;
				if (degree >= interior_degrees) degree = interior_degrees-1;
				has_interior_degree[degree] = 1;
			}
		}
	}
	for (i = 0; i < mesh->vs->size(); i++)
	{
		v = (CFvertex*) mesh->vs->elementAt(i);
		degree = v->marker;
		num_vertex_degrees++;
		sum_vertex_degrees+=degree;
		if (v->active)
		{
			num_b_vertex_degrees++;
			sum_b_vertex_degrees+=degree;
			if (degree > 19) degree = 19;
			b_vertex_degrees[degree]++;
		}
		else
		{
			num_i_vertex_degrees++;
			sum_i_vertex_degrees+=degree;
			if (degree > 19) degree = 19;
			i_vertex_degrees[degree]++;
		}
	}
	cout << "faces: " << num_faces << " = b " << num_b_faces << " + i " << num_i_faces << endl;
	cout << "edge_degrees: " << num_edge_degrees << " = b " << num_b_edge_degrees << " + i " << num_i_edge_degrees << endl;
//	cout << "sum         : " << sum_edge_degrees << " = b " << sum_b_edge_degrees << " + i " << sum_i_edge_degrees << endl;
	cout << "b_edge_degrees: ";
	for (i = 0; i < 20; i++)
	{
		cout << b_edge_degrees[i] << " ";
	}
	cout << endl;
	cout << "i_edge_degrees: ";
	for (i = 0; i < 20; i++)
	{
		cout << i_edge_degrees[i] << " ";
	}
	cout << endl;
	cout << "vertex_degrees: " << num_vertex_degrees << " = b " << num_b_vertex_degrees << " + i " << num_i_vertex_degrees << endl;
	cout << "sum           : " << sum_vertex_degrees << " = b " << sum_b_vertex_degrees << " + i " << sum_i_vertex_degrees << endl;
	cout << "avg           : " << (float)sum_vertex_degrees/num_vertex_degrees << " = b " << (float)sum_b_vertex_degrees/num_b_vertex_degrees << " + i " << (float)sum_i_vertex_degrees/num_i_vertex_degrees << endl;
//	cout << "b_vertex_degrees: ";
	num_b_vertex_degrees = 0;
	for (i = 0; i < 20; i++)
	{
		num_b_vertex_degrees += b_vertex_degrees[i];
//		cout << b_vertex_degrees[i] << " ";
	}
//	cout << " = " << num_b_vertex_degrees << endl;
//	cout << "i_vertex_degrees: ";
	num_i_vertex_degrees = 0;
	for (i = 0; i < 20; i++)
	{
		num_i_vertex_degrees += i_vertex_degrees[i];
//		cout << i_vertex_degrees[i] << " ";
	}
//	cout << " = " << num_i_vertex_degrees << endl;

	// edge degree entropy
	float entropy_b_edge_degrees = 0.0f;
	float entropy_i_edge_degrees = 0.0f;
	for (i = 0; i < 20; i++)
	{
		if (b_edge_degrees[i] > 0)
		{
			entropy_b_edge_degrees -= float(log(float(b_edge_degrees[i])/float(num_b_edge_degrees)) * b_edge_degrees[i]);
		}
		if (i_edge_degrees[i] > 0)
		{
			entropy_i_edge_degrees -= float(log(float(i_edge_degrees[i])/float(num_i_edge_degrees)) * i_edge_degrees[i]);
		}
	}
	cout << "b_edge_degrees entropy: " << entropy_b_edge_degrees/float(log(2.0f)) <<  " i_edge_degrees entropy: " << entropy_i_edge_degrees/float(log(2.0f)) << endl;
	if (mesh->ts->size())
	{
		cout << "bits per tetrahedra   : " << ((entropy_b_edge_degrees+entropy_i_edge_degrees)/float(log(2.0f)))/mesh->ts->size() << endl;
	}
	else
	{
		cout << "bits per hexahedra    : " << ((entropy_b_edge_degrees+entropy_i_edge_degrees)/float(log(2.0f)))/mesh->hs->size() << endl;
	}
}

int encodeStep()
{
	if (codec == 0)
	{
		codec = new CFcodec();
		if (pq != 0)
		{
			delete pq;
		}
		pq = new PositionQuantizer();
		pq->SetPrecision(12);

    mesh->enQuantize(pq);
    mesh->deQuantize(pq);

		codec->initEnc(boundary_degrees, has_boundary_degree, interior_degrees, has_interior_degree, pq->m_uBits);
	}
	if (encoder->encodeStep(mesh, codec, pq) == 0)
	{
		int i;

		cout << "be: " << codec->be << " ie: " << codec->ie << endl;
		cout << "ib: " << codec->ib_off << " / " << codec->ib_on << endl;
		cout << "av: " << codec->av_off << " / " << codec->av_on << endl;
		cout << "case: " << encoder->case0 << " " << encoder->case1 << " " << encoder->case2a << " " << encoder->case2b << " " << encoder->case3a << " " << encoder->case3b << " " << encoder->case4 << endl;
		cout << "pred: null " << encoder->pred_null << " last " << encoder->pred_last << " ray " << encoder->pred_ray << " para " << encoder->pred_para << " lorenzo " << encoder->pred_lorenzo << endl;
		cout << "dist: null " << encoder->dist_null/encoder->pred_null << " last " << encoder->dist_last/encoder->pred_last << " ray " << encoder->dist_ray/encoder->pred_ray << " para " << encoder->dist_para/encoder->pred_para << " lorenzo " << encoder->dist_lorenzo/encoder->pred_lorenzo << endl;
		codec->doneEnc();
		cout << "bound: ";
		for (i = 0; i < boundary_degrees; i++)
		{
			cout << " " << codec->boundary_degree_dist[i];
		}
		cout << endl << "inter: ";
		for (i = 0; i < interior_degrees; i++)
		{
			cout << " " << codec->interior_degree_dist[i];
		}
		cout << endl;

		cout << "# conn bits: " << codec->ae_conn->getNumberBits() << " " << (float)codec->ae_conn->getNumberBits() / (float)mesh->hs->size() << endl;
		cout << "# geom bits: " << codec->ae_geom->getNumberBits() << " " << (float)codec->ae_geom->getNumberBits() / (float)mesh->vs->size() << endl;

    if (mesh->hs->size())
		{
      cout << "verticesHEXcount: " << encoder->verticesHEXcount << endl;
    }
    else
    {
      cout << "verticesTETcount: " << encoder->verticesTETcount << endl;
    }
		delete codec;
		codec = 0;
		return 0;
	}
	if (mesh->active)
	{
		CodingAnimationOn = 0;
	}
	return 1;
}

void displayMessage( )
{
  glColor3f( 0.7f, 0.7f, 0.7f );  // Set colour to grey
  glMatrixMode( GL_PROJECTION );
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D( 0.0, 1.0, 0.0, 1.0 );
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity();
  glRasterPos2f( 0.03f, 0.95f );
	
  if( InteractionMode == 0 )
	{
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'r');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'o');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'e');
	}
  else if( InteractionMode == 1 )
	{    
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'r');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'n');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 's');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'l');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'e');
	}
  else if( InteractionMode == 2 )
	{
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'z');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'o');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'o');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'm');
	}
  else if( InteractionMode == 3 )
	{
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'n');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'e');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'r');
	}
  else  if( InteractionMode == 4 )
	{
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'f');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'r');
	}
	
	glPopMatrix();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
}

void myTetDisplay()
{
	int i;
	glEnable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);  
	
	if (BoundaryOn)
	{
		CFmaterial* m = mesh->boundary_material;
		glMaterialfv(GL_FRONT,GL_AMBIENT,m->a);
		glMaterialfv(GL_FRONT,GL_DIFFUSE,m->d);
		glMaterialfv(GL_FRONT,GL_SPECULAR,m->s);
		glMaterialfv(GL_FRONT,GL_EMISSION,m->e);

		if (BoundaryOn ==	1 || BoundaryOn == 2) // draw polygons = 1 or lines = 2 or points = 3
		{
			CFspinEdge *e0, *e1, *e2;
			if (BoundaryOn == 1)
			{
				glPolygonMode(GL_FRONT, GL_FILL);
			}
			else
			{
				glPolygonMode(GL_FRONT, GL_LINE);
			}
			glBegin(GL_TRIANGLES);
			for (int i = mesh->interior_faces; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				e1 = e0->next;
				e2 = e1->next;
				glNormal3fv(e0->normal->n);
				glVertex3fv(e0->vertex->v);
				glVertex3fv(e2->vertex->v);
				glVertex3fv(e1->vertex->v);
			}
			glEnd();
		}
		else
		{
			CFvertex *v;
			CFspinEdge *e;
			for  (i = mesh->interior_edges; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				e->vertex->marker = 1;
			}
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->total_vertices; i++)
			{
				v = (CFvertex*) mesh->vs->elementAt(i);
				if (v->marker)
				{
					#define DRAWVERTEX(x,y,z,s) glNormal3f( 1,0,0); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y+s,z-s); glNormal3f(-1,0,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y+s,z-s); glVertex3f(x-s,y+s,z+s); glVertex3f(x-s,y-s,z+s); glNormal3f(0, 1,0); glVertex3f(x-s,y+s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x+s,y+s,z+s); glVertex3f(x-s,y+s,z+s); glNormal3f(0,-1,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y-s,z+s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y-s,z-s); glNormal3f(0,0, 1); glVertex3f(x-s,y-s,z+s); glVertex3f(x-s,y+s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y-s,z+s); glNormal3f(0,0,-1); glVertex3f(x-s,y-s,z-s); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x-s,y+s,z-s);
					DRAWVERTEX(v->v[0], v->v[1],v->v[2], 0.005f);
					#undef DRAWVERTEX
					v->marker = 0;
				}
			}
			glEnd();
		}
	}
	
	if (InteriorOn)
	{
		CFmaterial* m = mesh->interior_material;
		glMaterialfv(GL_FRONT,GL_AMBIENT,m->a);
		glMaterialfv(GL_FRONT,GL_DIFFUSE,m->d);
		glMaterialfv(GL_FRONT,GL_SPECULAR,m->s);
		glMaterialfv(GL_FRONT,GL_EMISSION,m->e);

		if (InteriorOn ==	1 || InteriorOn == 2) // draw polygons = 1 or lines = 2 or points = 3
		{
			CFspinEdge *e0, *e1, *e2;
			if (InteriorOn == 1)
			{
				glPolygonMode(GL_FRONT, GL_FILL);
			}
			else
			{
				glPolygonMode(GL_FRONT, GL_LINE);
			}
			glBegin(GL_TRIANGLES);
			for (int i = 0; i < mesh->interior_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				e1 = e0->next;
				e2 = e1->next;
				glNormal3fv(e0->normal->n);
				glVertex3fv(e0->vertex->v);
				glVertex3fv(e2->vertex->v);
				glVertex3fv(e1->vertex->v);
			}
			glEnd();
		}
		else
		{
			CFvertex *v;
			CFspinEdge *e;
			for  (i = mesh->interior_edges; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				e->vertex->marker = 1;
			}
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->total_vertices; i++)
			{
				v = (CFvertex*) mesh->vs->elementAt(i);
				if (v->marker == 0)
				{
					#define DRAWVERTEX(x,y,z,s) glNormal3f( 1,0,0); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y+s,z-s); glNormal3f(-1,0,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y+s,z-s); glVertex3f(x-s,y+s,z+s); glVertex3f(x-s,y-s,z+s); glNormal3f(0, 1,0); glVertex3f(x-s,y+s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x+s,y+s,z+s); glVertex3f(x-s,y+s,z+s); glNormal3f(0,-1,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y-s,z+s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y-s,z-s); glNormal3f(0,0, 1); glVertex3f(x-s,y-s,z+s); glVertex3f(x-s,y+s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y-s,z+s); glNormal3f(0,0,-1); glVertex3f(x-s,y-s,z-s); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x-s,y+s,z-s);
					DRAWVERTEX(v->v[0], v->v[1],v->v[2], 0.0025f);
					#undef DRAWVERTEX
				}
				else
				{
					v->marker = 0;
				}
			}
			glEnd();
		}
	}

	if ( CodingOn )
	{
		// show next gate
		if (1)
		{
			CFspinEdge *e0, *e1, *e2;
			float temp1[3] = { 1.0f, 0.0f, 1.0f };
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_TRIANGLES);
			if (mesh->gate)
			{
				e0 = mesh->gate;
				e1 = e0->next;
				e2 = e1->next;
				glNormal3fv(e0->normal->n);
				glVertex3fv(e0->vertex->v);
				glVertex3fv(e2->vertex->v);
				glVertex3fv(e1->vertex->v);
			}
			glEnd();
		}
		// show coding boundary (faces)
		if (1)
		{
			CFspinEdge *e0, *e1, *e2;
			float temp1[3] = { 0.0f, 0.0f, 1.0f };
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_TRIANGLES);
			for (int i = 0; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				if (e0->inv->component && !e0->component)
				{
					e0 = e0;
					e1 = e0->next;
					e2 = e1->next;
					glNormal3fv(e0->normal->n);
					glVertex3fv(e0->vertex->v);
					glVertex3fv(e2->vertex->v);
					glVertex3fv(e1->vertex->v);
				}
			}
			glEnd();
		}
		// show coding boundary (edges)
		if (1)
		{
			CFspinEdge *e0, *e1, *e2;
			float temp1[3] = { 0.0f, 0.0f, 0.0f };
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
			glBegin(GL_LINES);
			for (int i = 0; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				if (e0->inv->component && !e0->component)
				{
					e0 = e0;
					e1 = e0->next;
					e2 = e1->next;
					glNormal3fv(e0->normal->n);
					glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
					glVertex3f(e0->next->vertex->v[0]+e0->next->normal->n[0]*0.001,e0->next->vertex->v[1]+e0->next->normal->n[1]*0.001,e0->next->vertex->v[2]+e0->next->normal->n[2]*0.001);
					glVertex3f(e0->next->vertex->v[0]+e0->next->normal->n[0]*0.001,e0->next->vertex->v[1]+e0->next->normal->n[1]*0.001,e0->next->vertex->v[2]+e0->next->normal->n[2]*0.001);
					glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
					glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
					glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
				}
			}
			glEnd();
		}
		// show edge degrees
		if (1)
		{
			CFspinEdge *e;
			float temp1[3] = { 1.0f, 0.0f, 0.0f }, temp2[3], temp3[3];
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);

			for  (i = encoder->verticesTETdone; i < encoder->verticesTETnumber; i++)
			{
				e = encoder->verticesTET[i];
				while (e)
				{
					if (e->slots > 0)
					{
						int slots = e->slots;
						CFspinEdge *e_run = e->inv;
						while (slots--)
						{
							e_run = e_run->spin;

							glNormal3fv(e_run->normal->n);

							VecSubtract3fv(temp1, e_run->inv->vertex->v, e_run->vertex->v);
							VecSelfScalarMult3fv(temp1, 0.3f);
							VecAdd3fv(temp2, e_run->vertex->v, temp1);
							glVertex3fv(temp2);
							
							VecSubtract3fv(temp1, e_run->inv->vertex->v, e_run->vertex->v);
							VecSelfScalarMult3fv(temp1, 0.7f);
							VecAdd3fv(temp3, e_run->vertex->v, temp1);
							glVertex3fv(temp3);

							VecSubtract3fv(temp1, e_run->next->vertex->v, temp3);
							VecSelfScalarMult3fv(temp1, 0.2f);
							VecSelfAdd3fv(temp3, temp1);
							glVertex3fv(temp3);

							VecSubtract3fv(temp1, e_run->next->vertex->v, temp2);
							VecSelfScalarMult3fv(temp1, 0.2f);
							VecSelfAdd3fv(temp2, temp1);
							glVertex3fv(temp2);
						}
					}
					e = e->Lnext;
				}
			}
			glEnd();
		}
/*
			for  (i = 0; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				if (e->vertex->active && e->inv->vertex->active && !e->component  && !e->inv->component)
				{
					glNormal3fv(e->normal->n);

					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.3f);
					VecAdd3fv(temp2, e->vertex->v, temp1);
					glVertex3fv(temp2);
					
					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.7f);
					VecAdd3fv(temp3, e->vertex->v, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->vertex->v, temp3);
					VecSelfScalarMult3fv(temp1, 0.2f);
					VecSelfAdd3fv(temp3, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->vertex->v, temp2);
					VecSelfScalarMult3fv(temp1, 0.2f);
					VecSelfAdd3fv(temp2, temp1);
					glVertex3fv(temp2);
				}
			}
			glEnd();
		}
*/
		// show neighboring tetrahedrons
		if (1)
		{
			CFspinEdge *e;
			CFvertex *v0, *v1, *v2, *v3;
			float temp0[3] = { 0.0f, 1.0f, 0.0f }, temp1[3], temp2[3], temp3[3];
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp0);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_TRIANGLES);
			for  (i = 0; i < mesh->ts->size(); i++)
			{
				e = (CFspinEdge*) mesh->ts->elementAt(i);

				if ((!e->component) && (e->inv->component || e->spin->component || e->next->spin->component || e->next->next->spin->component))
				{
					v0 = e->vertex;
					v1 = e->next->next->vertex;
					v2 = e->next->vertex;
					v3 = e->spin->inv->next->vertex;

					VecAdd3fv(temp0, v1->v, v2->v);
					VecSelfAdd3fv(temp0, v3->v);
					VecSelfScalarDiv3fv(temp0, 3.0f);
					VecSelfSubtract3fv(temp0, v0->v);
					VecSelfScalarMult3fv(temp0, 0.4f);
					VecSelfAdd3fv(temp0, v0->v);

					VecAdd3fv(temp1, v0->v, v2->v);
					VecSelfAdd3fv(temp1, v3->v);
					VecSelfScalarDiv3fv(temp1, 3.0f);
					VecSelfSubtract3fv(temp1, v1->v);
					VecSelfScalarMult3fv(temp1, 0.4f);
					VecSelfAdd3fv(temp1, v1->v);
					
					VecAdd3fv(temp2, v0->v, v1->v);
					VecSelfAdd3fv(temp2, v3->v);
					VecSelfScalarDiv3fv(temp2, 3.0f);
					VecSelfSubtract3fv(temp2, v2->v);
					VecSelfScalarMult3fv(temp2, 0.4f);
					VecSelfAdd3fv(temp2, v2->v);

					VecAdd3fv(temp3, v0->v, v1->v);
					VecSelfAdd3fv(temp3, v2->v);
					VecSelfScalarDiv3fv(temp3, 3.0f);
					VecSelfSubtract3fv(temp3, v3->v);
					VecSelfScalarMult3fv(temp3, 0.4f);
					VecSelfAdd3fv(temp3, v3->v);

					glNormal3fv(e->inv->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp2);
					glVertex3fv(temp1);

					glNormal3fv(e->spin->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp1);
					glVertex3fv(temp3);

					glNormal3fv(e->next->spin->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp3);
					glVertex3fv(temp2);

					glNormal3fv(e->next->next->spin->normal->n);
					glVertex3fv(temp1);
					glVertex3fv(temp2);
					glVertex3fv(temp3);
				}
			}
			glEnd();
		}
	}

	if ( EdgeDegreesOn )
	{
		CFspinEdge *e;
		float temp1[3] = { 1.0f, 0.0f, 0.0f };
		float temp2[3];
		float temp3[3];
		glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
		glPolygonMode(GL_FRONT, GL_FILL);
		glBegin(GL_QUADS);
		if ( BoundaryOn )
		{
			for  (i = 0; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				if (e->is_on_boundary)
				{
					glNormal3fv(e->normal->n);

					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.3f);
					VecAdd3fv(temp2, e->vertex->v, temp1);
					glVertex3fv(temp2);
					
					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.7f);
					VecAdd3fv(temp3, e->vertex->v, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->vertex->v, temp3);
					VecSelfScalarMult3fv(temp1, 0.2f);
					VecSelfAdd3fv(temp3, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->vertex->v, temp2);
					VecSelfScalarMult3fv(temp1, 0.2f);
					VecSelfAdd3fv(temp2, temp1);
					glVertex3fv(temp2);
				}
			}
		}
		if ( InteriorOn )
		{
			for  (i = 0; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				if (!e->is_on_boundary)
				{
					glNormal3fv(e->normal->n);

					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.3f);
					VecAdd3fv(temp2, e->vertex->v, temp1);
					glVertex3fv(temp2);
					
					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.7f);
					VecAdd3fv(temp3, e->vertex->v, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->vertex->v, temp3);
					VecSelfScalarMult3fv(temp1, 0.2f);
					VecSelfAdd3fv(temp3, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->vertex->v, temp2);
					VecSelfScalarMult3fv(temp1, 0.2f);
					VecSelfAdd3fv(temp2, temp1);
					glVertex3fv(temp2);
				}
			}
		}
		glEnd();
	}

	if ( TetrahedraOn )
	{
		CFspinEdge *e;
		CFvertex *v0, *v1, *v2, *v3;
		float temp0[3] = { 0.0f, 1.0f, 0.0f }, temp1[3], temp2[3], temp3[3];
		glMaterialfv(GL_FRONT,GL_DIFFUSE,temp0);
		glPolygonMode(GL_FRONT, GL_FILL);
		glBegin(GL_TRIANGLES);
		for  (i = 0; i < mesh->ts->size(); i++)
		{
			e = (CFspinEdge*) mesh->ts->elementAt(i);

			v0 = e->vertex;
			v1 = e->next->next->vertex;
			v2 = e->next->vertex;
			v3 = e->spin->inv->next->vertex;

			VecAdd3fv(temp0, v1->v, v2->v);
			VecSelfAdd3fv(temp0, v3->v);
			VecSelfScalarDiv3fv(temp0, 3.0f);
			VecSelfSubtract3fv(temp0, v0->v);
			VecSelfScalarMult3fv(temp0, 0.4f);
			VecSelfAdd3fv(temp0, v0->v);

			VecAdd3fv(temp1, v0->v, v2->v);
			VecSelfAdd3fv(temp1, v3->v);
			VecSelfScalarDiv3fv(temp1, 3.0f);
			VecSelfSubtract3fv(temp1, v1->v);
			VecSelfScalarMult3fv(temp1, 0.4f);
			VecSelfAdd3fv(temp1, v1->v);
			
			VecAdd3fv(temp2, v0->v, v1->v);
			VecSelfAdd3fv(temp2, v3->v);
			VecSelfScalarDiv3fv(temp2, 3.0f);
			VecSelfSubtract3fv(temp2, v2->v);
			VecSelfScalarMult3fv(temp2, 0.4f);
			VecSelfAdd3fv(temp2, v2->v);

			VecAdd3fv(temp3, v0->v, v1->v);
			VecSelfAdd3fv(temp3, v2->v);
			VecSelfScalarDiv3fv(temp3, 3.0f);
			VecSelfSubtract3fv(temp3, v3->v);
			VecSelfScalarMult3fv(temp3, 0.4f);
			VecSelfAdd3fv(temp3, v3->v);

			glNormal3fv(e->inv->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp2);
			glVertex3fv(temp1);

			glNormal3fv(e->spin->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp1);
			glVertex3fv(temp3);

			glNormal3fv(e->next->spin->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp3);
			glVertex3fv(temp2);

			glNormal3fv(e->next->next->spin->normal->n);
			glVertex3fv(temp1);
			glVertex3fv(temp2);
			glVertex3fv(temp3);
		}
		glEnd();
	}

	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);  
	
	if ( VertexNormalsOn )
	{
		CFspinEdge *e;
		glColor3f(1.0f, 1.0f, 1.0f);
		glLineWidth(1.0f);
		glBegin(GL_LINES);
		if ( BoundaryOn )
		{
			for  (i = mesh->interior_edges; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				glVertex3fv(e->vertex->v);
				glVertex3f(e->vertex->v[0]+e->normal->n[0]*0.05,e->vertex->v[1]+e->normal->n[1]*0.05,e->vertex->v[2]+e->normal->n[2]*0.05);
			}
		}
		if ( InteriorOn )
		{
			for  (i = 0; i < mesh->interior_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				glVertex3fv(e->vertex->v);
				glVertex3f(e->vertex->v[0]+e->normal->n[0]*0.05,e->vertex->v[1]+e->normal->n[1]*0.05,e->vertex->v[2]+e->normal->n[2]*0.05);
			}
		}
		glEnd();
	}
	
	if ( WireFrameOn )
	{
		CFspinEdge *e0, *e1, *e2;
		glLineWidth(1.0f);
		glBegin(GL_LINES);
		if ( BoundaryOn )
		{
			if ( BoundaryOn == 3 )
			{
				glColor3f(1.0f, 1.0f, 1.0f);
			}
			else
			{
				glColor3f(0.0f, 0.0f, 0.0f);
			}

			for  (i = mesh->interior_faces; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				e1 = e0->next;
				e2 = e1->next;
				glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
				glVertex3f(e0->next->vertex->v[0]+e0->next->normal->n[0]*0.001,e0->next->vertex->v[1]+e0->next->normal->n[1]*0.001,e0->next->vertex->v[2]+e0->next->normal->n[2]*0.001);
				glVertex3f(e0->next->vertex->v[0]+e0->next->normal->n[0]*0.001,e0->next->vertex->v[1]+e0->next->normal->n[1]*0.001,e0->next->vertex->v[2]+e0->next->normal->n[2]*0.001);
				glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
				glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
				glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
			}
		}
		if ( InteriorOn )
		{
			if ( InteriorOn == 3 )
			{
				glColor3f(1.0f, 1.0f, 1.0f);
			}
			else
			{
				glColor3f(0.0f, 0.0f, 0.0f);
			}

			for  (i = 0; i < mesh->interior_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				e1 = e0->next;
				e2 = e1->next;
				glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
				glVertex3f(e0->next->vertex->v[0]+e0->next->normal->n[0]*0.001,e0->next->vertex->v[1]+e0->next->normal->n[1]*0.001,e0->next->vertex->v[2]+e0->next->normal->n[2]*0.001);
				glVertex3f(e0->next->vertex->v[0]+e0->next->normal->n[0]*0.001,e0->next->vertex->v[1]+e0->next->normal->n[1]*0.001,e0->next->vertex->v[2]+e0->next->normal->n[2]*0.001);
				glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
				glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
				glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
			}
		}
		glEnd();
	}
	
	if (mesh->active)
	{
		glColor3f(0.5f, 0.5f, 0.9f);
		glLineWidth(6.0);
		glBegin(GL_LINES);
		glVertex3fv(mesh->active->vertex->v);
		glVertex3fv(mesh->active->inv->vertex->v);
		glEnd();
	}
}

void myHexDisplay()
{
	int i;

	glEnable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);  

	if (BoundaryOn)
	{
		CFmaterial* m = mesh->boundary_material;
		glMaterialfv(GL_FRONT,GL_AMBIENT,m->a);
		glMaterialfv(GL_FRONT,GL_DIFFUSE,m->d);
		glMaterialfv(GL_FRONT,GL_SPECULAR,m->s);
		glMaterialfv(GL_FRONT,GL_EMISSION,m->e);

		if (BoundaryOn ==	1 || BoundaryOn == 2) // draw polygons = 1 or lines = 2 or points = 3
		{
			CFspinEdge *e0, *e1, *e2, *e3;
			if (BoundaryOn == 1)
			{
				glPolygonMode(GL_FRONT, GL_FILL);
			}
			else
			{
				glPolygonMode(GL_FRONT, GL_LINE);
			}
			glBegin(GL_QUADS);
			for (int i = mesh->interior_faces; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				e1 = e0->next;
				e2 = e1->next;
				e3 = e2->next;
				glNormal3fv(e0->normal->n);
				glVertex3fv(e3->vertex->v);
				glVertex3fv(e2->vertex->v);
				glVertex3fv(e1->vertex->v);
				glVertex3fv(e0->vertex->v);
			}
			glEnd();
		}
		else
		{
			CFvertex *v;
			CFspinEdge *e;
			for  (i = mesh->interior_edges; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				e->vertex->marker = 1;
			}
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->total_vertices; i++)
			{
				v = (CFvertex*) mesh->vs->elementAt(i);
				if (v->marker)
				{
					#define DRAWVERTEX(x,y,z,s) glNormal3f( 1,0,0); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y+s,z-s); glNormal3f(-1,0,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y+s,z-s); glVertex3f(x-s,y+s,z+s); glVertex3f(x-s,y-s,z+s); glNormal3f(0, 1,0); glVertex3f(x-s,y+s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x+s,y+s,z+s); glVertex3f(x-s,y+s,z+s); glNormal3f(0,-1,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y-s,z+s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y-s,z-s); glNormal3f(0,0, 1); glVertex3f(x-s,y-s,z+s); glVertex3f(x-s,y+s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y-s,z+s); glNormal3f(0,0,-1); glVertex3f(x-s,y-s,z-s); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x-s,y+s,z-s);
					DRAWVERTEX(v->v[0], v->v[1],v->v[2], 0.005f);
					#undef DRAWVERTEX
					v->marker = 0;
				}
			}
			glEnd();
		}
	}
	if (InteriorOn)
	{
		CFmaterial* m = mesh->interior_material;
		glMaterialfv(GL_FRONT,GL_AMBIENT,m->a);
		glMaterialfv(GL_FRONT,GL_DIFFUSE,m->d);
		glMaterialfv(GL_FRONT,GL_SPECULAR,m->s);
		glMaterialfv(GL_FRONT,GL_EMISSION,m->e);

		if (InteriorOn ==	1 || InteriorOn == 2) // draw polygons = 1 or lines = 2 or points = 3
		{
			CFspinEdge *e0, *e1, *e2, *e3;
			if (InteriorOn == 1)
			{
				glPolygonMode(GL_FRONT, GL_FILL);
			}
			else
			{
				glPolygonMode(GL_FRONT, GL_LINE);
			}
			glBegin(GL_QUADS);
			for (int i = 0; i < mesh->interior_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				e1 = e0->next;
				e2 = e1->next;
				e3 = e2->next;
				glNormal3fv(e0->normal->n);
				glVertex3fv(e3->vertex->v);
				glVertex3fv(e2->vertex->v);
				glVertex3fv(e1->vertex->v);
				glVertex3fv(e0->vertex->v);
			}
			glEnd();
		}
		else
		{
			CFvertex *v;
			CFspinEdge *e;
			for  (i = mesh->interior_edges; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				e->vertex->marker = 1;
			}
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->total_vertices; i++)
			{
				v = (CFvertex*) mesh->vs->elementAt(i);
				if (v->marker == 0)
				{
					#define DRAWVERTEX(x,y,z,s) glNormal3f( 1,0,0); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y+s,z-s); glNormal3f(-1,0,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y+s,z-s); glVertex3f(x-s,y+s,z+s); glVertex3f(x-s,y-s,z+s); glNormal3f(0, 1,0); glVertex3f(x-s,y+s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x+s,y+s,z+s); glVertex3f(x-s,y+s,z+s); glNormal3f(0,-1,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y-s,z+s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y-s,z-s); glNormal3f(0,0, 1); glVertex3f(x-s,y-s,z+s); glVertex3f(x-s,y+s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y-s,z+s); glNormal3f(0,0,-1); glVertex3f(x-s,y-s,z-s); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x-s,y+s,z-s);
					DRAWVERTEX(v->v[0], v->v[1],v->v[2], 0.0025f);
					#undef DRAWVERTEX
				}
				else
				{
					v->marker = 0;
				}
			}
			glEnd();
		}
	}
	if ( TetrahedraOn )
	{
		CFspinEdge *e;
		CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
		float temp[3] = { 0.0f, 1.0f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
		glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
		glPolygonMode(GL_FRONT, GL_FILL);
		glBegin(GL_QUADS);
		for  (i = 0; i < mesh->hs->size(); i++)
		{
			e = (CFspinEdge*) mesh->hs->elementAt(i);

			v0 = e->vertex;
			v1 = e->next->next->next->vertex;
			v2 = e->next->next->vertex;
			v3 = e->next->vertex;
			v4 = e->next->next->spin->inv->next->vertex;
			v5 = e->next->next->spin->inv->next->next->vertex;
			v6 = e->spin->inv->next->vertex;
			v7 = e->spin->inv->next->next->vertex;

			VecAdd3fv(temp, v0->v, v1->v);
			VecSelfAdd3fv(temp, v2->v);
			VecSelfAdd3fv(temp, v3->v);
			VecSelfAdd3fv(temp, v4->v);
			VecSelfAdd3fv(temp, v5->v);
			VecSelfAdd3fv(temp, v6->v);
			VecSelfAdd3fv(temp, v7->v);
			VecSelfScalarDiv3fv(temp, 8.0f);

			VecSubtract3fv(temp0, temp, v0->v);
			VecSelfScalarMult3fv(temp0, 0.3f);
			VecSelfAdd3fv(temp0, v0->v);

			VecSubtract3fv(temp1, temp, v1->v);
			VecSelfScalarMult3fv(temp1, 0.3f);
			VecSelfAdd3fv(temp1, v1->v);

			VecSubtract3fv(temp2, temp, v2->v);
			VecSelfScalarMult3fv(temp2, 0.3f);
			VecSelfAdd3fv(temp2, v2->v);

			VecSubtract3fv(temp3, temp, v3->v);
			VecSelfScalarMult3fv(temp3, 0.3f);
			VecSelfAdd3fv(temp3, v3->v);

			VecSubtract3fv(temp4, temp, v4->v);
			VecSelfScalarMult3fv(temp4, 0.3f);
			VecSelfAdd3fv(temp4, v4->v);

			VecSubtract3fv(temp5, temp, v5->v);
			VecSelfScalarMult3fv(temp5, 0.3f);
			VecSelfAdd3fv(temp5, v5->v);

			VecSubtract3fv(temp6, temp, v6->v);
			VecSelfScalarMult3fv(temp6, 0.3f);
			VecSelfAdd3fv(temp6, v6->v);

			VecSubtract3fv(temp7, temp, v7->v);
			VecSelfScalarMult3fv(temp7, 0.3f);
			VecSelfAdd3fv(temp7, v7->v);

			glNormal3fv(e->inv->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp3);
			glVertex3fv(temp2);
			glVertex3fv(temp1);

			glNormal3fv(e->spin->normal->n);
			glVertex3fv(temp1);
			glVertex3fv(temp6);
			glVertex3fv(temp7);
			glVertex3fv(temp0);

			glNormal3fv(e->next->spin->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp7);
			glVertex3fv(temp4);
			glVertex3fv(temp3);

			glNormal3fv(e->next->next->spin->normal->n);
			glVertex3fv(temp3);
			glVertex3fv(temp4);
			glVertex3fv(temp5);
			glVertex3fv(temp2);

			glNormal3fv(e->next->next->next->spin->normal->n);
			glVertex3fv(temp2);
			glVertex3fv(temp5);
			glVertex3fv(temp6);
			glVertex3fv(temp1);

			glNormal3fv(e->spin->inv->next->next->spin->normal->n);
			glVertex3fv(temp6);
			glVertex3fv(temp5);
			glVertex3fv(temp4);
			glVertex3fv(temp7);
		}
		glEnd();
	}
	if ( CodingOn )
	{
		// show next gate
		if (1)
		{
			CFspinEdge *e0, *e1, *e2, *e3;
			float temp[3] = { 0.9f, 0.4f, 0.9f };
			float offset = 0.001f;
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			if (mesh->gate)
			{
				e0 = mesh->gate;
				e1 = e0->next;
				e2 = e1->next;
				e3 = e2->next;
				glNormal3fv(e0->normal->n);
				glVertex3f(e2->vertex->v[0]+offset*e0->normal->n[0],e2->vertex->v[1]+offset*e0->normal->n[1],e2->vertex->v[2]+offset*e0->normal->n[2]);
				glVertex3f(e1->vertex->v[0]+offset*e0->normal->n[0],e1->vertex->v[1]+offset*e0->normal->n[1],e1->vertex->v[2]+offset*e0->normal->n[2]);
				temp[1] = 0.0f;
				glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
				glVertex3f(e0->vertex->v[0]+offset*e0->normal->n[0],e0->vertex->v[1]+offset*e0->normal->n[1],e0->vertex->v[2]+offset*e0->normal->n[2]);
				glVertex3f(e3->vertex->v[0]+offset*e0->normal->n[0],e3->vertex->v[1]+offset*e0->normal->n[1],e3->vertex->v[2]+offset*e0->normal->n[2]);
			}
			glEnd();
		}
		// show coding boundary (faces)
		if (1)
		{
			CFspinEdge *e0, *e1, *e2, *e3;
			float temp1[3] = { 0.0f, 0.0f, 1.0f };
			float temp4[3] = { 0.0f, 0.0f, 2.0f };
			float temp2[3] = { 0.75f, 0.75f, 1.5f };
			float temp3[3] = { 0.1f, 0.1f, 0.2f };
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp1);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for (int i = 0; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				if (e0->inv->component && !e0->component)
				{
					if (e0->group == -1)
					{
						glMaterialfv(GL_FRONT,GL_AMBIENT,temp4);
						glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
					}
					else
					{
						glMaterialfv(GL_FRONT,GL_AMBIENT,temp2);
						glMaterialfv(GL_FRONT,GL_DIFFUSE,temp3);
					}
					e0 = e0;
					e1 = e0->next;
					e2 = e1->next;
					e3 = e2->next;
					glNormal3fv(e0->normal->n);
					glVertex3fv(e3->vertex->v);
					glVertex3fv(e2->vertex->v);
					glVertex3fv(e1->vertex->v);
					glVertex3fv(e0->vertex->v);
				}
			}
			glEnd();
		}
		// show coding boundary (edges)
		if (1)
		{
			CFspinEdge *e0, *e1, *e2, *e3;
			float temp[3] = { 0.0f, 0.0f, 0.0f };
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glBegin(GL_LINES);
			for (int i = 0; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				if (e0->inv->component && !e0->component)
				{
					e0 = e0;
					e1 = e0->next;
					e2 = e1->next;
					e3 = e2->next;
					glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
					glVertex3f(e1->vertex->v[0]+e1->normal->n[0]*0.001,e1->vertex->v[1]+e1->normal->n[1]*0.001,e1->vertex->v[2]+e1->normal->n[2]*0.001);
					glVertex3f(e1->vertex->v[0]+e1->normal->n[0]*0.001,e1->vertex->v[1]+e1->normal->n[1]*0.001,e1->vertex->v[2]+e1->normal->n[2]*0.001);
					glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
					glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
					glVertex3f(e3->vertex->v[0]+e3->normal->n[0]*0.001,e3->vertex->v[1]+e3->normal->n[1]*0.001,e3->vertex->v[2]+e3->normal->n[2]*0.001);
					glVertex3f(e3->vertex->v[0]+e3->normal->n[0]*0.001,e3->vertex->v[1]+e3->normal->n[1]*0.001,e3->vertex->v[2]+e3->normal->n[2]*0.001);
					glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
				}
			}
			glEnd();
		}
		// show edge degrees
		if (1)
		{
			CFspinEdge *e;
			float temp1[3] = { 1.0f, 0.0f, 0.0f }, temp2[3], temp3[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp1);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);

			for  (i = encoder->verticesHEXdone; i < encoder->verticesHEXnumber; i++)
			{
				e = encoder->verticesHEX[i];
				while (e)
				{
					if (e->slots > 0)
					{
						int slots = e->slots;
						CFspinEdge *e_run = e->inv;
						while (slots--)
						{
							e_run = e_run->spin;

							glNormal3fv(e_run->normal->n);

							VecSubtract3fv(temp1, e_run->inv->vertex->v, e_run->vertex->v);
							VecSelfScalarMult3fv(temp1, 0.3f);
							VecAdd3fv(temp2, e_run->vertex->v, temp1);
							glVertex3fv(temp2);
							
							VecSubtract3fv(temp1, e_run->inv->vertex->v, e_run->vertex->v);
							VecSelfScalarMult3fv(temp1, 0.7f);
							VecAdd3fv(temp3, e_run->vertex->v, temp1);
							glVertex3fv(temp3);

							VecSubtract3fv(temp1, e_run->next->vertex->v, temp3);
							VecSelfScalarMult3fv(temp1, 0.2f);
							VecSelfAdd3fv(temp3, temp1);
							glVertex3fv(temp3);

							VecSubtract3fv(temp1, e_run->next->vertex->v, temp2);
							VecSelfScalarMult3fv(temp1, 0.2f);
							VecSelfAdd3fv(temp2, temp1);
							glVertex3fv(temp2);
						}
					}
					e = e->Lnext;
				}
			}
			glEnd();
		}

/*
			for  (i = 0; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				if (e->vertex->active && e->inv->vertex->active && !e->component  && !e->inv->component)
				{
					glNormal3fv(e->normal->n);

					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.2f);
					VecAdd3fv(temp2, e->vertex->v, temp1);
					glVertex3fv(temp2);
					
					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.8f);
					VecAdd3fv(temp3, e->vertex->v, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->vertex->v, temp3);
					VecSelfScalarMult3fv(temp1, 0.15f);
					VecSelfAdd3fv(temp3, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->next->vertex->v, temp2);
					VecSelfScalarMult3fv(temp1, 0.15f);
					VecSelfAdd3fv(temp2, temp1);
					glVertex3fv(temp2);
				}
			}
			glEnd();
*/
		// show neighboring hexahedrons (faces)
		if (1)
		{
			CFspinEdge *e;
			CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
			float temp[3] = { 0.0f, 1.0f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->hs->size(); i++)
			{
				e = (CFspinEdge*) mesh->hs->elementAt(i);

				if ((!e->component) && (e->inv->component || e->spin->component || e->next->spin->component || e->next->next->spin->component || e->next->next->next->spin->component || e->spin->inv->next->next->spin->component))
				{
					v0 = e->vertex;
					v1 = e->next->next->next->vertex;
					v2 = e->next->next->vertex;
					v3 = e->next->vertex;
					v4 = e->next->next->spin->inv->next->vertex;
					v5 = e->next->next->spin->inv->next->next->vertex;
					v6 = e->spin->inv->next->vertex;
					v7 = e->spin->inv->next->next->vertex;

					VecAdd3fv(temp, v0->v, v1->v);
					VecSelfAdd3fv(temp, v2->v);
					VecSelfAdd3fv(temp, v3->v);
					VecSelfAdd3fv(temp, v4->v);
					VecSelfAdd3fv(temp, v5->v);
					VecSelfAdd3fv(temp, v6->v);
					VecSelfAdd3fv(temp, v7->v);
					VecSelfScalarDiv3fv(temp, 8.0f);

#define SCALE_NEIGHBORS 0.5f

					VecSubtract3fv(temp0, temp, v0->v);
					VecSelfScalarMult3fv(temp0, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp0, v0->v);

					VecSubtract3fv(temp1, temp, v1->v);
					VecSelfScalarMult3fv(temp1, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp1, v1->v);

					VecSubtract3fv(temp2, temp, v2->v);
					VecSelfScalarMult3fv(temp2, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp2, v2->v);

					VecSubtract3fv(temp3, temp, v3->v);
					VecSelfScalarMult3fv(temp3, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp3, v3->v);

					VecSubtract3fv(temp4, temp, v4->v);
					VecSelfScalarMult3fv(temp4, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp4, v4->v);

					VecSubtract3fv(temp5, temp, v5->v);
					VecSelfScalarMult3fv(temp5, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp5, v5->v);

					VecSubtract3fv(temp6, temp, v6->v);
					VecSelfScalarMult3fv(temp6, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp6, v6->v);

					VecSubtract3fv(temp7, temp, v7->v);
					VecSelfScalarMult3fv(temp7, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp7, v7->v);

					glNormal3fv(e->inv->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp3);
					glVertex3fv(temp2);
					glVertex3fv(temp1);

					glNormal3fv(e->spin->normal->n);
					glVertex3fv(temp1);
					glVertex3fv(temp6);
					glVertex3fv(temp7);
					glVertex3fv(temp0);

					glNormal3fv(e->next->spin->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp7);
					glVertex3fv(temp4);
					glVertex3fv(temp3);

					glNormal3fv(e->next->next->spin->normal->n);
					glVertex3fv(temp3);
					glVertex3fv(temp4);
					glVertex3fv(temp5);
					glVertex3fv(temp2);

					glNormal3fv(e->next->next->next->spin->normal->n);
					glVertex3fv(temp2);
					glVertex3fv(temp5);
					glVertex3fv(temp6);
					glVertex3fv(temp1);

					glNormal3fv(e->spin->inv->next->next->spin->normal->n);
					glVertex3fv(temp6);
					glVertex3fv(temp5);
					glVertex3fv(temp4);
					glVertex3fv(temp7);
				}
			}
			glEnd();
		}
		// show neighboring hexahedrons (edges)
		if (1)
		{
			CFspinEdge *e;
			CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
			float temp[3] = { 0.0f, 0.5f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_LINE);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->hs->size(); i++)
			{
				e = (CFspinEdge*) mesh->hs->elementAt(i);

				if ((!e->component) && (e->inv->component || e->spin->component || e->next->spin->component || e->next->next->spin->component || e->next->next->next->spin->component || e->spin->inv->next->next->spin->component))
				{
					v0 = e->vertex;
					v1 = e->next->next->next->vertex;
					v2 = e->next->next->vertex;
					v3 = e->next->vertex;
					v4 = e->next->next->spin->inv->next->vertex;
					v5 = e->next->next->spin->inv->next->next->vertex;
					v6 = e->spin->inv->next->vertex;
					v7 = e->spin->inv->next->next->vertex;

					VecAdd3fv(temp, v0->v, v1->v);
					VecSelfAdd3fv(temp, v2->v);
					VecSelfAdd3fv(temp, v3->v);
					VecSelfAdd3fv(temp, v4->v);
					VecSelfAdd3fv(temp, v5->v);
					VecSelfAdd3fv(temp, v6->v);
					VecSelfAdd3fv(temp, v7->v);
					VecSelfScalarDiv3fv(temp, 8.0f);

#define SCALE_NEIGHBORS_LINE 0.4995f

					VecSubtract3fv(temp0, temp, v0->v);
					VecSelfScalarMult3fv(temp0, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp0, v0->v);

					VecSubtract3fv(temp1, temp, v1->v);
					VecSelfScalarMult3fv(temp1, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp1, v1->v);

					VecSubtract3fv(temp2, temp, v2->v);
					VecSelfScalarMult3fv(temp2, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp2, v2->v);

					VecSubtract3fv(temp3, temp, v3->v);
					VecSelfScalarMult3fv(temp3, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp3, v3->v);

					VecSubtract3fv(temp4, temp, v4->v);
					VecSelfScalarMult3fv(temp4, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp4, v4->v);

					VecSubtract3fv(temp5, temp, v5->v);
					VecSelfScalarMult3fv(temp5, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp5, v5->v);

					VecSubtract3fv(temp6, temp, v6->v);
					VecSelfScalarMult3fv(temp6, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp6, v6->v);

					VecSubtract3fv(temp7, temp, v7->v);
					VecSelfScalarMult3fv(temp7, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp7, v7->v);

					glNormal3fv(e->inv->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp3);
					glVertex3fv(temp2);
					glVertex3fv(temp1);

					glNormal3fv(e->spin->normal->n);
					glVertex3fv(temp1);
					glVertex3fv(temp6);
					glVertex3fv(temp7);
					glVertex3fv(temp0);

					glNormal3fv(e->next->spin->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp7);
					glVertex3fv(temp4);
					glVertex3fv(temp3);

					glNormal3fv(e->next->next->spin->normal->n);
					glVertex3fv(temp3);
					glVertex3fv(temp4);
					glVertex3fv(temp5);
					glVertex3fv(temp2);

					glNormal3fv(e->next->next->next->spin->normal->n);
					glVertex3fv(temp2);
					glVertex3fv(temp5);
					glVertex3fv(temp6);
					glVertex3fv(temp1);

					glNormal3fv(e->spin->inv->next->next->spin->normal->n);
					glVertex3fv(temp6);
					glVertex3fv(temp5);
					glVertex3fv(temp4);
					glVertex3fv(temp7);
				}
			}
			glEnd();
		}
		// show mesh->gate hexahedron (faces)
		if (mesh->gate)
		{
			CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
			float temp[3] = { 0.0f, 1.0f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			v0 = mesh->gate->vertex;
			v1 = mesh->gate->next->next->next->vertex;
			v2 = mesh->gate->next->next->vertex;
			v3 = mesh->gate->next->vertex;
			v4 = mesh->gate->next->next->spin->inv->next->vertex;
			v5 = mesh->gate->next->next->spin->inv->next->next->vertex;
			v6 = mesh->gate->spin->inv->next->vertex;
			v7 = mesh->gate->spin->inv->next->next->vertex;

			VecAdd3fv(temp, v0->v, v1->v);
			VecSelfAdd3fv(temp, v2->v);
			VecSelfAdd3fv(temp, v3->v);
			VecSelfAdd3fv(temp, v4->v);
			VecSelfAdd3fv(temp, v5->v);
			VecSelfAdd3fv(temp, v6->v);
			VecSelfAdd3fv(temp, v7->v);
			VecSelfScalarDiv3fv(temp, 8.0f);

			VecSubtract3fv(temp0, temp, v0->v);
			VecSelfScalarMult3fv(temp0, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp0, v0->v);

			VecSubtract3fv(temp1, temp, v1->v);
			VecSelfScalarMult3fv(temp1, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp1, v1->v);

			VecSubtract3fv(temp2, temp, v2->v);
			VecSelfScalarMult3fv(temp2, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp2, v2->v);

			VecSubtract3fv(temp3, temp, v3->v);
			VecSelfScalarMult3fv(temp3, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp3, v3->v);

			VecSubtract3fv(temp4, temp, v4->v);
			VecSelfScalarMult3fv(temp4, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp4, v4->v);

			VecSubtract3fv(temp5, temp, v5->v);
			VecSelfScalarMult3fv(temp5, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp5, v5->v);

			VecSubtract3fv(temp6, temp, v6->v);
			VecSelfScalarMult3fv(temp6, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp6, v6->v);

			VecSubtract3fv(temp7, temp, v7->v);
			VecSelfScalarMult3fv(temp7, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp7, v7->v);

			glNormal3fv(mesh->gate->inv->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp3);
			glVertex3fv(temp2);
			glVertex3fv(temp1);

			glNormal3fv(mesh->gate->spin->normal->n);
			glVertex3fv(temp1);
			glVertex3fv(temp6);
			glVertex3fv(temp7);
			glVertex3fv(temp0);

			glNormal3fv(mesh->gate->next->spin->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp7);
			glVertex3fv(temp4);
			glVertex3fv(temp3);

			glNormal3fv(mesh->gate->next->next->spin->normal->n);
			glVertex3fv(temp3);
			glVertex3fv(temp4);
			glVertex3fv(temp5);
			glVertex3fv(temp2);

			glNormal3fv(mesh->gate->next->next->next->spin->normal->n);
			glVertex3fv(temp2);
			glVertex3fv(temp5);
			glVertex3fv(temp6);
			glVertex3fv(temp1);

			glNormal3fv(mesh->gate->spin->inv->next->next->spin->normal->n);
			glVertex3fv(temp6);
			glVertex3fv(temp5);
			glVertex3fv(temp4);
			glVertex3fv(temp7);
			glEnd();
		}
				// show mesh->gate hexahedron (edges)
		if (mesh->gate)
		{
			CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
			float temp[3] = { 0.0f, 0.5f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_LINE);
			glBegin(GL_QUADS);
			v0 = mesh->gate->vertex;
			v1 = mesh->gate->next->next->next->vertex;
			v2 = mesh->gate->next->next->vertex;
			v3 = mesh->gate->next->vertex;
			v4 = mesh->gate->next->next->spin->inv->next->vertex;
			v5 = mesh->gate->next->next->spin->inv->next->next->vertex;
			v6 = mesh->gate->spin->inv->next->vertex;
			v7 = mesh->gate->spin->inv->next->next->vertex;

			VecAdd3fv(temp, v0->v, v1->v);
			VecSelfAdd3fv(temp, v2->v);
			VecSelfAdd3fv(temp, v3->v);
			VecSelfAdd3fv(temp, v4->v);
			VecSelfAdd3fv(temp, v5->v);
			VecSelfAdd3fv(temp, v6->v);
			VecSelfAdd3fv(temp, v7->v);
			VecSelfScalarDiv3fv(temp, 8.0f);

			VecSubtract3fv(temp0, temp, v0->v);
			VecSelfScalarMult3fv(temp0, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp0, v0->v);

			VecSubtract3fv(temp1, temp, v1->v);
			VecSelfScalarMult3fv(temp1, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp1, v1->v);

			VecSubtract3fv(temp2, temp, v2->v);
			VecSelfScalarMult3fv(temp2, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp2, v2->v);

			VecSubtract3fv(temp3, temp, v3->v);
			VecSelfScalarMult3fv(temp3, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp3, v3->v);

			VecSubtract3fv(temp4, temp, v4->v);
			VecSelfScalarMult3fv(temp4, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp4, v4->v);

			VecSubtract3fv(temp5, temp, v5->v);
			VecSelfScalarMult3fv(temp5, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp5, v5->v);

			VecSubtract3fv(temp6, temp, v6->v);
			VecSelfScalarMult3fv(temp6, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp6, v6->v);

			VecSubtract3fv(temp7, temp, v7->v);
			VecSelfScalarMult3fv(temp7, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp7, v7->v);

			glNormal3fv(mesh->gate->inv->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp3);
			glVertex3fv(temp2);
			glVertex3fv(temp1);

			glNormal3fv(mesh->gate->spin->normal->n);
			glVertex3fv(temp1);
			glVertex3fv(temp6);
			glVertex3fv(temp7);
			glVertex3fv(temp0);

			glNormal3fv(mesh->gate->next->spin->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp7);
			glVertex3fv(temp4);
			glVertex3fv(temp3);

			glNormal3fv(mesh->gate->next->next->spin->normal->n);
			glVertex3fv(temp3);
			glVertex3fv(temp4);
			glVertex3fv(temp5);
			glVertex3fv(temp2);

			glNormal3fv(mesh->gate->next->next->next->spin->normal->n);
			glVertex3fv(temp2);
			glVertex3fv(temp5);
			glVertex3fv(temp6);
			glVertex3fv(temp1);

			glNormal3fv(mesh->gate->spin->inv->next->next->spin->normal->n);
			glVertex3fv(temp6);
			glVertex3fv(temp5);
			glVertex3fv(temp4);
			glVertex3fv(temp7);
			glEnd();
		}
	}
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE); 

	if (mesh->active)
	{
		glColor3f(0.5f, 0.5f, 0.9f);
		glLineWidth(6.0);
		glBegin(GL_LINES);
		glVertex3fv(mesh->active->vertex->v);
		glVertex3fv(mesh->active->inv->vertex->v);
		glEnd();
	}

	if ( WireFrameOn )
	{
		CFspinEdge *e0, *e1, *e2, *e3;
		glLineWidth(1.0f);
		glBegin(GL_LINES);
		if ( BoundaryOn )
		{
			if ( BoundaryOn == 3 )
			{
				glColor3f(1.0f, 1.0f, 1.0f);
			}
			else
			{
				glColor3f(0.0f, 0.0f, 0.0f);
			}

			for  (i = mesh->interior_faces; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				e1 = e0->next;
				e2 = e1->next;
				e3 = e2->next;
				glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
				glVertex3f(e1->vertex->v[0]+e1->normal->n[0]*0.001,e1->vertex->v[1]+e1->normal->n[1]*0.001,e1->vertex->v[2]+e1->normal->n[2]*0.001);
				glVertex3f(e1->vertex->v[0]+e1->normal->n[0]*0.001,e1->vertex->v[1]+e1->normal->n[1]*0.001,e1->vertex->v[2]+e1->normal->n[2]*0.001);
				glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
				glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
				glVertex3f(e3->vertex->v[0]+e3->normal->n[0]*0.001,e3->vertex->v[1]+e3->normal->n[1]*0.001,e3->vertex->v[2]+e3->normal->n[2]*0.001);
				glVertex3f(e3->vertex->v[0]+e3->normal->n[0]*0.001,e3->vertex->v[1]+e3->normal->n[1]*0.001,e3->vertex->v[2]+e3->normal->n[2]*0.001);
				glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
			}
		}
		glEnd();
	}

	if ( VertexNormalsOn )
	{
		CFspinEdge *e;
		glColor3f(1.0f, 1.0f, 1.0f);
		glLineWidth(1.0f);
		glBegin(GL_LINES);
		if ( BoundaryOn )
		{
			for  (i = mesh->interior_edges; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				glVertex3fv(e->vertex->v);
				glVertex3f(e->vertex->v[0]+e->normal->n[0]*0.05,e->vertex->v[1]+e->normal->n[1]*0.05,e->vertex->v[2]+e->normal->n[2]*0.05);
			}
		}
		if ( InteriorOn )
		{
			for  (i = 0; i < mesh->interior_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				glVertex3fv(e->vertex->v);
				glVertex3f(e->vertex->v[0]+e->normal->n[0]*0.05,e->vertex->v[1]+e->normal->n[1]*0.05,e->vertex->v[2]+e->normal->n[2]*0.05);
			}
		}
		glEnd();
	}
}

void myPolDisplay()
{
	int i;

	glEnable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);  

	if (BoundaryOn)
	{
		CFmaterial* m = mesh->boundary_material;
		glMaterialfv(GL_FRONT,GL_AMBIENT,m->a);
		glMaterialfv(GL_FRONT,GL_DIFFUSE,m->d);
		glMaterialfv(GL_FRONT,GL_SPECULAR,m->s);
		glMaterialfv(GL_FRONT,GL_EMISSION,m->e);

		if (BoundaryOn ==	1 || BoundaryOn == 2) // draw polygons = 1 or lines = 2 or points = 3
		{
			CFspinEdge *e0, *e1;
			if (BoundaryOn == 1)
			{
				glPolygonMode(GL_FRONT, GL_FILL);
			}
			else
			{
				glPolygonMode(GL_FRONT, GL_LINE);
			}
			glBegin(GL_TRIANGLES);
			for (int i = mesh->interior_faces; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				glNormal3fv(e0->normal->n);
				e1 = e0->next;
				while (e1->next != e0)
				{
					glVertex3fv(e0->vertex->v);
					glVertex3fv(e1->next->vertex->v);
					glVertex3fv(e1->vertex->v);
					e1 = e1->next;
				}
			}
			glEnd();

			if (1 )
			{
				float rouge[4] = {2.0f, 0.0f, 0.0f, 1.0f};
				glMaterialfv(GL_FRONT,GL_DIFFUSE,rouge);
				glBegin(GL_LINES);
				for (int i = mesh->interior_faces; i < mesh->total_faces; i++)
				{
					e0 = (CFspinEdge*) mesh->fs->elementAt(i);
					int counter = 2;
					float temp[3];

					VecAdd3fv(temp, e0->vertex->v, e0->next->vertex->v);

					e1 = e0->next->next;
					while (e1 != e0)
					{
						VecSelfAdd3fv(temp, e1->vertex->v);
						counter++;
						e1 = e1->next;
					}
					VecSelfScalarDiv3fv(temp, counter);
					glVertex3fv(temp);
					VecSelfAddScalarMult3fv(temp, e0->normal->n, 0.01f);
					glVertex3fv(temp);
				}
				glEnd();
			}
		}
		else
		{
			CFvertex *v;
			CFspinEdge *e;
			for  (i = mesh->interior_edges; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				e->vertex->marker = 1;
			}
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->total_vertices; i++)
			{
				v = (CFvertex*) mesh->vs->elementAt(i);
				if (v->marker)
				{
					#define DRAWVERTEX(x,y,z,s) glNormal3f( 1,0,0); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y+s,z-s); glNormal3f(-1,0,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y+s,z-s); glVertex3f(x-s,y+s,z+s); glVertex3f(x-s,y-s,z+s); glNormal3f(0, 1,0); glVertex3f(x-s,y+s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x+s,y+s,z+s); glVertex3f(x-s,y+s,z+s); glNormal3f(0,-1,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y-s,z+s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y-s,z-s); glNormal3f(0,0, 1); glVertex3f(x-s,y-s,z+s); glVertex3f(x-s,y+s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y-s,z+s); glNormal3f(0,0,-1); glVertex3f(x-s,y-s,z-s); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x-s,y+s,z-s);
					DRAWVERTEX(v->v[0], v->v[1],v->v[2], 0.005f);
					#undef DRAWVERTEX
					v->marker = 0;
				}
			}
			glEnd();
		}
	}
	if (InteriorOn)
	{
		CFmaterial* m = mesh->interior_material;
		glMaterialfv(GL_FRONT,GL_AMBIENT,m->a);
		glMaterialfv(GL_FRONT,GL_DIFFUSE,m->d);
		glMaterialfv(GL_FRONT,GL_SPECULAR,m->s);
		glMaterialfv(GL_FRONT,GL_EMISSION,m->e);

		if (InteriorOn ==	1 || InteriorOn == 2) // draw polygons = 1 or lines = 2 or points = 3
		{
			CFspinEdge *e0, *e1;
			if (InteriorOn == 1)
			{
				glPolygonMode(GL_FRONT, GL_FILL);
			}
			else
			{
				glPolygonMode(GL_FRONT, GL_LINE);
			}
			glBegin(GL_TRIANGLES);
			for (int i = 0; i < mesh->interior_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				glNormal3fv(e0->normal->n);
				e1 = e0->next;
				while (e1->next != e0)
				{
					glVertex3fv(e0->vertex->v);
					glVertex3fv(e1->next->vertex->v);
					glVertex3fv(e1->vertex->v);
					e1 = e1->next;
				}
			}
			glEnd();
		}
		else
		{
			CFvertex *v;
			CFspinEdge *e;
			for  (i = mesh->interior_edges; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				e->vertex->marker = 1;
			}
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->total_vertices; i++)
			{
				v = (CFvertex*) mesh->vs->elementAt(i);
				if (v->marker == 0)
				{
					#define DRAWVERTEX(x,y,z,s) glNormal3f( 1,0,0); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y+s,z-s); glNormal3f(-1,0,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y+s,z-s); glVertex3f(x-s,y+s,z+s); glVertex3f(x-s,y-s,z+s); glNormal3f(0, 1,0); glVertex3f(x-s,y+s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x+s,y+s,z+s); glVertex3f(x-s,y+s,z+s); glNormal3f(0,-1,0); glVertex3f(x-s,y-s,z-s); glVertex3f(x-s,y-s,z+s); glVertex3f(x+s,y-s,z+s); glVertex3f(x+s,y-s,z-s); glNormal3f(0,0, 1); glVertex3f(x-s,y-s,z+s); glVertex3f(x-s,y+s,z+s); glVertex3f(x+s,y+s,z+s); glVertex3f(x+s,y-s,z+s); glNormal3f(0,0,-1); glVertex3f(x-s,y-s,z-s); glVertex3f(x+s,y-s,z-s); glVertex3f(x+s,y+s,z-s); glVertex3f(x-s,y+s,z-s);
					DRAWVERTEX(v->v[0], v->v[1],v->v[2], 0.0025f);
					#undef DRAWVERTEX
				}
				else
				{
					v->marker = 0;
				}
			}
			glEnd();
		}
	}
	if ( TetrahedraOn )
	{
		CFspinEdge *e;
		CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
		float temp[3] = { 0.0f, 1.0f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
		glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
		glPolygonMode(GL_FRONT, GL_FILL);
		glBegin(GL_QUADS);
		for  (i = 0; i < mesh->hs->size(); i++)
		{
			e = (CFspinEdge*) mesh->hs->elementAt(i);

			v0 = e->vertex;
			v1 = e->next->next->next->vertex;
			v2 = e->next->next->vertex;
			v3 = e->next->vertex;
			v4 = e->next->next->spin->inv->next->vertex;
			v5 = e->next->next->spin->inv->next->next->vertex;
			v6 = e->spin->inv->next->vertex;
			v7 = e->spin->inv->next->next->vertex;

			VecAdd3fv(temp, v0->v, v1->v);
			VecSelfAdd3fv(temp, v2->v);
			VecSelfAdd3fv(temp, v3->v);
			VecSelfAdd3fv(temp, v4->v);
			VecSelfAdd3fv(temp, v5->v);
			VecSelfAdd3fv(temp, v6->v);
			VecSelfAdd3fv(temp, v7->v);
			VecSelfScalarDiv3fv(temp, 8.0f);

			VecSubtract3fv(temp0, temp, v0->v);
			VecSelfScalarMult3fv(temp0, 0.3f);
			VecSelfAdd3fv(temp0, v0->v);

			VecSubtract3fv(temp1, temp, v1->v);
			VecSelfScalarMult3fv(temp1, 0.3f);
			VecSelfAdd3fv(temp1, v1->v);

			VecSubtract3fv(temp2, temp, v2->v);
			VecSelfScalarMult3fv(temp2, 0.3f);
			VecSelfAdd3fv(temp2, v2->v);

			VecSubtract3fv(temp3, temp, v3->v);
			VecSelfScalarMult3fv(temp3, 0.3f);
			VecSelfAdd3fv(temp3, v3->v);

			VecSubtract3fv(temp4, temp, v4->v);
			VecSelfScalarMult3fv(temp4, 0.3f);
			VecSelfAdd3fv(temp4, v4->v);

			VecSubtract3fv(temp5, temp, v5->v);
			VecSelfScalarMult3fv(temp5, 0.3f);
			VecSelfAdd3fv(temp5, v5->v);

			VecSubtract3fv(temp6, temp, v6->v);
			VecSelfScalarMult3fv(temp6, 0.3f);
			VecSelfAdd3fv(temp6, v6->v);

			VecSubtract3fv(temp7, temp, v7->v);
			VecSelfScalarMult3fv(temp7, 0.3f);
			VecSelfAdd3fv(temp7, v7->v);

			glNormal3fv(e->inv->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp3);
			glVertex3fv(temp2);
			glVertex3fv(temp1);

			glNormal3fv(e->spin->normal->n);
			glVertex3fv(temp1);
			glVertex3fv(temp6);
			glVertex3fv(temp7);
			glVertex3fv(temp0);

			glNormal3fv(e->next->spin->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp7);
			glVertex3fv(temp4);
			glVertex3fv(temp3);

			glNormal3fv(e->next->next->spin->normal->n);
			glVertex3fv(temp3);
			glVertex3fv(temp4);
			glVertex3fv(temp5);
			glVertex3fv(temp2);

			glNormal3fv(e->next->next->next->spin->normal->n);
			glVertex3fv(temp2);
			glVertex3fv(temp5);
			glVertex3fv(temp6);
			glVertex3fv(temp1);

			glNormal3fv(e->spin->inv->next->next->spin->normal->n);
			glVertex3fv(temp6);
			glVertex3fv(temp5);
			glVertex3fv(temp4);
			glVertex3fv(temp7);
		}
		glEnd();
	}
	if ( CodingOn )
	{
		// show next gate
		if (1)
		{
			CFspinEdge *e0, *e1, *e2, *e3;
			float temp[3] = { 0.9f, 0.4f, 0.9f };
			float offset = 0.001f;
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			if (mesh->gate)
			{
				e0 = mesh->gate;
				e1 = e0->next;
				e2 = e1->next;
				e3 = e2->next;
				glNormal3fv(e0->normal->n);
				glVertex3f(e2->vertex->v[0]+offset*e0->normal->n[0],e2->vertex->v[1]+offset*e0->normal->n[1],e2->vertex->v[2]+offset*e0->normal->n[2]);
				glVertex3f(e1->vertex->v[0]+offset*e0->normal->n[0],e1->vertex->v[1]+offset*e0->normal->n[1],e1->vertex->v[2]+offset*e0->normal->n[2]);
				temp[1] = 0.0f;
				glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
				glVertex3f(e0->vertex->v[0]+offset*e0->normal->n[0],e0->vertex->v[1]+offset*e0->normal->n[1],e0->vertex->v[2]+offset*e0->normal->n[2]);
				glVertex3f(e3->vertex->v[0]+offset*e0->normal->n[0],e3->vertex->v[1]+offset*e0->normal->n[1],e3->vertex->v[2]+offset*e0->normal->n[2]);
			}
			glEnd();
		}
		// show coding boundary (faces)
		if (1)
		{
			CFspinEdge *e0, *e1, *e2, *e3;
			float temp1[3] = { 0.0f, 0.0f, 1.0f };
			float temp4[3] = { 0.0f, 0.0f, 2.0f };
			float temp2[3] = { 0.75f, 0.75f, 1.5f };
			float temp3[3] = { 0.1f, 0.1f, 0.2f };
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp1);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for (int i = 0; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				if (e0->inv->component && !e0->component)
				{
					if (e0->group == -1)
					{
						glMaterialfv(GL_FRONT,GL_AMBIENT,temp4);
						glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
					}
					else
					{
						glMaterialfv(GL_FRONT,GL_AMBIENT,temp2);
						glMaterialfv(GL_FRONT,GL_DIFFUSE,temp3);
					}
					e0 = e0;
					e1 = e0->next;
					e2 = e1->next;
					e3 = e2->next;
					glNormal3fv(e0->normal->n);
					glVertex3fv(e3->vertex->v);
					glVertex3fv(e2->vertex->v);
					glVertex3fv(e1->vertex->v);
					glVertex3fv(e0->vertex->v);
				}
			}
			glEnd();
		}
		// show coding boundary (edges)
		if (1)
		{
			CFspinEdge *e0, *e1, *e2, *e3;
			float temp[3] = { 0.0f, 0.0f, 0.0f };
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glBegin(GL_LINES);
			for (int i = 0; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				if (e0->inv->component && !e0->component)
				{
					e0 = e0;
					e1 = e0->next;
					e2 = e1->next;
					e3 = e2->next;
					glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
					glVertex3f(e1->vertex->v[0]+e1->normal->n[0]*0.001,e1->vertex->v[1]+e1->normal->n[1]*0.001,e1->vertex->v[2]+e1->normal->n[2]*0.001);
					glVertex3f(e1->vertex->v[0]+e1->normal->n[0]*0.001,e1->vertex->v[1]+e1->normal->n[1]*0.001,e1->vertex->v[2]+e1->normal->n[2]*0.001);
					glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
					glVertex3f(e2->vertex->v[0]+e2->normal->n[0]*0.001,e2->vertex->v[1]+e2->normal->n[1]*0.001,e2->vertex->v[2]+e2->normal->n[2]*0.001);
					glVertex3f(e3->vertex->v[0]+e3->normal->n[0]*0.001,e3->vertex->v[1]+e3->normal->n[1]*0.001,e3->vertex->v[2]+e3->normal->n[2]*0.001);
					glVertex3f(e3->vertex->v[0]+e3->normal->n[0]*0.001,e3->vertex->v[1]+e3->normal->n[1]*0.001,e3->vertex->v[2]+e3->normal->n[2]*0.001);
					glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
				}
			}
			glEnd();
		}
		// show edge degrees
		if (1)
		{
			CFspinEdge *e;
			float temp1[3] = { 1.0f, 0.0f, 0.0f }, temp2[3], temp3[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp1);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp1);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				if (e->vertex->active && e->inv->vertex->active && !e->component  && !e->inv->component)
				{
					glNormal3fv(e->normal->n);

					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.2f);
					VecAdd3fv(temp2, e->vertex->v, temp1);
					glVertex3fv(temp2);
					
					VecSubtract3fv(temp1, e->inv->vertex->v, e->vertex->v);
					VecSelfScalarMult3fv(temp1, 0.8f);
					VecAdd3fv(temp3, e->vertex->v, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->vertex->v, temp3);
					VecSelfScalarMult3fv(temp1, 0.15f);
					VecSelfAdd3fv(temp3, temp1);
					glVertex3fv(temp3);

					VecSubtract3fv(temp1, e->next->next->vertex->v, temp2);
					VecSelfScalarMult3fv(temp1, 0.15f);
					VecSelfAdd3fv(temp2, temp1);
					glVertex3fv(temp2);
				}
			}
			glEnd();
		}

		// show neighboring hexahedrons (faces)
		if (1)
		{
			CFspinEdge *e;
			CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
			float temp[3] = { 0.0f, 1.0f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->hs->size(); i++)
			{
				e = (CFspinEdge*) mesh->hs->elementAt(i);

				if ((!e->component) && (e->inv->component || e->spin->component || e->next->spin->component || e->next->next->spin->component || e->next->next->next->spin->component || e->spin->inv->next->next->spin->component))
				{
					v0 = e->vertex;
					v1 = e->next->next->next->vertex;
					v2 = e->next->next->vertex;
					v3 = e->next->vertex;
					v4 = e->next->next->spin->inv->next->vertex;
					v5 = e->next->next->spin->inv->next->next->vertex;
					v6 = e->spin->inv->next->vertex;
					v7 = e->spin->inv->next->next->vertex;

					VecAdd3fv(temp, v0->v, v1->v);
					VecSelfAdd3fv(temp, v2->v);
					VecSelfAdd3fv(temp, v3->v);
					VecSelfAdd3fv(temp, v4->v);
					VecSelfAdd3fv(temp, v5->v);
					VecSelfAdd3fv(temp, v6->v);
					VecSelfAdd3fv(temp, v7->v);
					VecSelfScalarDiv3fv(temp, 8.0f);

#define SCALE_NEIGHBORS 0.5f

					VecSubtract3fv(temp0, temp, v0->v);
					VecSelfScalarMult3fv(temp0, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp0, v0->v);

					VecSubtract3fv(temp1, temp, v1->v);
					VecSelfScalarMult3fv(temp1, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp1, v1->v);

					VecSubtract3fv(temp2, temp, v2->v);
					VecSelfScalarMult3fv(temp2, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp2, v2->v);

					VecSubtract3fv(temp3, temp, v3->v);
					VecSelfScalarMult3fv(temp3, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp3, v3->v);

					VecSubtract3fv(temp4, temp, v4->v);
					VecSelfScalarMult3fv(temp4, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp4, v4->v);

					VecSubtract3fv(temp5, temp, v5->v);
					VecSelfScalarMult3fv(temp5, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp5, v5->v);

					VecSubtract3fv(temp6, temp, v6->v);
					VecSelfScalarMult3fv(temp6, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp6, v6->v);

					VecSubtract3fv(temp7, temp, v7->v);
					VecSelfScalarMult3fv(temp7, SCALE_NEIGHBORS);
					VecSelfAdd3fv(temp7, v7->v);

					glNormal3fv(e->inv->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp3);
					glVertex3fv(temp2);
					glVertex3fv(temp1);

					glNormal3fv(e->spin->normal->n);
					glVertex3fv(temp1);
					glVertex3fv(temp6);
					glVertex3fv(temp7);
					glVertex3fv(temp0);

					glNormal3fv(e->next->spin->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp7);
					glVertex3fv(temp4);
					glVertex3fv(temp3);

					glNormal3fv(e->next->next->spin->normal->n);
					glVertex3fv(temp3);
					glVertex3fv(temp4);
					glVertex3fv(temp5);
					glVertex3fv(temp2);

					glNormal3fv(e->next->next->next->spin->normal->n);
					glVertex3fv(temp2);
					glVertex3fv(temp5);
					glVertex3fv(temp6);
					glVertex3fv(temp1);

					glNormal3fv(e->spin->inv->next->next->spin->normal->n);
					glVertex3fv(temp6);
					glVertex3fv(temp5);
					glVertex3fv(temp4);
					glVertex3fv(temp7);
				}
			}
			glEnd();
		}
		// show neighboring hexahedrons (edges)
		if (1)
		{
			CFspinEdge *e;
			CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
			float temp[3] = { 0.0f, 0.5f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_LINE);
			glBegin(GL_QUADS);
			for  (i = 0; i < mesh->hs->size(); i++)
			{
				e = (CFspinEdge*) mesh->hs->elementAt(i);

				if ((!e->component) && (e->inv->component || e->spin->component || e->next->spin->component || e->next->next->spin->component || e->next->next->next->spin->component || e->spin->inv->next->next->spin->component))
				{
					v0 = e->vertex;
					v1 = e->next->next->next->vertex;
					v2 = e->next->next->vertex;
					v3 = e->next->vertex;
					v4 = e->next->next->spin->inv->next->vertex;
					v5 = e->next->next->spin->inv->next->next->vertex;
					v6 = e->spin->inv->next->vertex;
					v7 = e->spin->inv->next->next->vertex;

					VecAdd3fv(temp, v0->v, v1->v);
					VecSelfAdd3fv(temp, v2->v);
					VecSelfAdd3fv(temp, v3->v);
					VecSelfAdd3fv(temp, v4->v);
					VecSelfAdd3fv(temp, v5->v);
					VecSelfAdd3fv(temp, v6->v);
					VecSelfAdd3fv(temp, v7->v);
					VecSelfScalarDiv3fv(temp, 8.0f);

#define SCALE_NEIGHBORS_LINE 0.4995f

					VecSubtract3fv(temp0, temp, v0->v);
					VecSelfScalarMult3fv(temp0, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp0, v0->v);

					VecSubtract3fv(temp1, temp, v1->v);
					VecSelfScalarMult3fv(temp1, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp1, v1->v);

					VecSubtract3fv(temp2, temp, v2->v);
					VecSelfScalarMult3fv(temp2, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp2, v2->v);

					VecSubtract3fv(temp3, temp, v3->v);
					VecSelfScalarMult3fv(temp3, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp3, v3->v);

					VecSubtract3fv(temp4, temp, v4->v);
					VecSelfScalarMult3fv(temp4, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp4, v4->v);

					VecSubtract3fv(temp5, temp, v5->v);
					VecSelfScalarMult3fv(temp5, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp5, v5->v);

					VecSubtract3fv(temp6, temp, v6->v);
					VecSelfScalarMult3fv(temp6, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp6, v6->v);

					VecSubtract3fv(temp7, temp, v7->v);
					VecSelfScalarMult3fv(temp7, SCALE_NEIGHBORS_LINE);
					VecSelfAdd3fv(temp7, v7->v);

					glNormal3fv(e->inv->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp3);
					glVertex3fv(temp2);
					glVertex3fv(temp1);

					glNormal3fv(e->spin->normal->n);
					glVertex3fv(temp1);
					glVertex3fv(temp6);
					glVertex3fv(temp7);
					glVertex3fv(temp0);

					glNormal3fv(e->next->spin->normal->n);
					glVertex3fv(temp0);
					glVertex3fv(temp7);
					glVertex3fv(temp4);
					glVertex3fv(temp3);

					glNormal3fv(e->next->next->spin->normal->n);
					glVertex3fv(temp3);
					glVertex3fv(temp4);
					glVertex3fv(temp5);
					glVertex3fv(temp2);

					glNormal3fv(e->next->next->next->spin->normal->n);
					glVertex3fv(temp2);
					glVertex3fv(temp5);
					glVertex3fv(temp6);
					glVertex3fv(temp1);

					glNormal3fv(e->spin->inv->next->next->spin->normal->n);
					glVertex3fv(temp6);
					glVertex3fv(temp5);
					glVertex3fv(temp4);
					glVertex3fv(temp7);
				}
			}
			glEnd();
		}
		// show mesh->gate hexahedron (faces)
		if (mesh->gate)
		{
			CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
			float temp[3] = { 0.0f, 1.0f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_FILL);
			glBegin(GL_QUADS);
			v0 = mesh->gate->vertex;
			v1 = mesh->gate->next->next->next->vertex;
			v2 = mesh->gate->next->next->vertex;
			v3 = mesh->gate->next->vertex;
			v4 = mesh->gate->next->next->spin->inv->next->vertex;
			v5 = mesh->gate->next->next->spin->inv->next->next->vertex;
			v6 = mesh->gate->spin->inv->next->vertex;
			v7 = mesh->gate->spin->inv->next->next->vertex;

			VecAdd3fv(temp, v0->v, v1->v);
			VecSelfAdd3fv(temp, v2->v);
			VecSelfAdd3fv(temp, v3->v);
			VecSelfAdd3fv(temp, v4->v);
			VecSelfAdd3fv(temp, v5->v);
			VecSelfAdd3fv(temp, v6->v);
			VecSelfAdd3fv(temp, v7->v);
			VecSelfScalarDiv3fv(temp, 8.0f);

			VecSubtract3fv(temp0, temp, v0->v);
			VecSelfScalarMult3fv(temp0, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp0, v0->v);

			VecSubtract3fv(temp1, temp, v1->v);
			VecSelfScalarMult3fv(temp1, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp1, v1->v);

			VecSubtract3fv(temp2, temp, v2->v);
			VecSelfScalarMult3fv(temp2, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp2, v2->v);

			VecSubtract3fv(temp3, temp, v3->v);
			VecSelfScalarMult3fv(temp3, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp3, v3->v);

			VecSubtract3fv(temp4, temp, v4->v);
			VecSelfScalarMult3fv(temp4, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp4, v4->v);

			VecSubtract3fv(temp5, temp, v5->v);
			VecSelfScalarMult3fv(temp5, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp5, v5->v);

			VecSubtract3fv(temp6, temp, v6->v);
			VecSelfScalarMult3fv(temp6, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp6, v6->v);

			VecSubtract3fv(temp7, temp, v7->v);
			VecSelfScalarMult3fv(temp7, SCALE_NEIGHBORS);
			VecSelfAdd3fv(temp7, v7->v);

			glNormal3fv(mesh->gate->inv->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp3);
			glVertex3fv(temp2);
			glVertex3fv(temp1);

			glNormal3fv(mesh->gate->spin->normal->n);
			glVertex3fv(temp1);
			glVertex3fv(temp6);
			glVertex3fv(temp7);
			glVertex3fv(temp0);

			glNormal3fv(mesh->gate->next->spin->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp7);
			glVertex3fv(temp4);
			glVertex3fv(temp3);

			glNormal3fv(mesh->gate->next->next->spin->normal->n);
			glVertex3fv(temp3);
			glVertex3fv(temp4);
			glVertex3fv(temp5);
			glVertex3fv(temp2);

			glNormal3fv(mesh->gate->next->next->next->spin->normal->n);
			glVertex3fv(temp2);
			glVertex3fv(temp5);
			glVertex3fv(temp6);
			glVertex3fv(temp1);

			glNormal3fv(mesh->gate->spin->inv->next->next->spin->normal->n);
			glVertex3fv(temp6);
			glVertex3fv(temp5);
			glVertex3fv(temp4);
			glVertex3fv(temp7);
			glEnd();
		}
				// show mesh->gate hexahedron (edges)
		if (mesh->gate)
		{
			CFvertex *v0, *v1, *v2, *v3, *v4, *v5, *v6, *v7;
			float temp[3] = { 0.0f, 0.5f, 0.0f }, temp0[3], temp1[3], temp2[3], temp3[3], temp4[3], temp5[3], temp6[3], temp7[3];
			glMaterialfv(GL_FRONT,GL_AMBIENT,temp);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,temp);
			glPolygonMode(GL_FRONT, GL_LINE);
			glBegin(GL_QUADS);
			v0 = mesh->gate->vertex;
			v1 = mesh->gate->next->next->next->vertex;
			v2 = mesh->gate->next->next->vertex;
			v3 = mesh->gate->next->vertex;
			v4 = mesh->gate->next->next->spin->inv->next->vertex;
			v5 = mesh->gate->next->next->spin->inv->next->next->vertex;
			v6 = mesh->gate->spin->inv->next->vertex;
			v7 = mesh->gate->spin->inv->next->next->vertex;

			VecAdd3fv(temp, v0->v, v1->v);
			VecSelfAdd3fv(temp, v2->v);
			VecSelfAdd3fv(temp, v3->v);
			VecSelfAdd3fv(temp, v4->v);
			VecSelfAdd3fv(temp, v5->v);
			VecSelfAdd3fv(temp, v6->v);
			VecSelfAdd3fv(temp, v7->v);
			VecSelfScalarDiv3fv(temp, 8.0f);

			VecSubtract3fv(temp0, temp, v0->v);
			VecSelfScalarMult3fv(temp0, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp0, v0->v);

			VecSubtract3fv(temp1, temp, v1->v);
			VecSelfScalarMult3fv(temp1, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp1, v1->v);

			VecSubtract3fv(temp2, temp, v2->v);
			VecSelfScalarMult3fv(temp2, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp2, v2->v);

			VecSubtract3fv(temp3, temp, v3->v);
			VecSelfScalarMult3fv(temp3, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp3, v3->v);

			VecSubtract3fv(temp4, temp, v4->v);
			VecSelfScalarMult3fv(temp4, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp4, v4->v);

			VecSubtract3fv(temp5, temp, v5->v);
			VecSelfScalarMult3fv(temp5, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp5, v5->v);

			VecSubtract3fv(temp6, temp, v6->v);
			VecSelfScalarMult3fv(temp6, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp6, v6->v);

			VecSubtract3fv(temp7, temp, v7->v);
			VecSelfScalarMult3fv(temp7, SCALE_NEIGHBORS_LINE);
			VecSelfAdd3fv(temp7, v7->v);

			glNormal3fv(mesh->gate->inv->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp3);
			glVertex3fv(temp2);
			glVertex3fv(temp1);

			glNormal3fv(mesh->gate->spin->normal->n);
			glVertex3fv(temp1);
			glVertex3fv(temp6);
			glVertex3fv(temp7);
			glVertex3fv(temp0);

			glNormal3fv(mesh->gate->next->spin->normal->n);
			glVertex3fv(temp0);
			glVertex3fv(temp7);
			glVertex3fv(temp4);
			glVertex3fv(temp3);

			glNormal3fv(mesh->gate->next->next->spin->normal->n);
			glVertex3fv(temp3);
			glVertex3fv(temp4);
			glVertex3fv(temp5);
			glVertex3fv(temp2);

			glNormal3fv(mesh->gate->next->next->next->spin->normal->n);
			glVertex3fv(temp2);
			glVertex3fv(temp5);
			glVertex3fv(temp6);
			glVertex3fv(temp1);

			glNormal3fv(mesh->gate->spin->inv->next->next->spin->normal->n);
			glVertex3fv(temp6);
			glVertex3fv(temp5);
			glVertex3fv(temp4);
			glVertex3fv(temp7);
			glEnd();
		}
	}
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE); 

	if (mesh->active)
	{
		glColor3f(0.5f, 0.5f, 0.9f);
		glLineWidth(6.0);
		glBegin(GL_LINES);
		glVertex3fv(mesh->active->vertex->v);
		glVertex3fv(mesh->active->inv->vertex->v);
		glEnd();
	}

	if ( WireFrameOn )
	{
		CFspinEdge *e0, *e1, *e2;
		glLineWidth(1.0f);
		glBegin(GL_LINES);
		if ( BoundaryOn )
		{
			if ( BoundaryOn == 3 )
			{
				glColor3f(1.0f, 1.0f, 1.0f);
			}
			else
			{
				glColor3f(0.0f, 0.0f, 0.0f);
			}

			for  (i = mesh->interior_faces; i < mesh->total_faces; i++)
			{
				e0 = (CFspinEdge*) mesh->fs->elementAt(i);
				e1 = e0;
				do
				{
					e2 = e1->next;
					glVertex3f(e1->vertex->v[0]+e0->normal->n[0]*0.001,e1->vertex->v[1]+e0->normal->n[1]*0.001,e1->vertex->v[2]+e0->normal->n[2]*0.001);
					glVertex3f(e2->vertex->v[0]+e0->normal->n[0]*0.001,e2->vertex->v[1]+e0->normal->n[1]*0.001,e2->vertex->v[2]+e0->normal->n[2]*0.001);
					e1 = e2;
				} while (e1 != e0);
			}
		}
		glEnd();
	}

	if ( VertexNormalsOn )
	{
		CFspinEdge *e;
		glColor3f(1.0f, 1.0f, 1.0f);
		glLineWidth(1.0f);
		glBegin(GL_LINES);
		if ( BoundaryOn )
		{
			for  (i = mesh->interior_edges; i < mesh->total_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				glVertex3fv(e->vertex->v);
				glVertex3f(e->vertex->v[0]+e->normal->n[0]*0.05,e->vertex->v[1]+e->normal->n[1]*0.05,e->vertex->v[2]+e->normal->n[2]*0.05);
			}
		}
		if ( InteriorOn )
		{
			for  (i = 0; i < mesh->interior_edges; i++)
			{
				e = (CFspinEdge*) mesh->es->elementAt(i);
				glVertex3fv(e->vertex->v);
				glVertex3f(e->vertex->v[0]+e->normal->n[0]*0.05,e->vertex->v[1]+e->normal->n[1]*0.05,e->vertex->v[2]+e->normal->n[2]*0.05);
			}
		}
		glEnd();
	}
}

void myDisplay()
{
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
  glViewport(0,0,WindowW,WindowH);
  glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(30,(float)WindowW/WindowH,NearPlane,FarPlane);
  glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

//	glRotatef(-90,0,0,1);

	glLightfv(GL_LIGHT0,GL_POSITION,LightFrom);

	gluLookAt(DistX,DistY,DistZ, DistX,DistY,0, 0,1,0);
	glRotatef(Elevation,1,0,0);
	glRotatef(Azimuth,0,1,0);
	
	if (mesh->ts->size())
	{
		myTetDisplay();
	}
	else if (mesh->hs->size())
	{
		myHexDisplay();
	}
	else
	{
		myPolDisplay();
	}

//	displayMessage( );
    
  glutSwapBuffers();
}

void myReshape(int w, int h)
{
  glutReshapeWindow(WindowW,WindowH);
}

void myIdle()
{
  if (CodingAnimationOn)
  {
		if (CodingAnimationSpeed == 0)
		{
			CodingAnimationOn &= encodeStep();
		}
    else if (CodingAnimationSpeed == 1)
		{
			for (int i = 0; i < 10; i++)
			{
				if (CodingAnimationOn)
				{
					CodingAnimationOn &= encodeStep();
				}
			}
		}
		else if (CodingAnimationSpeed == 2)
		{
			for (int i = 0; i < 100; i++)
			{
				if (CodingAnimationOn)
				{
					CodingAnimationOn &= encodeStep();
				}
			}
		}
		else if (CodingAnimationSpeed == 3)
		{
			for (int i = 0; i < 250; i++)
			{
				if (CodingAnimationOn)
				{
					CodingAnimationOn &= encodeStep();
				}
			}
		}
		else if (CodingAnimationSpeed == 4)
		{
			for (int i = 0; i < 2500; i++)
			{
				if (CodingAnimationOn)
				{
					CodingAnimationOn &= encodeStep();
				}
			}
		}
	}
  glutPostRedisplay();
}

void myMouseFunc(int button, int state, int x, int y)
{
  NewX=x; NewY=y;

  if (button == GLUT_LEFT_BUTTON)
    {
      LeftButtonDown =! LeftButtonDown;
    }
  else if (button == GLUT_RIGHT_BUTTON)
    {
      RightButtonDown =! RightButtonDown;
    }
  else if (button == GLUT_MIDDLE_BUTTON)
    {
      MiddleButtonDown =! MiddleButtonDown;
    }
}

void myMotionFunc(int x, int y)
{
  OldX=NewX; OldY=NewY;
  NewX=x;    NewY=y;
  
  float RelX = (NewX-OldX) / (float)glutGet((GLenum)GLUT_WINDOW_WIDTH);
  float RelY = (NewY-OldY) / (float)glutGet((GLenum)GLUT_WINDOW_HEIGHT);
  if (LeftButtonDown) 
  { 
    if (InteractionMode == 0)
		{
			Azimuth += (RelX*180);
			Elevation += (RelY*180);
		}
    else if (InteractionMode == 1)
		{
			DistX-=RelX;
			DistY+=RelY;
		}
    else if (InteractionMode == 2)
		{
			DistZ-=RelY*DistZ;
		}
    else if (InteractionMode == 3)
		{
			NearPlane *= (1.0f-RelY);
			if (NearPlane < 0.005f) NearPlane = 0.005f;
			if (NearPlane > FarPlane) NearPlane = FarPlane;
		}
    else if (InteractionMode == 4)
		{
			FarPlane *= (1.0f-RelY);
			if (FarPlane < NearPlane) FarPlane = NearPlane;
			if (FarPlane > 5.0f) FarPlane = 5.0f;
		}
  }
  else if (MiddleButtonDown)
  {
    DistX-=RelX*1.0f;
    DistY+=RelY*1.0f;
  }
}

void InitLights()
{
  GLfloat ambient[] = {0.5, 0.5, 0.5, 1.0};
  GLfloat diffuse[] = {1.0, 1.0, 1.0, 1.0};
  GLfloat specular[] = {0.5, 0.5, 0.5, 1.0};

  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
}

void MyMenuFunc(int value)
{
  if (value == 1)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("bump2");
      loader->loadHex("hex/bump2.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
			Azimuth =  180.0f;
			Elevation =  136.876f;
			DistX =  -0.0296875f; DistY = 0.142708f; DistZ = 2.91108f;
	}
  else if (value == 2)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("hanger.hex");
      loader->loadHex("hex/hanger.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
			Azimuth =  705.703f;
			Elevation =  -141.41f;
			DistX =  0.0200521f; DistY = -0.0582841f; DistZ = 2.22766f;
    }
  else if (value == 3)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("test");
      loader->loadHex("hex/test.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
			Azimuth =  232.874f;
			Elevation =  203.492f;
			DistX =  0.0757812f; DistY = -0.182292f; DistZ = 3.00785f;
	}
  else if (value == 4)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("shaft");
      loader->loadHex("hex/shaft.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
			Azimuth =  237.234f;
			Elevation =  4.313f;
			DistX =  0.0148437f; DistY = 0.0427083f; DistZ = 3.13438f;
	}
  else if (value == 5)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("mdg-1b");
      loader->loadHex("hex/mdg-1b.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
			Azimuth =  408.938f;
			Elevation =  21.8761f;
			DistX =  0.0335938f; DistY = -0.0937999f; DistZ = 3.11323f;
	}
  else if (value == 6)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("fru");
      loader->loadHex("hex/fru.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
			Azimuth =  539.859f;
			Elevation =  -240.159f;
			DistX =  -0.0361979f; DistY = -0.0842762f; DistZ = 3.37027f;
	}
  else if (value == 7)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("hutch.hex");
      loader->loadHex("hex/hutch.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
			Azimuth =  720.281f;
			Elevation =  294.59f;
			DistX =  0.00937501f; DistY = -0.165229f; DistZ = 3.01486f;
	}
  else if (value == 8)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("random.tet");
      loader->loadTet("tet/random.tet", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesTETnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
			Azimuth =  0.0f;
			Elevation =  0.0f;
			DistX =  0.0f; DistY = 0.0f; DistZ = 3.0f;
   }
  else if (value == 9)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("silva13k.tet");
      loader->loadTet("tet/silva13k.tet", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesTETnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
			Azimuth =  0.0f;
			Elevation =  0.0f;
			DistX =  0.0f; DistY = 0.0f; DistZ = 3.0f;
    }
  else if (value == 201)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("out2e+07.tet");
      loader->loadTet("tet/out2e+07.tet", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesTETnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
    }
  else if (value == 202)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("tet9.tet");
      loader->loadTet("tet/tet9.tet", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesTETnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
   }
  else if (value == 203)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("reg005.tet");
      loader->loadTet("tet/reg005.tet", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesTETnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
    }
  else if (value == 204)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("warped");
      loader->loadHex("hex/warped.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
    }
  else if (value == 205)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("ra.hex");
      loader->loadHex("hex/ra.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
    }
  else if (value == 206)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("c2");
      loader->loadHex("hex/c2.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
    }
  else if (value == 207)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("c1");
      loader->loadHex("hex/c1.hex", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
    }
  else if (value == 208)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("isa_grid3");
      loader->loadPol("isa_grid3.pol", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
//			countMeshDegrees(mesh);
      if (codec) delete codec;
      codec = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
    }
  else if (value == 209)
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("isa_test");
      loader->loadPol("isa_test.pol", mesh);
      mesh->computePerFaceNormals();
      mesh->normalize();
//			countMeshDegrees(mesh);
      if (codec) delete codec;
      codec = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
			BoundaryOn = 1;
    }
  else if (211) // curve linear grids were only added for reviewing a paper
    {
      if (mesh) delete mesh;
      mesh = new CFmesh("post.clg");
//      loader->loadClg("clg/blunt.x", mesh); 
//      loader->loadClg("clg/delta-40.x", mesh); 
      loader->loadClg("clg/post.x", mesh); 
      mesh->computePerFaceNormals();
      mesh->normalize();
			countMeshDegrees(mesh);
			encoder->verticesTETnumber = 0;
			encoder->verticesHEXnumber = 0;
      if (codec) delete codec;
      codec = 0;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
    }
  else if (value == 10)
    {
      VertexNormalsOn = 0;
    }
  else if (value == 11)
    {
      VertexNormalsOn = 1;
    }
  else if (value == 12)
    {
      VertexNormalsOn = 2;
    }
  else if (value == 13)
    {
      VertexNormalsOn = 3;
    }
  else if (value == 14)
    {
//      mesh->computePerVertexNormals();
    }
  else if (value == 15)
    {
      mesh->computePerFaceNormals();
    }
  else if (value == 30)
    {
      ColourMode = 0; // per material
    }
  else if (value == 31)
    {
      ColourMode = 1; // per group
    }
  else if (value == 32)
    {
      ColourMode = 2; // per component
    }
  else if (value == 33)
    {
      ColourMode = 3; // show triangles
    }
  else if (value == 34)
    {
      ColourMode = 4; // show quadrangles
    }
  else if (value == 35)
    {
      ColourMode = 5; // show pentagons
    }
  else if (value == 36)
    {
      ColourMode = 6; // show hexagons
    }
  else if (value == 37)
    {
      ColourMode = 7; // show biggergons
    }
  else if (value == 40)
    {
      WireFrameOn =! WireFrameOn;
    }
  else if (value == 41)
    {
      ViewMode = 0; // view entire model
    }
  else if (value == 42)
    {
      ViewMode = 1; // view single group
    }
  else if (value == 43)
    {
      ViewMode = 2; // view single component
    }
  else if (value == 44)
    {
      if (ViewMode == 1)
			{
				ViewModeGroup = 1; // view first group
			}
      else if (ViewMode == 2)
			{
				ViewModeComponent = 1; // view first component
			}
	}
  else if (value == 45)
	{
		if (ViewMode == 1)
		{
			ViewModeGroup++; // view next group
			if (ViewModeGroup >= mesh->gs->size())
			{
				ViewModeGroup = 1;
			}
		}
		else if (ViewMode == 2)
		{
			ViewModeComponent++; // view next component
			if (ViewModeComponent > mesh->NumComponents)
			{
				ViewModeComponent = 1;
			}
		}
	}
  else if (value == 51)
	{
		CodingAnimationStep++;
//		encodeMeshAbs(CodingAnimationStep);
	}
  else if (value == 52)
	{
		CodingAnimationStep--;
		if (CodingAnimationStep < 0)
		{
			CodingAnimationStep = 0;
		}
//		encodeMeshAbs(CodingAnimationStep);
	}
  else if (value == 53)
	{
		if (CodingAnimationDirection == 1)
		{
			CodingAnimationDirection = 0;
		}
		else
		{
			CodingAnimationDirection = 1;
		}
	}
  else if (value == 54)
	{
		CodingAnimationOn = 1;
		CodingAnimationSpeed = 0;
	}
  else if (value == 55)
	{
		CodingAnimationOn = 1;
		CodingAnimationSpeed = 1;
	}
  else if (value == 56)
	{
		CodingAnimationOn = 1;
		CodingAnimationSpeed = 2;
	}
  else if (value == 57)
	{
		CodingAnimationOn = 0;
	}
  else if (value == 58)
	{
		CodingAnimationOn = 0;
		CodingAnimationSpeed = 0;
		CodingAnimationDirection = 1;
		CodingAnimationStep = 0;

		CFspinEdge* e;
		if (codec) delete codec;
		codec = 0;
		for (int i = 0; i < mesh->es->size(); i++)
		{
			e = (CFspinEdge*) mesh->es->elementAt(i);
		}
	}
  else if (value == 100)
	{
		InteractionMode = 0; // rotate
	}
  else if (value == 101)
	{
		InteractionMode = 1; // translate
	}
  else if (value == 102)
	{
		InteractionMode = 2; // zoom      
	}
  else if (value == 103)
	{
  }
  else if (value == 104)
    {
      Elevation = 0;
      Azimuth = 0;
      DistX = 0;
      DistY = 0;
      DistZ = 3;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      InteractionMode = 0;
      WireFrameOn = 0;
      VertexNormalsOn = 0;
      InteractionMode = 0;
      ColourMode = 0;
      MyMenuFunc(58);
   }
  else if (value == 105)
    {
      exit(0);
    }
}

void myKeyboard(unsigned char Key, int x, int y)
{
  switch(Key)
    {
    case 27 :
    case 'Q':
    case 'q': exit(0);
    case ' ': // rotate, translate, or zoom
      if (InteractionMode > 1)
			{
				InteractionMode = 0;
			}
			else
			{
				InteractionMode += 1;
			}
      break;
    case 'R': // recenter
			if (mesh) mesh->normalizeUsingCell(mesh->gate);
      break;
    case 'r': // recenter
			if (mesh) mesh->normalizeUsingActiveVertices();
      break;
    case '>': //zoom in
      DistZ-=0.1f*DistZ;
      break;
    case '<': //zoom out
      DistZ+=0.1f*DistZ;
      break;
    case ',': // next
      if (mesh->active) mesh->active = mesh->active->next;
			cout << "active: " << mesh->active->vertex->index << " " << mesh->active->inv->vertex->index << " " << mesh->active->slots << " (" << mesh->active->degree << ")" << endl;
      break;
    case '.': // inv
      if (mesh->active) mesh->active = mesh->active->inv;
			cout << "active: " << mesh->active->vertex->index << " " << mesh->active->inv->vertex->index << " " << mesh->active->slots << " (" << mesh->active->degree << ")" << endl;
      break;
    case '/': // spin
      if (mesh->active) mesh->active = mesh->active->spin;
			cout << "active: " << mesh->active->vertex->index << " " << mesh->active->inv->vertex->index << " " << mesh->active->slots << " (" << mesh->active->degree << ")" << endl;
      break;
    case 'G':
    case 'g': // view single <g>roup
      if (ViewMode == 1)
			{
				MyMenuFunc(41);
			}
      else
			{
				MyMenuFunc(42);
			}
      break;
    case 'T':
    case 't':
      if (TetrahedraOn)
			{
				TetrahedraOn = 0;
			}
			else
			{
				TetrahedraOn = 1;
			}
      break;
      break;
    case 'Y':
    case 'y': // encode step by step plus 5
      MyMenuFunc(52);
      break;
    case 'U':
    case 'u': // encode step by step plus 10
      MyMenuFunc(58);
      break;
    case 'P':
    case 'p': // play encode animation at different speeds
      if (CodingAnimationOn)
			{
				if (CodingAnimationSpeed == 4)
				{
					CodingAnimationSpeed = 0;
				}
				else
				{
					CodingAnimationSpeed++;
				}
			}
      else
			{
				CodingAnimationSpeed = 0;
				CodingAnimationOn = 1;
				CodingOn = 1;
				if (BoundaryOn == 1) BoundaryOn++;
			}
      break;
    case 'O':
    case 'o': // stop encode animation
      CodingAnimationOn = 0;
			CodingAnimationSpeed = 0;
      break;
    case 'B':
    case 'b':
      if (BoundaryOn == 3)
			{
				BoundaryOn = 0;
			}
			else
			{
				BoundaryOn++;
			}
      break;
    case 'I':
    case 'i':
      if (InteriorOn == 3)
			{
				InteriorOn = 0;
			}
			else
			{
				InteriorOn++;
			}
      break;
    case 'C':
      if (CodingOn)
			{
				CodingOn = 0;
			}
			else
			{
				CodingOn = 1;
			}
      break;
    case 'c': // view single <c>omponent
			encodeStep();
			BoundaryOn = 2;
			if (CodingOn == 0)
			{
				CodingOn = 1;
//				mesh->normalizeUsingActiveVertices();
			}
      break;
    case 'W':
    case 'w': // view <w>ireframe
      MyMenuFunc(40);
      break;
    case 'D':
    case 'd': // <d>ump
      MyMenuFunc(103);
      break;
    case 'F':
			if (InteractionMode == 4)
			{
				InteractionMode = 0;
			}
			else
			{
				InteractionMode = 4;
			}
		break;
    case 'f': // view <f>irst
      MyMenuFunc(44);
      break;
    case 'V':
    case 'v':
      if (VertexNormalsOn == 1)
			{
				VertexNormalsOn = 0;
			}
      else
			{
				VertexNormalsOn = 1;
			}
			break;
      break;
    case 'N':
			if (InteractionMode == 3)
			{
				InteractionMode = 0;
			}
			else
			{
				InteractionMode = 3;
			}
			break;
    case 'n': // view <n>ext
      MyMenuFunc(45);
      break;
    case 'S':
    case 's': // u<s>e materials
      MyMenuFunc(30);
      break;
    case 'E':
      if (EdgeDegreesOn == 1)
			{
				EdgeDegreesOn = 0;
			}
      else
			{
				EdgeDegreesOn = 1;
			}
			break;
    case 'e': // random p<e>r group
      MyMenuFunc(31);
      break;
    case 'M':
    case 'm': // random per co<m>ponent
      MyMenuFunc(32);
      break;
    case 'Z':
    case 'z': // red triangles <z>
      MyMenuFunc(33);
      break;
    case 'X':
    case 'x': // red quadrangles <x>
			mesh->gate = 0;
			CodingOn = 0;
      break;
    case 'H':
    case 'h': // red pentagons <h>
      MyMenuFunc(35);
      break;
    case 'J':
    case 'j': // red hexagons <j>
      MyMenuFunc(36);
      break;
    case 'K':
    case 'k': // red biggergons <k>
      cout << "Azimuth = "  << " " << Azimuth << "f;" << endl;
      cout << "Elevation = " << " " << Elevation << "f;" << endl;
      cout << "DistX = " << " " << DistX << "f; DistY = " << DistY << "f; DistZ = " << DistZ << "f;" << endl;
      break;
    case 'L':
    case 'l':
      break;
    case 'A':
    case 'a':
      break;
    case '1':
      MyMenuFunc(1);
      break;
    case '2':
      MyMenuFunc(2);
      break;
    case '3':
      MyMenuFunc(3);
      break;
    case '#':
      MyMenuFunc(203);
      break;
    case '4':
      MyMenuFunc(4);
      break;
    case '$':
      MyMenuFunc(204);
      break;
    case '5':
      MyMenuFunc(5);
      break;
    case '%':
      MyMenuFunc(205);
      break;
    case '6':
      MyMenuFunc(6);
      break;
    case '^':
      MyMenuFunc(206);
      break;
    case '7':
      MyMenuFunc(7);
      break;
    case '&':
      MyMenuFunc(207);
      break;
    case '8':
      MyMenuFunc(8);
      break;
    case '*':
      MyMenuFunc(208);
      break;
    case '9':
      MyMenuFunc(9);
      break;
    case '(':
      MyMenuFunc(209);
      break;
    case ')':
      MyMenuFunc(210);
      break;
    case '0':
      MyMenuFunc(211);
      break;
  };
}

void InitLightModel()
{
  float GlobalAmbient[] = {0.0f, 0.3f, 0.0f, 1.0f};

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,GlobalAmbient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
}

void InitMaterials()
{
  colours_ambient[0][0] = 0.0f; colours_ambient[0][1] = 0.0f; colours_ambient[0][2] = 0.0f; colours_ambient[0][3] = 1.0f; // black
  colours_ambient[1][0] = 0.3f; colours_ambient[1][1] = 0.0f; colours_ambient[1][2] = 0.0f; colours_ambient[1][3] = 1.0f; // red
  colours_ambient[2][0] = 0.0f; colours_ambient[2][1] = 0.3f; colours_ambient[2][2] = 0.0f; colours_ambient[2][3] = 1.0f; // green
  colours_ambient[3][0] = 0.0f; colours_ambient[3][1] = 0.0f; colours_ambient[3][2] = 0.3f; colours_ambient[3][3] = 1.0f; // blue
  colours_ambient[4][0] = 0.3f; colours_ambient[4][1] = 0.3f; colours_ambient[4][2] = 0.0f; colours_ambient[4][3] = 1.0f; // yellow
  colours_ambient[5][0] = 0.3f; colours_ambient[5][1] = 0.0f; colours_ambient[5][2] = 0.3f; colours_ambient[5][3] = 1.0f; // purple
  colours_ambient[6][0] = 0.0f; colours_ambient[6][1] = 0.3f; colours_ambient[6][2] = 0.3f; colours_ambient[6][3] = 1.0f; // cyan
  colours_ambient[7][0] = 0.3f; colours_ambient[7][1] = 0.3f; colours_ambient[7][2] = 0.3f; colours_ambient[7][3] = 1.0f; // white
  colours_ambient[8][0] = 0.2f; colours_ambient[8][1] = 0.2f; colours_ambient[8][2] = 0.3f; colours_ambient[8][3] = 1.0f; // light blue
  colours_ambient[9][0] = 0.4f; colours_ambient[9][1] = 0.2f; colours_ambient[9][2] = 0.3f; colours_ambient[9][3] = 1.0f; // violett

  colours_diffuse[0][0] = 0.0f; colours_diffuse[0][1] = 0.0f; colours_diffuse[0][2] = 0.0f; colours_diffuse[0][3] = 1.0f; // black
  colours_diffuse[1][0] = 0.6f; colours_diffuse[1][1] = 0.0f; colours_diffuse[1][2] = 0.0f; colours_diffuse[1][3] = 1.0f; // red
  colours_diffuse[2][0] = 0.0f; colours_diffuse[2][1] = 0.6f; colours_diffuse[2][2] = 0.0f; colours_diffuse[2][3] = 1.0f; // green
  colours_diffuse[3][0] = 0.0f; colours_diffuse[3][1] = 0.0f; colours_diffuse[3][2] = 0.6f; colours_diffuse[3][3] = 1.0f; // blue
  colours_diffuse[4][0] = 0.6f; colours_diffuse[4][1] = 0.6f; colours_diffuse[4][2] = 0.0f; colours_diffuse[4][3] = 1.0f; // yellow
  colours_diffuse[5][0] = 0.6f; colours_diffuse[5][1] = 0.0f; colours_diffuse[5][2] = 0.6f; colours_diffuse[5][3] = 1.0f; // purple
  colours_diffuse[6][0] = 0.0f; colours_diffuse[6][1] = 0.6f; colours_diffuse[6][2] = 0.6f; colours_diffuse[6][3] = 1.0f; // cyan
  colours_diffuse[7][0] = 0.6f; colours_diffuse[7][1] = 0.6f; colours_diffuse[7][2] = 0.6f; colours_diffuse[7][3] = 1.0f; // white
  colours_diffuse[8][0] = 0.2f; colours_diffuse[8][1] = 0.2f; colours_diffuse[8][2] = 0.6f; colours_diffuse[8][3] = 1.0f; // light blue
  colours_diffuse[9][0] = 0.9f; colours_diffuse[9][1] = 0.4f; colours_diffuse[9][2] = 0.7f; colours_diffuse[9][3] = 1.0f; // violett

  colours_specular[0][0] = 0.0f; colours_specular[0][1] = 0.0f; colours_specular[0][2] = 0.0f; colours_specular[0][3] = 1.0f; // none
  colours_specular[1][0] = 0.3f; colours_specular[1][1] = 0.3f; colours_specular[1][2] = 0.3f; colours_specular[1][3] = 1.0f; // little
  colours_specular[2][0] = 0.7f; colours_specular[2][1] = 0.7f; colours_specular[2][2] = 0.7f; colours_specular[2][3] = 1.0f; // fairly
  colours_specular[3][0] = 1.0f; colours_specular[3][1] = 1.0f; colours_specular[3][2] = 1.0f; colours_specular[3][3] = 1.0f; // full

  colours_emission[0][0] = 0.0f; colours_emission[0][1] = 0.0f; colours_emission[0][2] = 0.0f; colours_emission[0][3] = 0.0f; // none
  colours_emission[1][0] = 0.5f; colours_emission[1][1] = 0.0f; colours_emission[1][2] = 0.0f; colours_emission[1][3] = 0.0f; // little red
  colours_emission[2][0] = 0.0f; colours_emission[2][1] = 0.5f; colours_emission[2][2] = 0.0f; colours_emission[2][3] = 0.0f; // little green
  colours_emission[3][0] = 0.0f; colours_emission[3][1] = 0.0f; colours_emission[3][2] = 0.5f; colours_emission[3][3] = 0.0f; // little blue
  colours_emission[4][0] = 0.4f; colours_emission[4][1] = 0.2f; colours_emission[4][2] = 0.3f; colours_emission[4][3] = 1.0f; // violett

  glMaterialf(GL_FRONT,GL_SHININESS,50);
}

void main(void)
{
  loader = new CFloader();
  encoder = new CFencoder();
  
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(WindowW,WindowH);
  glutInitWindowPosition(180,100);
  glutCreateWindow("Edge degree-based Volume Mesh Compressor");

  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);
  
  InitLightModel();
  InitLights();
  InitMaterials();

  glutDisplayFunc(myDisplay);
  glutReshapeFunc(myReshape);
  glutIdleFunc(myIdle);

  glutMouseFunc(myMouseFunc);
  glutMotionFunc(myMotionFunc);
  glutKeyboardFunc(myKeyboard);


  MyMenuFunc(1);

  glutMainLoop();
}
