/*
===============================================================================

	FILE:	spinedge.h
	
	CONTENTS:
	
    An extension of the "twin-edge" (or half-edge) from the surface mesh case
    to the volume mesh case. The "spin-edge" has an additional field besides
    and inv and a next pointer (which connect edges and faces into a cell). The
    spin pointer connects cells around an edge together.
   
	PROGRAMMERS:
	
		martin isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		(c) 2002 martin isenburg@cs.unc.edu
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
		20 May 2002 -- created
	
===============================================================================
*/
#ifndef CF_SPINEDGE_H
#define CF_SPINEDGE_H

class CFvertex;
class CFnormal;

class CFspinEdge
{
public:
  int group;
  int material;
  int component;
  int degree;
  int slots;
	int is_on_boundary;

  __int64 marker; // used to visualize encoded faces

  CFvertex* vertex; // corner properties
  CFnormal* normal;

  CFspinEdge* next; // connectivity
  CFspinEdge* inv;
  CFspinEdge* spin;

  CFspinEdge* Lnext;

  CFspinEdge();
  CFspinEdge(int g, int m);
  CFspinEdge(int g, int m, CFvertex* v);
  CFspinEdge(int g, int m, CFvertex* v, CFnormal* n);
  ~CFspinEdge();
};

#endif
